#!/usr/bin/clisp -ansi -q -q
(setq *print-case* :downcase)
;;;
;;; Design note:
;;;
;;; Normally, one prefers the `-modern' option for a `case-sensitive'
;;; world; in which case, the top two lines above should instead read:
;;;
;;;   #!/usr/bin/clisp -ansi -modern -q -q
;;;
;;; However, I got errors when running in `--gui' mode (under clisp 2.49):
;;;
;;;   (root-shell)# wp-mirror --gui
;;;   WARNING: One should not change the case sensitiveness of #<package xlib>.
;;;   WARNING: One should not change the case inversion of #<package xlib>.
;;;   *** - subtypep: invalid type specification xlib::|x-type-error|
;;;
;;; First two errors appear while executing    `(require "clx")'.
;;; Last error (fatal) appears while executing `(xlib:display-roots display)'.
;;;
;;; I filed Debian Bug Report: #691519
;;; 

;;;;--------------------------------------------------------------------------+
;;;; WP-MIRROR is a free utility for mirroring a set of Wikipedias.           |
;;;;--------------------------------------------------------------------------+


(defconstant *copyright-license-and-disclaimer*
"Copyright (C) 2011,2012 Dr. Kent L. Miller.  All rights reserved.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY\; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>")


(defconstant *version* "wp-mirror 0.5")


;;;;--------------------------------------------------------------------------+
;;;; Internationalization:                                                    |
;;;;--------------------------------------------------------------------------+


(setf (i18n:textdomaindir "wp-mirror") "./")
(defun _ (msgid) (i18n:gettext msgid "wp-mirror"))


;;;;--------------------------------------------------------------------------+
;;;; Libraries:                                                               |
;;;;  asdf   - process other libraries                                        |
;;;;  getopt - process command-line arguments                                 |
;;;;  md5    - compute md5sum                                                 |
;;;; Modules:                                                                 |
;;;;  clx    - provide xlib bindings                                          |
;;;;--------------------------------------------------------------------------+

;;;
;;; Design note:
;;;
;;; `clisp' uses `common-lisp-controller' to load libraries.
;;;
;;; Libraries are loaded with  (clc:clc-require :name-of-library).
;;; Modules   are loaded with  (require "clx").
;;;

(eval-when (:compile-toplevel :execute :load-toplevel)
	   (unless (member :common-lisp-controller *features*)
	     (load #P"/usr/share/common-lisp/source/common-lisp-controller/common-lisp-controller.lisp")))

(unless (member :common-lisp-controller *features*)
  (common-lisp-controller:compile-common-lisp-controller-v5 "clisp")
  (common-lisp-controller:init-common-lisp-controller-v5 "clisp"))

;;;
;;; Design note:
;;;
;;; `common-lisp-controller' makes use of `asdf' (another system
;;; definition facility).  It can happen that an error message similar
;;; to the following appears:
;;;
;;;   *** - read from #<input buffered file-stream character
;;;   #P"./wp-mirror-0.2.lisp" @47>: there is no package with name "ASDF"
;;;
;;; Most likely, the `cl-asdf' package was either not installed or not
;;; configured.  Its documentation is probably under
;;; `/usr/share/doc/cl-asdf/'.  Look for the file `README.Debian' or a
;;; directory containing HTML pages.  They should have instructions for
;;; configuring ASDF.  For Debian GNU/Linux (and its derivatives) the
;;; next two lines should work.
;;;

;;(unless (member :asdf2 *features*)
;;  (load #P"/usr/share/common-lisp/source/cl-asdf/asdf.lisp"))
;;(when (null asdf:*central-registry*)
;;  (push #P"/usr/share/common-lisp/systems/" asdf:*central-registry*))

;;; If you are installing your own packages elsewhere, then provide
;;; additional lines that push the relevant path(s) onto
;;; `custom:*load-paths*'.  Note that a directory name must have a
;;; trailing slash.

;;(push #P"/usr/share/common-lisp/systems/" custom:*load-paths*)


;;; The `asdf2' library should have been loaded by `common-lisp-controller'.

(unless (member :asdf2 *features*)
  (error (_ "The :asdf2 package is missing.  Aborting.")))

;;; Now we load the remaining libraries.

(clc:clc-require :getopt)

;;; The md5 package runs far too slowly, so 
;;; o we use it only for short strings (such as file names),
;;; o we do not use it for file contents (we use shell commands instead).
(clc:clc-require :md5)

;;; Debian packages: cl-sql, cl-sql-uffi, and cl-sql-mysql all seem
;;; broken, so 
;;; o we do not use it for database access,
;;; o we use shell commands instead.

;;; Now we load the modules.

;;; The clx module is needed for the --gui option
(require "clx")


;;;;--------------------------------------------------------------------------+
;;;; Constants:                                                               |
;;;;   *author*                                                               |
;;;;   *copyright-license-and-disclaimer* (see above)                         |
;;;;   *help*                                                                 |
;;;;   *version* (see above)                                                  |
;;;;   *wtf*                                                                  |
;;;; The values of these constants are used by the build process.             |
;;;; The utility `help2man' generates the `wp-mirror.1.gz' man page out       |
;;;; of `--help' and `--version' output.  See also the functions              |
;;;; `put-help-and-die' and `put-version-and-die' below.                      |
;;;;--------------------------------------------------------------------------+


(defconstant *author*
"Written by Dr. Kent L. Miller.")  


(defconstant *help* 
"The WikiMedia Foundation offers Wikipedias in nearly 300 languages.

Wp-mirror is a free utility for mirroring any desired set of these
wikipedias. That is, it builds a wikipedia farm that the user can
browse locally.  Many users need such off-line access, often for
reasons of availability, mobility, and privacy.  Wp-mirror builds a
complete mirror with original size images. Wp-mirror is robust and
uses check-pointing to resume after interruption.
 
Wp-mirror by default mirrors the `simple' wikipedia (Simple English
means shorter sentences).  The default should work `out-of-the-box'
with no user configuration.  It should build in 200ks (two days),
occupy 60G of disk space, be served locally by a virtual host
<http://simple.wpmirror.site/>, and update automatically every week.
The default should be suitable for anyone who learned English as a
second language (ESL).

The top ten wikipedias are: en, de, fr, nl, it, es, pl, ru, ja, and
pt. Because wp-mirror uses original size image files, the top ten are
too large to fit on a laptop with a single 500G disk, unless the user
does not need the images (and this is configurable).  The `en'
wikipedia is the most demanding case.  It should build in 5Ms (two
months), occupy 3T of disk space, be served locally by a virtual host
<http://en.wpmirror.site/>, and update automatically every month.

Homepage: <http://www.nongnu.org/wp-mirror/>

Usage: wp-mirror [OPTIONS] [CONFIGFILE]

Options:

  --add language-code 
                     add given language to mirror, and exit.
  --drop language-code 
                     drop given language from mirror, and exit.
  --help             display this help, and exit.
  --mirror           build mirror, and exit.
  --version          display version and copyright, and exit.

Options for monitor mode:

  --gui              operate monitor in gui mode (default).
  --monitor          operate in monitor mode (gui default).
  --screen           operate monitor in screen mode.
  --text             operate monitor in text mode.

Options used during build and install:

  --copyright        display copyright file, and exit.

Options for developers:

  --debug            run in debug mode (more verbose).
  --delete language-code  
                     delete files and state info, for given language, and exit.
  --dump language-code   
                     dump the database, to file `xxwiki.sql' in working directory (where `xx' stands for the language code), for a given language, and exit.  If the language-code is `template', then the empty database `wikidb' will be dumped to `database_farm.sql'.
  --info             display parameters, and exit.
  --restore-default  drop all databases and files (except images), and start over (dangerous)
  --update language-code
                     update the database, to latest mediawiki schema, for given language, and exit

Files:

  /etc/cron.d/wp-mirror  weekly cron job

  /etc/logrotate.d/wp-mirror  logrotate configuration

  /etc/mediawiki/LocalSettings_account.php  database credentials for building and using a wikipedia farm

  /etc/mediawiki/LocalSettings_wpmirror.php  configuration file for building and using a wikipedia farm

  /etc/wp-mirror/default.conf  default configuration file which lists all configurable parameters (do not edit this file)

  /etc/wp-mirror/local.conf  local configuration file (edit this file instead)

  /usr/share/doc/wp-mirror/doc/wp-mirror-x.y.pdf.gz  WP-MIRROR Reference Manual

  /usr/share/mediawiki/maintenance/database_farm.sql  template for wikimedia databases

  /usr/share/mediawiki/maintenance/importDump_farm.php  wikipedia farm wrapper for importDump.php

  /usr/share/mediawiki/maintenance/rebuildImages_farm.php  wikipedia farm wrapper for rebuildImages.php

  /usr/share/mediawiki/maintenance/update_farm.php  wikipedia farm wrapper for update.php

  /usr/share/pixmaps/wp-mirror.xpm  icon for menu system

  /var/lib/mediawiki/favicon.ico  favicon.ico file for browser

  /var/lib/mediawiki/wp-mirror.png  WP-MIRROR logotype for browser 

  /var/lib/mediawiki/images/  image files are stored here

  /var/lib/mediawiki/images/wp-mirror/  working files are placed here

  /var/lib/mysql/wikidb  empty wikimedia database (loaded from database_farm.sql)

  /var/lib/mysql/wpmirror  database for wp-mirror state information

  /var/lib/mysql/xxwiki  one database for each mirror in the wikipedia farm (where `xx' is replaced by the language code, such as `en' or `simple')

  /var/log/wp-mirror.log  log file

  /var/run/wp-mirror.pid  pid file

Examples: 

Installation is described in the WP-MIRROR Reference Manual, usually
to be found under `/usr/share/doc/wp-mirror/doc/'.  It should be
considered essential reading.

Wp-mirror uses two configuration files: `/etc/wp-mirror/default.conf'
and `/etc/wp-mirror/local.conf'.  The first shows all the configurable
parameters.  The second is where you should make your edits (if any).
By default, `wp-mirror' builds a mirror of the `simple' wiki.  The
`simple' wiki is the collection of articles written using Simple
English. That is, the articles are written using shorter sentences, as
would be appropriate for those for whom English is a Second Language
(ESL).  The `simple' wiki is also suitable for children.  If you wish
for a different wiki, or a set of wiki's, then edit
`/etc/wp-mirror/local.conf'.

After you edit /etc/wp-mirror/local.conf' you should run as root:

      # wp-mirror

You may run multiple instances of `wp-mirror' concurrently.  The first
instance runs in mirror mode, downloading and installing the articles
and images.  Each additional instance of `wp-mirror' runs in monitor
mode.  Instances running in monitor mode display the state of each
mirror.  If a suitable windowing system is present, progress bars are
rendered using graphics in a separate window, and otherwise using
ASCII characters in a console.  If X complains that you lack
permission, first try:

      $ xhost local:

You may control the mode of each instance of `wp-mirror' by providing
either the `--mirror' or the `--monitor' option at the command line.
For a desktop PC, it is reasonable to run two or three instances in
mirror mode together with an instance in monitor mode.  However, for a
laptop PC, it is best to run just one instance in mirror mode together
with one instance in monitor mode.

Wp-mirror normally runs in background as a weekly cron job, updating
the mirrors whenever the WikiMedia Foundation posts new dump files.

Wp-mirror configures a virtual-host for your web-server:

      <VirtualHost *:80>

      ServerAdmin webmaster@localhost

      ServerName wpmirror.site

      ServerAlias *.wpmirror.site

      DocumentRoot /var/lib/mediawiki

      CustomLog ${APACHE_LOG_DIR}/wpmirror.site-access.log combined

      ErrorLog ${APACHE_LOG_DIR}/wpmirror.site-error.log

      </VirtualHost>

Add the following line to `/etc/hosts':

      ::1 wpmirror.site simple.wpmirror.site

Then just open a web browser to <http://simple.wpmirror.site/> and
enjoy.

Report bugs in `wp-mirror' to <wp-mirror-list@nongnu.org>.")


;;;;--------------------------------------------------------------------------+
;;;; Any number of mirrors and monitors can communicate state via database    |
;;;;                                                                          |
;;;;     +--------+                      +---------+                          |
;;;;     | mirror |------          ------| monitor |                          |
;;;;     +--------+      \        /      +---------+                          |
;;;;     +--------+     +----------+                                          |
;;;;     | mirror |-----| wpmirror |                                          |
;;;;     +--------+     +----------+                                          |
;;;;     +--------+      /        \      +---------+                          |
;;;;     | mirror |------          ------| monitor |                          |
;;;;     +--------+                      +---------+                          |
;;;;                                                                          |
;;;;--------------------------------------------------------------------------+


(defconstant *wtf*
  "For help, please try:
    $ wp-mirror --help")


;;;;--------------------------------------------------------------------------+
;;;; Print utilities:                                                         |
;;;;   The idea is to give all the messages a common look and feel, and to    |
;;;;   have few format strings outside of this section.                       |
;;;;   Note:  `put-and-log-string' should be the only way to *standard-output*|
;;;;--------------------------------------------------------------------------+


(defun put-and-log-string (str &key (put nil))
  "Showing given string, and copying it to log file (if any)."
  (unless (null put)
    (format *standard-output* "~a~%" str))
  (unless (null *log-enabled-p*)
    (with-open-file (f *whereis-file-log*
		       :direction :output
		       :if-exists :append
		       :if-does-not-exist :create)
      (format f "[~d]~a~%" (os:process-id) str))))

(defun die ()
  "Exiting immediately."
  (ext:exit 0))

(defun abort-and-die (&key (put nil))
  "Showing abort message, and die."
  (put-and-log-string (format nil "~a." (_ "Aborting")) :put put)
  (die))

(defun debug-message (message &key (put t))
  "Showing given message when non-zero debug-level."
  (when (>= *debug-level* 1)
    (put-message (format nil "~a~a" (_ ";debug: ") message) :put nil)))

(defun debug-message-done (message &key (put t))
  "Showing given message followed by [DONE] when non-zero debug-level."
  (when (>= *debug-level* 1)
    (put-message-done (format nil "~a~a" (_ ";debug: ") message) :put nil)))

(defun debug-message-start (message &key (put t))
  "Showing given message followed by ellipses when non-zero debug-level."
  (when (>= *debug-level* 1)
    (put-message-start (format nil "~a~a" (_ ";debug: ") message) :put nil)))

(defun debug-message-value (message value &key (put nil))
  "Showing given message and value when non-zero debug-level."
  (when (>= *debug-level* 1)
    (put-message-value (format nil "~a~a" (_ ";debug: ") message) value
		       :put nil)))

(defun format-integer-for-human (n)
  "Formatting a number for human (e.g. 4096 --> 4.1K).  
Return four character string."
  (cond ((<= n 999)             (format nil "~3d " n))
	((<= n 9949)            (format nil "~3,1,-3fK" n))
	((<= n 999499)          (format nil "~3dK" (round (/ n 1000))))
	((<= n 9949999)         (format nil "~3,1,-6fM" n))
	((<= n 999499999)       (format nil "~3dM" (round (/ n 1000000))))
	((<= n 9949999999)      (format nil "~3,1,-9fG" n))
	((<= n 999499999999)    (format nil "~3dG" (round (/ n 1000000000))))
	((<= n 9949999999999)   (format nil "~3,1,-9fT" n))
	((<= n 999499999999999) 
	 (format nil "~3dT" (round (/ n 1000000000000))))
	((<= n 9949999999999999)   
	 (format nil "~3,1,-9fP" n))
	((<= n 999499999999999999)
	 (format nil "~3dP" (round (/ n 1000000000000000))))
	((<= n 9949999999999999999)   
	 (format nil "~3,1,-12fE" n))
	((<= n 999499999999999999999)
	 (format nil "~3dE" (round (/ n 1000000000000000000))))
	((<= n 9949999999999999999999)   
	 (format nil "~3,1,-15fZ" n))
	((<= n 999499999999999999999999)
	 (format nil "~3dZ" (round (/ n 1000000000000000000000))))
	((<= n 9949999999999999999999999)   
	 (format nil "~3,1,-18fY" n))
	((<= n 999499999999999999999999999)
	 (format nil "~3dY" (round (/ n 1000000000000000000000000))))
	(t                      (format nil "~d" n))))

(defun format-progress-bar-pulse ()
  "Formatting a progress bar (pulse type) as a string"
  (format-progress-bar-string 0.5 :pulse t))

(defun format-progress-bar-string (val &key (pulse nil))
  "Formatting a progress bar as a string"
  (let* ((len-string        35)
	 (s                 (make-string len-string :initial-element #\Space))
	 ;; if pulse, then use clock to generate pulse motion
	 (val-float         (cond (pulse (float (/ (get-decoded-time) 60)))
				  ((null val) 0)
				  ((numberp val) 
				   (max 0 val (min val 1)))  ; [0,1]
				  (t val)))
	 (val-percent       (* val-float 100))
	 (val-chars         (floor (* val-float len-string)))
	 (pos-equal-left    0)
	 (pos-arrow         (1- (+ pos-equal-left val-chars)))
	 (pos-equal-right   (1- pos-arrow)))
    (when (>= pos-arrow 0)
      (loop 
	for i of-type integer from pos-equal-left to pos-equal-right by 1
	do (setf (aref s i) #\=))
      (setf (aref s pos-arrow) #\>))
    (cond (pulse               (format nil "[~a] pulse" s))
          ((= val-percent   0) (format nil "[~a] wait " s))
          ((= val-percent 100) (format nil "[~a] done " s))
          (t                   (format nil "[~a] ~5,2f%" s val-percent)))))

(defun put-message (message &key (put nil))
  "Showing given message"
  (put-and-log-string (format nil "~a" message) :put put))

(defun put-message-and-abort (message &key (put t))
  "Showing given message, and abort."
  (put-message message :put put)
  (abort-and-die))

(defun put-message-and-die (message &key (put nil))
  "Showing given message, and die."
  (put-message message :put put)
  (die))

(defun put-message-done (message &key (put nil))
  "Showing given message followed by [DONE]."
  (put-message-result message :DONE :put put))

(defun put-message-fail (message &key (put t))
  "Showing given message followed by [FAIL]."
  (put-message-result message :FAIL :put put))

(defun put-message-fail-and-abort (message &key (put t))
  "Showing given message followed by [FAIL], and abort."
  (put-message-result message :FAIL :put put)
  (abort-and-die))

(defun put-message-fail-and-die (message &key (put t))
  "Showing given message followed by [FAIL], and die."
  (put-message-result message :FAIL :put put)
  (die))

(defun put-message-pass (message &key (put nil))
  "Showing given message followed by [PASS]."
  (put-message-result message :PASS :put put))

(defun put-message-result (message result &key (put nil))
  "Showing given message followed by a result"
  (put-and-log-string (format nil "~70a[~a]" message result) :put put))

(defun put-message-start (message &key (put nil))
  "Showing given message followed by ellipses."
  (put-and-log-string (format nil "~a ..." message) :put put))

(defun put-message-value (message value &key (put nil))
  "Showing given message and a value."
  (put-and-log-string (format nil "~34a: ~a" message value) :put put))

(defun put-message-value-done (message value &key (put nil))
  "Showing given message and value, followed by [DONE]."
  (put-message-value-result message value :DONE :put put))

(defun put-message-value-fail (message value &key (put t))
  "Showing given message and value, followed by [FAIL]."
  (put-message-value-result message value :FAIL :put put))

(defun put-message-value-fail-and-abort (message value &key (put t))
  "Showing given message and value, followed by [FAIL], and abort."
  (put-message-value-result message value :FAIL :put put)
  (abort-and-die))

(defun put-message-value-fail-and-die (message value &key (put t))
  "Showing given message and value, followed by [FAIL], and die."
  (put-message-value-result message value :FAIL :put put)
  (die))

(defun put-message-value-pass (message value &key (put nil))
  "Showing given message and value, followed by [PASS]."
  (put-message-value-result message value :PASS :put put))

(defun put-message-value-result (message value result &key (put nil))
  "Showing given message and value, followed by a result"
  (put-and-log-string (format nil "~34a: ~34a[~a]" message value result)
		      :put put))

(defun put-symbol-value (symbol &key (put nil))
  "Showing symbol and its value."
  (put-and-log-string (format nil "~34s: ~a" symbol (symbol-value symbol)) 
		      :put put))

(defun put-symbol-value-fail (symbol &key (put nil))
  "Showing symbol, its value, and [FAIL]."
  (put-symbol-value-result symbol :FAIL :put put))

(defun put-symbol-value-pass (symbol &key (put nil))
  "Showing symbol, its value, and [PASS]."
  (put-symbol-value-result symbol :PASS :put put))

(defun put-symbol-value-result (symbol result &key (put nil))
  "Showing symbol, its value, and a result."
  (put-and-log-string (format nil "~34s: ~34a[~a]" 
			      symbol (symbol-value symbol) result)
		      :put put))

(defun put-symbol-value-table (symbol-list &key (put nil))
  "Showing table of symbols and their values."
  (dolist (symbol symbol-list t)
    (put-symbol-value symbol :put put)))


;;;;--------------------------------------------------------------------------+
;;;; Parameters:                                                              |
;;;;   May be changed by: 1) config file `/etc/wp-mirror/local.conf', or      |
;;;;                      2) by command-line options.                         |
;;;;--------------------------------------------------------------------------+

;;; Was `--mirror' given as command line options?
(defparameter *cmd-force-mirror*        nil)
;;; Were any of `--gui', `--monitor', `--screen', or `--text' 
;;; given as command line options?
(defparameter *cmd-force-monitor*       nil)

;;;
;;; Design note:
;;;
;;; 0) WP-MIRROR when first installed has no database credentials.
;;;
;;; 1) WP-MIRROR, therefore, first scrapes `debian-sys-maint'
;;; credentials from `/etc/mysql/debian.cnf'.  WP-MIRROR can now
;;; access the DBMS with ALL PRIVILEGES.
;;; 
;;; 2) WP-MIRROR generates the passwords for `wikiadmin' and
;;; `wikiuser', and writes them to
;;; `/etc/mediawiki/LocalSettings_account.php'.
;;;
;;; 3) WP-MIRROR accesses the DBMS to create accounts and grant 
;;; privileges to `wikiadmin' and `wikiuser'.
;;;    a) `wikiadmin' is granted ALL PRIVILEGES with WP-MIRROR related
;;;       databases (e.g. `wikidb', `simplewiki', and `wpmirror').
;;;    b) `wikiuser' is granted SELECT, INSERT, UPDATE, and
;;;       DELETE privileges with WP-MIRROR related databases.
;;;
;;; 4) Thereafter, both WP-MIRROR and MediaWiki obtain `wikiadmin' and
;;; `wikiuser' credentials by scraping:
;;; `/etc/mediawiki/LocalSettings_account.php'.
;;;
(defparameter *db-debian-user*          nil)         ; expect `debian-sys-maint'
(defparameter *db-debian-password*      nil)         ;
(defparameter *db-root-user*            "root")      ; expect `root'
(defparameter *db-root-password*        nil)         ;
(defparameter *db-wikiadmin-user*       "wikiadmin") ; expect `wikiadmin'
(defparameter *db-wikiadmin-password*   nil)         ;
(defparameter *db-wikiuser-user*        "wikiuser")  ; expect `wikiuser'
(defparameter *db-wikiuser-password*    nil)         ; 
(defparameter *db-name*                 "wikidb")     
(defparameter *db-server*               "localhost")  
(defparameter *db-type*                 "mysql")      
(defparameter *db-wpmirror-name*        "wpmirror")   
(defparameter *db-wpmirror-tables*      2)

(defparameter *debug-level*             0)
(defparameter *log-enabled-p*           nil)

;;; Main modes (:add :delete :drop :mirror :monitor :restore-default :update)
(defparameter *main-language*           nil)
(defparameter *main-mode*               nil)

;; MediaWiki configuration
(defparameter *mediawiki-wgUseImageMagick*          "false") ; "true"
(defparameter *mediawiki-wgCustomConvertCommand*
  "/usr/bin/gm convert %s -resize %wx%h %d")
(defparameter *mediawiki-wgSVGConverter*         "inkscape") ; "rsvg"

;;; Concurrency of mirroring
(defparameter *mirrors*                 1)      ; not implemented
(defparameter *mirror-concurrency-limit-xchunk* 1)

;;; Do we download and validate image files?  
(defparameter *mirror-image-download-p*                   t)
(defparameter *mirror-image-use-wikix-p*                nil)
(defparameter *mirror-image-validate-p*                   t)
(defparameter *mirror-importdump-timeout-sec-max*       nil) ; [1000-4000]
(defparameter *mirror-objectcache-threshold*          10000)
(defparameter *mirror-objectcache-delete-limit*        1000)
(defparameter *mirror-rebuildimages-threshold*        20000)
(defparameter *mirror-rebuildimages-timeout-sec-inc*    200)
(defparameter *mirror-rebuildimages-timeout-sec-min*   2000)
(defparameter *mirror-rebuildimages-timeout-sec-max*  20000)

;;; Monitors (if any) should poll mirror status every so many seconds.
(defparameter *monitor-gui-font-preference-list* ; most preferred first
'("-*-helvetica-medium-r-*-*-12-*-*-*-*-*-*"    
  "-*-lucida-medium-r-*-*-12-*-*-*-*-*-*"
  "-*-lucida-medium-r-*-*-14-*-*-*-*-*-*"))
(defparameter *monitor-gui-one-window-per-language*    t) ;
(defparameter *monitor-mode*          :auto)    ; (:auto, :gui, :screen, :text)
(defparameter *monitor-poll-sec*         10)
(defparameter *monitor-poll-sec-min*      1)
(defparameter *monitor-poll-sec-max*    100)
(defparameter *monitor-screen-stream*   nil)    ; if choose :screen
(defparameter *monitor-screen-height*   nil)    ; if choose :screen
(defparameter *monitor-screen-width*    nil)    ; if choose :screen

;;; Constants
(defconstant *kibibyte*        (expt 1024 1) "KiB")
(defconstant *mebibyte*        (expt 1024 2) "MiB")
(defconstant *gibibyte*        (expt 1024 3) "GiB")
(defconstant *tebibyte*        (expt 1024 4) "TiB")
(defconstant *kibobyte*        (expt 1000 1) "K or KB")
(defconstant *megabyte*        (expt 1000 2) "M or MB")
(defconstant *gigabyte*        (expt 1000 3) "G or GB")
(defconstant *terabyte*        (expt 1000 4) "T or TB")

;;; System resources minimums
(defparameter *system-cpu*                 1)
(defparameter *system-cpu-min*             1)
(defparameter *system-cpu-max*             3)
(defparameter *system-disk-name-list*    nil)
;;; Assert adequate disk and memory capacity for largest wiki's
(defparameter *system-partition-free-images-min*       (*   5 *gigabyte*))
(defparameter *system-partition-free-images-start-min* (* 100 *gigabyte*))
(defparameter *system-partition-free-images-warn*      (*  60 *gigabyte*))
(defparameter *system-partition-free-innodb-min*       (*   5 *gigabyte*))
(defparameter *system-physical-memory-min*             (*   4 *gigabyte*))
(defparameter *wikimedia-large-wiki-list*
  '("en" "de" "fa" "nl" "it" "es" "pl" "ru" "ja" "pt"))

;;; Most of the heavy lifting is done by invoking existing utilities.
(defparameter *whereis-apache2ctl*          "/usr/sbin/apache2ctl")
(defparameter *whereis-bashrc*              "/etc/bash.bashrc")
(defparameter *whereis-bunzip2*             "/bin/bunzip2")
(defparameter *whereis-cat*                 "/bin/cat")
(defparameter *whereis-cp*                  "/bin/cp")
(defparameter *whereis-chown*               "/bin/chown")
(defparameter *whereis-chmod*               "/bin/chmod")
(defparameter *whereis-cron*                "/etc/cron.d/wp-mirror")
(defparameter *whereis-curl*                "/usr/bin/curl")
(defparameter *whereis-curlrc*              "/root/.curlrc")
;;; Need directories
(defparameter *whereis-directory-apache-sites-available*
  "/etc/apache2/sites-available/")
(defparameter *whereis-directory-apache-sites-enabled*
  "/etc/apache2/sites-enabled/")
(defparameter *whereis-directory-images*        "/var/lib/mediawiki/images/")
(defparameter *whereis-directory-images-bad* 
  "/var/lib/mediawiki/images/bad-images/")
(defparameter *whereis-directory-images-math* 
  "/var/lib/mediawiki/images/math/")
(defparameter *whereis-directory-images-thumb* 
  "/var/lib/mediawiki/images/thumb/")
(defparameter *whereis-directory-images-tmp* 
  "/var/lib/mediawiki/images/tmp/")
(defparameter *whereis-directory-mediawiki*               "/var/lib/mediawiki/")
(defparameter *whereis-directory-mediawiki-config*        "/etc/mediawiki/")
(defparameter *whereis-directory-mediawiki-extensions-math* 
  "/var/lib/mediawiki/extensions/Math/math/")
(defparameter *whereis-directory-mediawiki-maintenance*
  "/usr/share/mediawiki/maintenance/")
(defparameter *whereis-directory-mysql-config*            "/etc/mysql/")
(defparameter *whereis-directory-mysql-config-conf.d*     "/etc/mysql/conf.d/")
(defparameter *whereis-directory-mysql-datadir*           "/var/lib/mysql/")
(defparameter *whereis-directory-tmp*                     "/tmp/")
(defparameter *whereis-directory-working*
  "/var/lib/mediawiki/images/wp-mirror/")
(defparameter *whereis-directory-wpmirror-config*         "/etc/wp-mirror/")
(defparameter *whereis-env*                               "/usr/bin/env")
;;; Need files
(defparameter *whereis-file-dev-null*                     "/dev/null")
(defparameter *whereis-file-etc-hosts*                    "/etc/hosts")
(defparameter *whereis-file-log*                  "/var/log/wp-mirror.log")
(defparameter *whereis-file-mediawiki-config-localsettings* "LocalSettings.php")
(defparameter *whereis-file-mediawiki-config-localsettings-account*
  "LocalSettings_account.php")
(defparameter *whereis-file-mediawiki-config-localsettings-wpmirror*
  "LocalSettings_wpmirror.php")
(defparameter *whereis-file-mediawiki-importdump*         "importDump.php")
(defparameter *whereis-file-mediawiki-rebuildimages*      "rebuildImages.php")
(defparameter *whereis-file-mediawiki-update*             "update.php")
(defparameter *whereis-file-mediawiki-farm-database*      "database_farm.sql")
(defparameter *whereis-file-mediawiki-farm-importdump*
  "importDump_farm.php")
(defparameter *whereis-file-mediawiki-farm-rebuildimages* 
  "rebuildImages_farm.php")
(defparameter *whereis-file-mediawiki-farm-update*        "update_farm.php")
(defparameter *whereis-file-mediawiki-favicon*            "favicon.ico")
(defparameter *whereis-file-mediawiki-logo*               "wp-mirror.png")
(defparameter *whereis-file-mysql-config-debian*          "debian.cnf")
(defparameter *whereis-file-mysql-config-wpmirror*        "wp-mirror.cnf")
(defparameter *whereis-file-mysql-log-file*               "ib_logfile*")
(defparameter *whereis-file-pid*               "/var/run/wp-mirror.pid")
(defparameter *whereis-file-tmp-http*                     "/tmp/http")
(defparameter *whereis-file-virtual-host*           "wpmirror.site.conf")
(defparameter *whereis-file-wpmirror-config-default*      "default.conf")
(defparameter *whereis-file-wpmirror-config-local*        "local.conf")
(defparameter *whereis-gm*                  "/usr/bin/gm")
(defparameter *whereis-grep*                "/bin/grep")
(defparameter *whereis-gunzip*              "/bin/gunzip")
(defparameter *whereis-hdparm*              "/sbin/hdparm")
(defparameter *whereis-inkscape*            "/usr/bin/inkscape")
(defparameter *whereis-invoke-rc.d*         "/usr/sbin/invoke-rc.d")
(defparameter *whereis-ls*                  "/bin/ls")
(defparameter *whereis-md5sum*              "/usr/bin/md5sum")
(defparameter *whereis-mv*                  "/bin/mv")
(defparameter *whereis-mysql*               "/usr/bin/mysql")
(defparameter *whereis-mysqldump*           "/usr/bin/mysqldump")
(defparameter *whereis-mysql-tzinfo-to-sql* "/usr/bin/mysql_tzinfo_to_sql")
(defparameter *whereis-openssl*             "/usr/bin/openssl")
(defparameter *whereis-php*                 "/usr/bin/php")
(defparameter *whereis-rm*                  "/bin/rm")
(defparameter *whereis-rsvg*                "/usr/bin/rsvg")
(defparameter *whereis-texvc*               "/usr/bin/texvc")
(defparameter *whereis-wc*                  "/usr/bin/wc")
(defparameter *whereis-wget*                "/usr/bin/wget")
(defparameter *whereis-wgetrc*              "/etc/wgetrc")
(defparameter *whereis-wikix*               "/usr/bin/wikix")
(defparameter *whereis-x*                   "/usr/bin/X")
(defparameter *whereis-zoneinfo*            "/usr/share/zoneinfo")

;;; paths to files that can be restored
(defparameter *wpmirror-directory-restore*  "/usr/share/doc/wp-mirror/restore/")
(defparameter *wpmirror-config-delete-list*       
 '((*whereis-directory-mediawiki-config*
    *whereis-file-mediawiki-config-localsettings*)
   (*whereis-directory-mediawiki-config*
    *whereis-file-mediawiki-config-localsettings-account*)
   (*whereis-directory-mediawiki-config*
    *whereis-file-mediawiki-config-localsettings-wpmirror*)
   (*whereis-directory-mysql-config-conf.d* 
    *whereis-file-mysql-config-wpmirror*)
   (*whereis-directory-wpmirror-config*  *whereis-file-wpmirror-config-default*)
   (*whereis-directory-wpmirror-config*  *whereis-file-wpmirror-config-local*)
   (*whereis-directory-mediawiki-maintenance*
    *whereis-file-mediawiki-farm-database*)
   (*whereis-directory-mediawiki*        *whereis-file-mediawiki-favicon*)
   (*whereis-directory-mediawiki*        *whereis-file-mediawiki-logo*)
   ))
(defparameter *wpmirror-config-restore-list*       
 '((*whereis-directory-mediawiki-config*
    *whereis-file-mediawiki-config-localsettings*)
   (*whereis-directory-mediawiki-config*
    *whereis-file-mediawiki-config-localsettings-wpmirror*)
   (*whereis-directory-mysql-config-conf.d* 
    *whereis-file-mysql-config-wpmirror*)
   (*whereis-directory-wpmirror-config*  *whereis-file-wpmirror-config-default*)
   (*whereis-directory-wpmirror-config*  *whereis-file-wpmirror-config-local*)
   (*whereis-directory-mediawiki-maintenance*
    *whereis-file-mediawiki-farm-database*)
   (*whereis-directory-mediawiki*        *whereis-file-mediawiki-favicon*)
   (*whereis-directory-mediawiki*        *whereis-file-mediawiki-logo*)
   ))

;;; templates and URLs for wikimedia and wikix
(defparameter *whereis-file-checksums-template*
  "xxwiki-latest-md5sums.txt")
(defparameter *whereis-file-dump-template*
  "xxwiki-yyyymmdd-pages-articles.xml.bz2")
(defparameter *wikimedia-checksums-template*
  "xxwiki/latest/xxwiki-latest-md5sums.txt")
(defparameter *wikimedia-dump-template*
  "xxwiki/yyyymmdd/xxwiki-yyyymmdd-pages-articles.xml.bz2")
(defparameter *wikimedia-site*       "http://dumps.wikimedia.org/")
(defparameter *wikimedia-wikix*      "http://wikix.ngen-cast.de/wikix.tar.gz")

;;; Used for splitting dump into small XML files `xchunks'
(defparameter *xchunk-page-count*               1000)
(defparameter *ichunk-curl-retry*                  0)   ; [0,1] are reasonable

;;; This is the key parameter
(defparameter *mirror-languages*                '("simple"))
(defparameter *virtual-host-name*               "wpmirror.site")


;;;;--------------------------------------------------------------------------+
;;;; Configuration File:                                                      |
;;;;   Read config file, apply sanity checks, redefine parameters             |
;;;;--------------------------------------------------------------------------+


(defun put-parameters (&key (put nil))
  "Showing parameter values"
  (put-symbol-value-table
   '(*debug-level* 
		   *mirrors*
		   *mirror-concurrency-limit-xchunk*
		   *mirror-image-download-p* *mirror-image-validate-p*
                   *mirror-importdump-timeout-sec-max*
		   *mirror-languages* *monitor-mode* *monitor-poll-sec*
		   *monitor-poll-sec-min* *monitor-poll-sec-max*
		   *mirror-rebuildimages-threshold*
		   *mirror-rebuildimages-timeout-sec-min*
		   *mirror-rebuildimages-timeout-sec-inc*
		   *mirror-rebuildimages-timeout-sec-max*
		   *system-cpu-min* *system-cpu-max* 
		   *system-partition-free-images-min*
		   *system-partition-free-innodb-min*
		   *system-partition-free-images-start-min* 
		   *system-physical-memory-min*
		   *wikimedia-large-wiki-list*
		   *whereis-bashrc* *whereis-bunzip2*
		   *whereis-cat* *whereis-cp* *whereis-chown* *whereis-chmod*
		   *whereis-cron* *whereis-curl*  *whereis-curlrc* 
		   *whereis-directory-apache-sites-available*
		   *whereis-directory-apache-sites-enabled*
		   *whereis-directory-images* *whereis-directory-images-bad*
		   *whereis-directory-mediawiki-maintenance*
		   *whereis-directory-mysql-config*        
		   *whereis-directory-mysql-config-conf.d* 
		   *whereis-directory-tmp*
		   *whereis-directory-working*
		   *whereis-directory-wpmirror-config*
		   *whereis-file-checksums-template*
		   *whereis-file-dump-template*
		   *whereis-file-log*
		   *whereis-file-mediawiki-config-localsettings*
		   *whereis-file-mediawiki-config-localsettings-account*
		   *whereis-file-mediawiki-config-localsettings-wpmirror*
		   *whereis-file-mediawiki-farm-database*
		   *whereis-file-mediawiki-farm-importdump*
		   *whereis-file-mediawiki-farm-rebuildimages*
		   *whereis-file-mediawiki-farm-update*
		   *whereis-file-mediawiki-importdump*
		   *whereis-file-mediawiki-rebuildimages*
		   *whereis-file-mediawiki-update*
		   *whereis-file-mysql-config-debian*
		   *whereis-file-mysql-config-wpmirror* 
		   *whereis-file-pid*
		   *whereis-file-tmp-http*
		   *whereis-file-virtual-host*
		   *whereis-file-wpmirror-config-default*
		   *whereis-file-wpmirror-config-local* 
		   *whereis-gm* *whereis-grep* *whereis-gunzip*
		   *whereis-hdparm* *whereis-invoke-rc.d*
		   *whereis-ls*
		   *whereis-md5sum*
		   *whereis-mv* *whereis-mysql* *whereis-mysqldump*
		   *whereis-mysql-tzinfo-to-sql*
		   *whereis-openssl* *whereis-php*
		   *whereis-rm* *whereis-rsvg* *whereis-texvc*
		   *whereis-wc* *whereis-wget* *whereis-wgetrc*
		   *whereis-wikix* *whereis-x* *whereis-zoneinfo*
		   *wikimedia-checksums-template*
		   *wikimedia-site* *wikimedia-wikix*
		   *wpmirror-directory-restore*
		   *wpmirror-config-delete-list*
		   *wpmirror-config-restore-list*
		   *xchunk-page-count*)
   :put put)
  t)

(defun process-configuration-files-or-die (&key (put nil))
  "Loading configuration file"
  ;; 0) load config files
  (let ((path-default (merge-pathnames
	       (parse-namestring *whereis-directory-wpmirror-config*)
	       (parse-namestring *whereis-file-wpmirror-config-default*)))
	(path-local   (merge-pathnames
	       (parse-namestring *whereis-directory-wpmirror-config*)
	       (parse-namestring *whereis-file-wpmirror-config-local*))))
    ;; default config file
    (put-message-value (_ "probing") path-default :put put)
    (unless (file-exists-p path-default :put put)
      (put-message-value-fail-and-abort (_ "file not found") path-default))
    ;;(load path-default :if-does-not-exist nil)
    ;;(put-message-done (_ "default config file loaded"))
    (put-message-done (_ "default config file found, not loaded") :put put)
    ;; local config file
    (put-message-value (_ "probing") path-local :put put)
    (unless (file-exists-p path-local :put put)
      (put-message-value-fail-and-abort (_ "file not found") path-local))
    (load path-local :if-does-not-exist nil)
    (put-message-done (_ "local config file loaded") :put put))

  ;; 1) keep the monitor polling rate within reasonable bounds
  (put-message-start (_ "checking range *monitor-poll-sec*") :put put)
  (put-message-value (_ "reasonable bounds would be")
		     (format nil "[~d,~d]" *monitor-poll-sec-min* 
			     *monitor-poll-sec-max*) :put put)
  (cond ((< *monitor-poll-sec* *monitor-poll-sec-min*)
	 (setq *monitor-poll-sec* *monitor-poll-sec-min*))
	((> *monitor-poll-sec* *monitor-poll-sec-max*)
	 (setq *monitor-poll-sec* *monitor-poll-sec-max*))
	(t t))
  (put-message-value-done (_ "checking range *monitor-poll-sec*")
			  *monitor-poll-sec* :put put)
  ;; 2) return
  (let ((path-default (merge-pathnames
	       (parse-namestring *whereis-directory-wpmirror-config*)
	       (parse-namestring *whereis-file-wpmirror-config-default*))))
    (file-exists-p path-default :put put)))


;;;;--------------------------------------------------------------------------+
;;;; Command Line:                                                            |
;;;;   Read command line options, apply sanity checks, redefine parameters    |
;;;;--------------------------------------------------------------------------+


(defun put-copyright-license-and-disclaimer-and-die (&key (put t))
  "Showing copyright, license, and disclaimer, and die."
  (format *standard-output* "~a~%" *copyright-license-and-disclaimer*)
  (die))

(defun put-help-and-die ()
  "Showing help message, and die."
  (format *standard-output* "~a~%" *help*)
  (die))

(defun put-version-and-die (&key (put t))
  "Showing version, copyright, and author, and die."
  (format *standard-output* "~a~%~%" *version*)
  (format *standard-output* "~a~%~%" *copyright-license-and-disclaimer*)
  (format *standard-output* "~a~%"   *author*)
  (die))

(defun put-wtf-and-die ()
  "Telling user how to get help, and die."
  (format *standard-output* "~a~%" *wtf*)
  (die))

(defun process-command-line-arguments-or-die (&key (put nil))
  "Processing any command line arguments using getopt."
  (multiple-value-bind (filtered opt-args unexpected)
      ;; 0) parse command-line
      ;;    o FILE (if given) should be in `filtered',
      ;;    o options (and their arguments) should be in `*opt-args*', and
      ;;    o anything invalid should be in `unexpected'
      (getopt:getopt ext:*args*
		     '(("add"                    :optional nil)
		       ("copyright"              :none     nil)
		       ("debug"                  :none     nil)
		       ("delete"                 :optional nil)
		       ("drop"                   :optional nil)
		       ("dump"                   :optional nil)
		       ("gui"                    :none     nil)
		       ("help"                   :none     nil)
		       ("info"                   :none     nil)
		       ("mirror"                 :none     nil)
		       ("monitor"                :none     nil)
		       ("restore-default"        :none     nil)
		       ("screen"                 :none     nil)
		       ("text"                   :none     nil)
		       ("update"                 :optional nil)
		       ("version"                :none     nil)
		       ))
    (when (assoc "debug"            opt-args :test #'string=) ; --debug
      (defparameter *debug-level* (1+ *debug-level*)))
    (debug-message-start (_ "Processing command line arguments"))
    (debug-message-start (_ "List packages and globals "))
    (debug-message-value "custom:*ansi*" custom:*ansi*)
    (debug-message-value "*package*"    *package*)
    (debug-message-value "*print-case*" *print-case*)
    (debug-message-start (_ "Parse command line"))
    (debug-message-value (_ "ext:*args*") ext:*args*)
    (debug-message-value (_ "filtered")   filtered)
    (debug-message-value (_ "opt-args")   opt-args)
    (debug-message-value (_ "unexpected") unexpected)
    (debug-message-done  (_ "Parse command line"))
    ;; 1) detect user who needs help and die
    (when (assoc "copyright"        opt-args :test #'string=) ; --copyright
      (put-copyright-license-and-disclaimer-and-die))
    (when (assoc "add"              opt-args :test #'string=) ; --add
      (put-message (assoc "add" opt-args :test #'string=) :put put)
      (let ((language-code (cdr (assoc "add" opt-args :test #'string=))))
	(setq *main-language*       language-code)
	(setq *mirror-languages*    (list language-code))
	(setq *main-mode*           :add)))
    (when (assoc "delete"           opt-args :test #'string=) ; --delete
      (put-message (assoc "delete" opt-args :test #'string=) :put put)
      (let ((language-code (cdr (assoc "delete" opt-args :test #'string=))))
	(setq *main-language*       language-code)
	(setq *main-mode*           :delete)))
    (when (assoc "drop"             opt-args :test #'string=) ; --drop
      (put-message (assoc "drop" opt-args :test #'string=) :put put)
      (let ((language-code (cdr (assoc "drop" opt-args :test #'string=))))
	(setq *main-language*       language-code)
	(setq *main-mode*           :drop)))
    (when (assoc "dump"             opt-args :test #'string=) ; --dump
      (put-message (assoc "dump" opt-args :test #'string=) :put put)
      (let ((language-code (cdr (assoc "dump" opt-args :test #'string=))))
	(setq *main-language*       language-code)
	(setq *main-mode*           :dump)))
    (when (assoc "restore-default"  opt-args :test #'string=) ;--restore-default
      (put-message (assoc "restore-default" opt-args :test #'string=) :put put)
      (setq *main-mode*             :restore-default))
    (when (assoc "help"             opt-args :test #'string=) ; --help
      (put-help-and-die))
    (when (assoc "info"             opt-args :test #'string=) ; --info
      (put-parameters :put t)
      (put-copyright-license-and-disclaimer-and-die))
    (when (assoc "update"           opt-args :test #'string=) ; --update
      (put-message (assoc "update" opt-args :test #'string=) :put put)
      (let ((language-code (cdr (assoc "update" opt-args :test #'string=))))
	(setq *main-language*       language-code)
	(setq *main-mode*           :update)))
    (when (assoc "version"          opt-args :test #'string=) ; --version
      (put-version-and-die))
    (when unexpected
      (put-message-start (_ "Error: command line") :put t)
      (put-message-value-fail (_ "unexpected option") unexpected)
      (put-wtf-and-die))
    (when (> (length filtered) 1)                      ; more than one FILE
      (put-message-start (_ "Error: command line") :put t)
      (put-message-value-fail (_ "saw more than one file") filtered)
      (put-wtf-and-die))
    (when (assoc "monitor"          opt-args :test #'string=) ; --monitor
      (setq *monitor-mode*          :auto)
      (setq *cmd-force-monitor*     t))
    (when (assoc "gui"              opt-args :test #'string=) ; --gui
      (setq *monitor-mode*          :gui)
      (setq *cmd-force-monitor*     t))
    (when (assoc "screen"           opt-args :test #'string=) ; --screen
      (setq *monitor-mode*          :screen)
      (setq *cmd-force-monitor*     t))
    (when (assoc "text"             opt-args :test #'string=) ; --text 
      (setq *monitor-mode*          :text)
      (setq *cmd-force-monitor*     t))
    (when (assoc "mirror"           opt-args :test #'string=) ; --mirror
      (setq *cmd-force-mirror*      t)
      (setq *cmd-force-monitor*     nil))                     ; override
    (when (assoc "debug"            opt-args :test #'string=) ; --debug
      (setq *monitor-mode*          :text))                   ; forces --text
    ;; 2) set configuration file (if given on command-line)
    (when filtered
      (defparameter *whereis-file-wpmirror-config-local* (first filtered))
      (assert-file-exists-p *whereis-file-wpmirror-config-local* :put put))
  (put-message-done (_ "processing command line arguments") :put put)
  t))


;;;;--------------------------------------------------------------------------+
;;;; Shell utilities:                                                         |
;;;;   Some tasks are more easily done with existing utilities.               |
;;;;   Avoid reinventing the wheel (absent compelling reasons).               |
;;;;--------------------------------------------------------------------------+


(defun shell-apache2ctl-grep-vhost-p (&key (put nil))
  "Determining if Apache2 was loaded with wpmirror virtual host"
  (let* ((command-string (format nil "~a -S 2>&1 | ~a ~s"
				 *whereis-apache2ctl*
				 *whereis-grep*
				 *virtual-host-name*)))
    (shell-command command-string :output :read-lines :put put)))

(defun shell-bunzip2 (file-name &key (put nil))
  "Decompressing `bzip2' file"
  (let* ((command-string (format nil "~a --force --keep ~a"
				 *whereis-bunzip2* file-name)))
    (shell-command command-string :put put)))

(defun shell-cat-grep (file-name pattern &key (put nil))
  "Grepping a file for a line matching pattern"
  (let* ((command-string (format nil "~a ~s | ~a ~s" 
				 *whereis-cat*
				 file-name
				 *whereis-grep*
				 pattern)))
    (shell-command command-string :output :read-lines :put nil)))

(defun shell-chmod (file-name &key (permissions "744") (put nil))
  "Chmod file"
  (let ((command-string (format nil "~a ~a ~a"
				*whereis-chmod*	permissions file-name)))
    (shell-command command-string :put nil)))

(defun shell-chown (file-name &key (owner "www-data:www-data") (put nil))
  "Chown file"
  (let ((command-string (format nil "~a ~a ~a"
				*whereis-chown* owner file-name)))
    (shell-command command-string :put nil)))

(defun shell-chown-dir (dir-name &key (owner "www-data:www-data") (put nil))
  "Chown directory recursively"
  (let ((command-string (format nil "~a --recursive ~a ~a"
				*whereis-chown*	owner dir-name)))
    (shell-command command-string :put nil)))

(defun shell-command (command-string &key (input :terminal)
				     (output :terminal)
				     (if-output-exists :overwrite) (put nil))
  "Run shell command, put messages before and after"
  ;;
  ;; Design note:
  ;;
  ;; Purpose: wrap `ext:run-shell-command' with messages; and to read
  ;; and return any output.
  ;;
  ;; command-string - is a valid shell commmand
  ;;
  ;; `:input' - where program's input is to come from: either
  ;; `:terminal' (`stdin', the default), or `:stream' (a Lisp `stream'
  ;; to be created), or a pathname designator (an input file), or
  ;; `nil' (no input at all).  Do not use `nil' as it messes up
  ;; certain command-strings (e.g. `env printf %s some-string |
  ;; openssl dgst -md5').
  ;;
  ;; `:output' - where the program's output is to be sent to: either
  ;; `:terminal' (`stdout', the default), or `:read' (a Lisp `stream'
  ;; to be created and read with `read'), or `:read-line' (a Lisp
  ;; `stream' to be created, read once with `read-line', returning a
  ;; string), or `:read-lines' (a Lisp `stream' to be created, read
  ;; until `EOF', returning list of strings), or a pathname designator
  ;; (an output file), or `nil' (no input at all).
  ;;
  ;; `:if-output-exists' - what to do if the `:output' file already
  ;; exists.  The possible values are `:overwrite' (the default),
  ;; `:append', and `:error', with the same meaning as for `open'.
  ;;
  ;; `:put' - where to write messages: either `nil' (write to log file
  ;; only), or `t' (write to both `stdout' and to log file).
  ;;
  ;;
  ;; Design note:
  ;;
  ;; This version fails due to LOCALE mismatch (clisp uses UTF-8,
  ;; shell uses POSIX).  Example, if command-string contain unicode:
  ;;
  ;;  (setq message (first (regexp:regexp-split " " command-string)))
  ;;
  ;;  *** - MAKE-ARRAY: dimension -1 is not of type `(INTEGER 0
  ;;  (,ARRAY-DIMENSION-LIMIT))
  ;;
  ;; So we use the following version:
  ;;
  ;;  (setq message (subseq command-string 0
  ;;		            (regexp:match-start
  ;;			     (regexp:match " " command-string))))
  ;;
  (let* ((match-p        (regexp:match " " command-string))
	 (message        (if (null match-p)
			     command-string
			   (subseq command-string 0
				   (regexp:match-start match-p))))
	 (message-return (format nil "~a returned" message))
	 (output-stream  nil)
	 (result         nil))
    (put-message-start command-string :put put)
    (case output
	  (:read
	   (setq output-stream  (ext:run-shell-command command-string
						       :input input
						       :output :stream))
	   (setq result         (with-open-stream (s output-stream)
						  (read s nil 'eof))))
	  (:read-line
	   (setq output-stream  (ext:run-shell-command command-string
						       :input input
						       :output :stream))
	   (setq result         (with-open-stream (s output-stream)
						  (read-line s nil 'eof))))
	  (:read-lines
	   (setq output-stream  (ext:run-shell-command command-string
						       :input input
						       :output :stream))
	   (setq result         (with-open-stream (s output-stream)
					     (do ((l (read-line s nil 'eof)
						     (read-line s nil 'eof)))
						 ((eq l 'eof) (nreverse result))
						 (push l result)))))
	  (otherwise   ; `:terminal', a pathname, or `nil'
	   (setq result         (ext:run-shell-command command-string
						       :input input
						       :output output
					:if-output-exists if-output-exists))))
    (when (streamp output-stream)
      (close output-stream))
    (case result
	  ('eof (put-message-value      message-return result :put put))
	  (nil  (put-message-value      message-return result :put put))
	  (t    (put-message-value      message-return result :put put)))
    result))

(defun shell-command-with-timeout (command-simple &key (timeout nil) (put nil))
  "Execute a command with a timeout"
  ;;
  ;; Design note:
  ;;
  ;; Problem.  Some processes can hang the system
  ;; (`rebuildImages.php') or loop forever (`importDump.php').
  ;;
  ;; Solution.  We impose a timeout on the process by killing it, if
  ;; it is still running, after a period of time (say 1ks).
  ;;
  ;; We spawn a sub-process and get its PID: 
  ;; 
  ;;     (php rebuildImages_farm.php lang --missing) & pid=$!
  ;;
  ;; and spawn a sub-process to timeout the PID:
  ;;
  ;;     (sleep 1000 && kill -15 $pid)
  ;;
  ;; and have the main process `wp-mirror' wait for it:
  ;;
  ;;     wait $pid
  ;;
  ;; Signals.  SIGTERM (15) was tested with success, and is preferred
  ;; to the heavy-handed SIGKILL (9).
  ;;
  (let* ((command-timeout (format nil
	 "(~a) & pid=$! && (sleep ~d && kill -15 $pid) & wait $pid" 
		                 command-simple
				 timeout))
	 (command-string   (if (null timeout)
			       command-simple
			     command-timeout)))
    (shell-command command-string :put put)))

(defun shell-copy-file (file-name dir-name &key (put nil))
  "Copying file to destination directory"
  (let* ((command-string (format nil "~a -a ~a ~a"
				 *whereis-cp* file-name dir-name))
	 (result         (shell-command command-string :put nil)))
    (when result (abort-and-die))))

(defun shell-count-dir (dir-name &key (put nil))
  "Counting files in directory"
  (let ((command-string (format nil "~a -U ~a | ~a -l"
				*whereis-ls* dir-name *whereis-wc*)))
    (shell-command command-string :output :read :put put)))

(defun shell-count-images (&key (put nil))
  "Counting images"
  (let ((command-string (format nil "~a -U ~a[0-9a-f]/* | ~a -l"
				*whereis-ls*
				*whereis-directory-images*
				*whereis-wc*)))
    (shell-command command-string :output :read :put put)))

(defun shell-curl-http-head (url file-name &key (silent t) (put nil))
  "Downloading HTTP head from URL, result stored in `file-name.head'"
  (let* ((command-string (concatenate 'string
			      (format nil "~a --head --fail " *whereis-curl*)
			      (unless (null silent) "--silent ")
			      url)))
    (shell-command command-string :output (format nil "~a.head" file-name)
		   :if-output-exists :overwrite :put nil)))

(defun shell-curl-http-file (url file-name &key (silent t) (put nil))
  "Downloading file from URL, save as file-name"
  (let* ((command-string (concatenate 'string
			      (format nil "~a --retry 1 --fail " *whereis-curl*)
			      (unless (null silent) "--silent ")
			      "--max-time 5000 "
			      url))
	 (wip-file  (format nil "~a.wip" file-name))
	 (result    1)) ; assume abnormal process termination
    ;; 1) create a temporary file to use to detect partial downloads
    (with-open-file (f wip-file :direction :probe :if-does-not-exist :create))
    (unless (ext:probe-pathname wip-file)
      (put-message-value-fail-and-abort (_ "unable to create wip file") 
					wip-file))
    ;; 2) download file (nil = success)
    (setq result (shell-command command-string :output file-name
				:if-output-exists :overwrite :put put))
    ;; 3) if partial download, remove file fragment
    (when (and result (file-exists-p file-name :put put))
      (shell-remove-file file-name :put put))
    ;; clean-up
    (shell-remove-file wip-file  :put put)
    result))

(defun shell-count-image (file-name &key (put nil))
  "Counting images `[[File:' in file"
  (let* ((file-type         (sql-select-file-type file-name))
	 (command-ichunk    (format nil "~a ~a | ~a \"^if\" | wc -l"  
				    *whereis-cat*
				    file-name
				    *whereis-grep*))
	 (command-xml-file  (format nil "~a ~a | ~a \"\\[\\[File:\" | wc -l"
				    *whereis-cat*
				    file-name
				    *whereis-grep*))
	 (command-xml-image (format nil "~a ~a | ~a \"\\[\\[Image:\" | wc -l"
				    *whereis-cat*
				    file-name
				    *whereis-grep*))
	 (command-xml-media (format nil "~a ~a | ~a \"\\[\\[Media:\" | wc -l"
				    *whereis-cat*
				    file-name
				    *whereis-grep*)))
    (cond ((string= file-type "ichunk")
	   (shell-command command-ichunk :output :read :put put))
	  ((member file-type '("xml" "xchunk") :test #'string=)
	   (+ (shell-command command-xml-file  :output :read :put put)
	      (shell-command command-xml-image :output :read :put put)
	      (shell-command command-xml-media :output :read :put put)))
	  (t 0))))

(defun shell-find-1k (dir-name &key (put nil))
  "Finding files smaller than 1k in given directory"
  (let ((command-string (format nil "find ~a -size 1k" dir-name)))
    (shell-command command-string :output :read-lines :put put)))

(defun shell-gunzip (file-name &key (put nil))
  "Decompressing `gz' file"
  (let* ((command-string (format nil "~a --force ~a"
				 *whereis-gunzip* file-name)))
    (shell-command command-string :put put)))

(defun shell-ls-grep (dir-name pattern &key (put nil))
  "Grepping a directory listing for file names matching a pattern"
  (let ((command-string (format nil "~a -U ~a | ~a ~s" 
				*whereis-ls* dir-name
				*whereis-grep* pattern)))
    (shell-command command-string :output :read-lines :put put)))

(defun shell-count-page (file-name &key (put nil))
  "Counting pages `<page>' in file"
  (let* ((file-type      (sql-select-file-type file-name))
	 (command-string (cond
			  ((string= file-type "xchunk")
			   (format nil "~a ~a | ~a \"  <page>\" | wc -l"  
				   *whereis-cat*
				   file-name
				   *whereis-grep*))
			  ((string= file-type "xml")
			   (format nil "~a ~a | ~a \"  <page>\" | wc -l"  
				   *whereis-cat*
				   file-name
				   *whereis-grep*))
			  (t nil))))
    (shell-command command-string :output :read :put put)))

(defun shell-gm-identify (pathname &key (put nil))
  "Determining if image file is incomplete or corrupt"
  (let* ((command-string (format nil
				"~a identify -verbose ~s 2>&1 | ~a ~s"
				*whereis-gm*   (namestring pathname)
				*whereis-grep* "gm identify"))
	 (result         (shell-command command-string :output :read-line
					:put nil)))
    (when (eq result 'eof) (setq result nil))
    result))

(defun shell-hdparm-write-cache-disabled-p (block-device &key (put nil))
  "Determining if hard drive write cache has been disabled"
  (let* ((command-string (format nil "~a -W0 ~a" *whereis-hdparm* block-device))
	 ;; attempting to disable HDD write cache
	 (lines          (shell-command command-string :output :read-lines
					:put nil))
	 ;; ignore first line : (which is blank)
	 ;;    and second     : /dev/sda:
	 ;;    and third      :  setting drive write-caching =  0 (off)
	 ;; keep fourth       :  write-caching =  0 (off)
	 (line           (fourth lines)))
    (if (regexp:match "off" line) t nil))) ; t means disabled, nil means on

(defun shell-md5sum (file-name &key (put nil))
  "Computing file checksum"
  ;;
  ;; Design note:
  ;;
  ;; Implementations of openssl differ.  Have seen two output formats:
  ;; 1) 567f7e60f0369351c5c56a5076cad182
  ;; 2) MD5(Flag_of_Saint_Helena.svg)= 567f7e60f0369351c5c56a5076cad182
  ;;
  (let* ((command-string (format nil "~a dgst -md5 ~a"
				 *whereis-openssl* file-name))
	 (result-string  (shell-command command-string :output :read-line
					:put put)))
    (when (regexp:match "=" result-string)
      (setq result-string (second (regexp:regexp-split " " result-string))))
    result-string))

(defun virtual-server-name (language-code &key (put nil))
  "Determining virtual server name given language-code"
  (format nil "~a.~a" language-code *virtual-host-name*))

(defun shell-mediawiki-farm-importdump (file-name &key (timeout nil) (put nil))
  "Importing xchunk into `xxwiki' database"
  ;;
  ;; Design note:
  ;;
  ;; Problem.  `importDump.php' sometimes uses 100% of one CPU
  ;; indefinitely (giving the impression of an infinite loop).  This
  ;; can be seen by running `top' and looking for the command `php'
  ;; (which will probably be the first process listed).
  ;;
  ;; Solution.  We can impose a timeout on the `importDump_farm.php'
  ;; process by killing it, if it is still running, after a period of
  ;; time (say 1ks).  
  ;;
  (let* ((language-code  (sql-select-file-language file-name))
	 (code-path      (merge-pathnames
		 (parse-namestring *whereis-directory-mediawiki-maintenance*)
		 (parse-namestring *whereis-file-mediawiki-farm-importdump*)))
	 (command-string (format nil "~a ~a ~a ~a"
				 *whereis-php*
				 (namestring code-path)
				 file-name
				 (virtual-server-name language-code))))
    (shell-command-with-timeout command-string :timeout timeout :put put)))

(defun shell-mediawiki-farm-rebuildimages (file-name &key (timeout nil)
						     (put nil))
  "Rebuilding image table in `xxwiki' database"
  ;;
  ;; Design note:
  ;;
  ;; Problem.  `rebuildImages.php' seems to have a memory leak.  As it
  ;; processes images it occupies ever more SDRAM.  This can be seen
  ;; by running `top' and looking under the `RES' column.  For every
  ;; 20,000 images it occupies 1G of SDRAM, until a memory shortage
  ;; results in GNU/Linux randomly killing off processes, and the
  ;; system hangs.
  ;;
  ;; Solution.  We impose a timeout.
  ;;
  (let* ((language-code  (sql-select-file-language file-name))
	 (code-path      (merge-pathnames
	       (parse-namestring *whereis-directory-mediawiki-maintenance*)
	       (parse-namestring *whereis-file-mediawiki-farm-rebuildimages*)))
         (command-string (format nil
				 "~a ~a ~a --missing > ~a" 
				 *whereis-php*
				 (namestring code-path)
				 (virtual-server-name language-code)
				 *whereis-file-dev-null*)))
    (shell-command-with-timeout command-string :timeout timeout :put put)))

(defun shell-mediawiki-farm-update (language-code &key (timeout nil)
						  (put nil))
  "Update `xxwiki' database to current MediaWiki schema"
  (let* ((code-path (merge-pathnames
	       (parse-namestring *whereis-directory-mediawiki-maintenance*)
	       (parse-namestring *whereis-file-mediawiki-farm-update*)))
         (command-string (format nil
				 "~a ~a ~a" 
				 *whereis-php*
				 (namestring code-path)
				 (virtual-server-name language-code)
				 *whereis-file-dev-null*)))
    (shell-command-with-timeout command-string :timeout timeout :put put)))

(defun shell-mysql-and-return-stream (arguments &key (put nil))
  "Querying MySQL database"
  (let ((s (ext:run-program *whereis-mysql*
			    :arguments arguments
			    :output :stream)))
  (if (null s)
      (put-message-value-fail-and-abort (_ "mysql query") arguments)
    s)))

(defun shell-mysql-load-database (db-name file-path &key (put nil))
  "Loading SQL dump file into database"
  (let ((command-string (concatenate 'string
			    (format nil "~a " *whereis-mysql*)
			    (format nil "--host=~a " *db-server*)
			    (format nil "--user=~a " *db-wikiadmin-user*)
			    (format nil "--password=~a " *db-wikiadmin-password*)
			    (format nil "--database=~a " db-name))))
    (shell-command command-string :input file-path :put put)))

(defun shell-mysql-restart (&key (put nil))
  "Restarting the `mysqld' server"
  (let ((path-name (merge-pathnames
		    (parse-namestring *whereis-directory-mysql-datadir*)
		    (parse-namestring *whereis-file-mysql-log-file*))))
    (shell-invoke-rc.d "mysql" "stop")
    (sleep 1)
    (shell-remove-file path-name             :put put)
    (shell-invoke-rc.d "mysql" "start")
    (sleep 5)))

(defun shell-mysql-tzinfo-to-sql-load-database (&key (put nil))
  "Loading time zone info into database"
  (let ((command-string (format nil "~a ~a 2>/dev/null | ~a ~a"
				*whereis-mysql-tzinfo-to-sql* 
				*whereis-zoneinfo*
				*whereis-mysql*
				(concatenate 'string 
		     (format nil "--host=~a "     *db-server*)
		     (format nil "--user=~a "     *db-debian-user*)
		     (format nil "--password=~a " *db-debian-password*)
		                 "--database=mysql"))))
    (shell-command command-string :put put)))

(defun shell-mysqldump-to-file (database-name file-path &key (put nil))
  "Dumping MySQL database to file"
  (let ((command-string (concatenate 'string
			   (format nil "~a " *whereis-mysqldump*)
			   (format nil "--host=~a "     *db-server*)
			   (format nil "--user=~a "     *db-wikiadmin-user*)
			   (format nil "--password=~a " *db-wikiadmin-password*)
			   database-name)))
    (shell-command command-string :output file-path
		   :if-output-exists :overwrite :put put)))

(defun shell-printf-md5sum (string &key (put nil))
  "Computing md5sum of a string"
  ;;
  ;; Design note:
  ;;
  ;; String may neither have spaces nor characters with special
  ;; meaning to `shell' (e.g. backquote, dollar, parentheses, etc.)
  ;; unless escaped with a backslash.
  ;;
  ;; 1) bad    Maria_Callas_(La_Traviata).jpg   
  ;; 2) good   Maria_Callas_\(La_Traviata\).jpg
  ;;
  (let* ((command-string (format nil "~a printf %s ~a | ~a dgst -md5"
				 *whereis-env*
				 string
				 *whereis-openssl*))
	 (result-string  (shell-command command-string :output :read-line
	                                :put put)))
    ;; Implementations of openssl differ.  Have seen two output formats:
    ;; 1) 00001111222233334444555566667777
    ;; 2) (stdin)= 00001111222233334444555566667777
    (when (regexp:match "=" result-string)
      (setq result-string (second (regexp:regexp-split " " result-string))))
    (debug-message-value (_ "md5sum trimmed") result-string)
    result-string))

(defun shell-invoke-rc.d (service action &key (put nil))
  "Invoking run-time service"
  (let ((command-string (format nil "~a ~a ~a" 
				*whereis-invoke-rc.d*
				service action)))
    (shell-command command-string :put nil)))

(defun shell-remove-directory (dir-name &key (put nil))
  "Removing file"
  (let ((command-string (format nil "~a --recursive --force ~a"
				*whereis-rm* dir-name)))
    (shell-command command-string :put nil)))

(defun shell-remove-file (file-name &key (put nil))
  "Removing file"
  (let ((command-string (format nil "~a --force ~a" *whereis-rm* file-name)))
    (shell-command command-string :put nil)))

(defun shell-remove-all-working-files (language-code &key (put t))
  "Removing all working files"
  (when (ext:probe-directory *whereis-directory-working*)
    ;; clean up .../images/wp-mirror/ directory
    (put-message-start (_ "cd to working directory")                :put put)
    (ext:cd *whereis-directory-working*)
    (put-message-value-done (_ "cd to working directory") (ext:cd)  :put put)
    (shell-remove-file (format nil "~a*" language-code)             :put put))
  (when (ext:probe-directory *whereis-directory-images*)
    ;; clean up .../images/ directory
    (put-message-start (_ "cd to images directory")                 :put put)
    (ext:cd *whereis-directory-images*)
    (put-message-value-done (_ "cd to images directory") (ext:cd)   :put put)
    (shell-remove-file "*log"                                       :put put))
  (when (ext:probe-directory *whereis-directory-working*)
    (put-message-start (_ "cd to working directory")                :put put)
    (ext:cd *whereis-directory-working*)
    (put-message-value-done (_ "cd to working directory") (ext:cd)  :put put))
  t)

(defun shell-remove-old-working-files (language &key (put t))
  "Removing decompressed dump file, xchunks, and ichunks given language"
  (when (ext:probe-directory *whereis-directory-working*)
    ;; clean up .../images/wp-mirror/ directory
    (put-message-start (_ "cd to working directory")                :put put)
    (ext:cd *whereis-directory-working*)
    (put-message-value-done (_ "cd to working directory") (ext:cd)  :put put)
    (shell-remove-file (format nil "~a*log" language)               :put put)
    (shell-remove-file (format nil "~a*sh" language)                :put put)
    (shell-remove-file (format nil "~a*xml" language)               :put put))
  (when (ext:probe-directory *whereis-directory-images*)
    ;; clean up .../images/ directory
    (put-message-start (_ "cd to images directory")                 :put put)
    (ext:cd *whereis-directory-images*)
    (put-message-value-done (_ "cd to images directory") (ext:cd)   :put put)
    (shell-remove-file "*log"                                       :put put))
  (when (ext:probe-directory *whereis-directory-working*)
    (put-message-start (_ "cd to working directory")                :put put)
    (ext:cd *whereis-directory-working*)
    (put-message-value-done (_ "cd to working directory") (ext:cd)  :put put))
  t)

(defun shell-rename-file (file-name-1 file-name-2 &key (put nil))
  "Renaming file"
  ;;
  ;; Design note:
  ;;
  ;; This routine will be removed in a later version.  This is because
  ;; it is used by `shell-wikix' only, which is deprecated.
  ;;
  (let ((command-string (format nil "~a ~a ~a"
				*whereis-mv* file-name-1 file-name-2)))
    (shell-command command-string :put t)
    (die)))

;  (let ((result (ext:run-program
;		 *whereis-mv*
;		 :arguments (list file-name-1 file-name-2))))
;    (if (null result)
;	(debug-message-value (_ "mv returned") result)
;      (put-message-value-fail-and-abort (_ "mv returned") result))
;    result))

(defun shell-script (file-name &key (put nil))
  "Running a shell script"
  (let ((command-string (format nil "./~a" file-name)))
    (shell-command command-string :put put)))

(defun shell-show-engine-innodb-status-grep (pattern &key (put nil))
  "Grepping `show engine innodb status' output"
  (let ((command-string (concatenate 'string
			  (format nil "~a " *whereis-mysql*)
			  (format nil "--host=~a "     *db-server*)
			  (format nil "--user=~a "     *db-wikiadmin-user*)
			  (format nil "--password=~a " *db-wikiadmin-password*)
			  "--execute=\"SHOW ENGINE INNODB STATUS\\G\" "
			  "| "
			  (format nil "~a ~s" *whereis-grep* pattern))))
    (shell-command command-string :output :read-line :put put)))

(defun shell-wikix (xchunk-name ichunk-name &key (put nil))
  "Generating `ichunk' from `xchunk'"
  ;;
  ;; Design note:
  ;;
  ;; This routine is used by `shell-wikix' only.
  ;;
  ;; As of WP-MIRROR 0.4, `shell-wikix' is deprecated because it
  ;; generates shell scripts that contain `bashisms' (i.e. not POSIX
  ;; compliant) which will not run under `/bin/dash' (the default
  ;; shell for scripts under Debian).
  ;;
  ;; In a later version, this code will be deleted.
  ;;
  (let ((command-string (format nil "~a" *whereis-wikix*)))
    (shell-command command-string :input xchunk-name :output ichunk-name
		   :if-output-exists :overwrite :put t)
    (die)))

;  (let ((result (ext:run-program *whereis-wikix*
;				 :input  xchunk-name
;				 :output ichunk-name
;				 :if-output-exists :overwrite)))
;    (if (null result)
;	(debug-message-value (_ "wikix returned") result)
;      (put-message-value-fail (_ "wikix returned") result))
;    result))


;;;;--------------------------------------------------------------------------+
;;;; Miscellaneous utilities                                                  |
;;;;--------------------------------------------------------------------------+


(defun sleep-while (time-out f &rest args)
  "Sleeping until earlier of time-out or function returns false"
  (loop for i from 0 to time-out
	do (put-message (_ "wait one second")) 
	   (sleep 1)
	always (apply f args)))                                ; <-- apply 

(defun sleep-until (time-out f &rest args)
  "Sleeping until earlier of time-out or function returns true"
  (loop for i from 0 to time-out
	do (put-message (_ "wait one second")) 
	   (sleep 1)
	thereis (apply f args)))                               ; <-- apply 


;;;;--------------------------------------------------------------------------+
;;;; Database utilities:                                                      |
;;;;   Debian packages cl-sql, cl-mysql, cl-uffi seem broken.                 |
;;;;   So 'query-database-*' are written using 'ext:run-program'.             |
;;;;--------------------------------------------------------------------------+


(defun grep-first (file-name pattern &key (put nil))
  "Grepping text file for first line matching pattern or nil"
  (first (shell-cat-grep file-name pattern :put put)))

(defun grep-first-non-comment (file-name pattern &key (put nil))
  "Grepping text file for first non-comment line matching pattern or nil"
  (with-open-file (f file-name
		     :direction :input
		     :if-does-not-exist nil)
    (when (streamp f)
      (do ((l (read-line f nil 'eof) (read-line f nil 'eof)))
	  ((eq l 'eof) nil)
	(unless (regexp:match "^[#;]" l)
	  (when (regexp:match pattern l)
	    (return l)))))))

(defun grep-last (file-name pattern &key (put nil))
  "Grepping text file for last line matching pattern or nil"
  (first (last (shell-cat-grep file-name pattern :put put))))

(defun grep-last-non-comment (file-name pattern &key (put nil))
  "Grepping text file for first non-comment line matching pattern or nil"
  (with-open-file (f file-name
		     :direction :input
		     :if-does-not-exist nil)
    (when (streamp f)
      (do ((l (read-line f nil 'eof) (read-line f nil 'eof))
	   (result nil))
	  ((eq l 'eof) result)
	(unless (regexp:match "^[#;]" l)
	  (when (regexp:match pattern l)
	    (setq result l)))))))

(defun query-database-and-return-stream (sql-string
					 &key 
					 (user     *db-wikiadmin-user*)
					 (database nil)
					 (put      nil))
  "Submitting SQL query, returning a stream containing the results"
  (debug-message-start (_ "query database return stream"))
  (debug-message sql-string)
  (let* ((password  (cond 
		     ((eql user *db-debian-user*)    *db-debian-password*)
		     ((eql user *db-root-user*)      *db-root-password*)
		     ((eql user *db-wikiadmin-user*) *db-wikiadmin-password*)
		     ((eql user *db-wikiuser-user*)  *db-wikiuser-password*)
		     (t nil)))
	 (arguments (list 
		     (concatenate 'string "--host="     *db-server*)
		     (concatenate 'string "--user="     user)
		     (concatenate 'string "--password=" password)
		     ;; --batch separates cells with Tab, rows with Newline
		     ;; we need Tabs for reliable parsing
		     "--batch"
		     "--skip-column-names"
		     ;; --raw means no escaping special characters
		     "--raw"
		     (concatenate 'string "--database=" database)
		     (concatenate 'string "--execute="  sql-string))))
    (debug-message (format nil "~a" (append (list "mysql") arguments)))
    (shell-mysql-and-return-stream arguments :put put)))

(defun query-database-and-return-list-of-strings (sql-string
				&key
				(user     *db-wikiadmin-user*)
				(database nil)
				(put      nil))
  "Submitting SQL query, returning a list of strings where 
each string is a row in the table"
  (let ((table-list nil))
    (with-open-stream (s (query-database-and-return-stream sql-string
							   :user     user
							   :database database
							   :put      put))
      (do ((l (read-line s nil 'eof) (read-line s nil 'eof)))
	  ((eq l 'eof) (nreverse table-list))
	(push l table-list)))))

(defun query-database-and-return-list-of-lists (sql-string
				&key
				(user     *db-wikiadmin-user*)
				(database nil)
				(put      nil))
  "Submitting SQL query, returning a list of lists of objects, 
where each object is an element in the table"  
  ;; this will misalign if there are NULL valued table elements not in
  ;; the last column (that is, NULL is not parsed)
  (let ((table-list nil)
	(result     nil))
    (with-open-stream (s (query-database-and-return-stream sql-string
							   :user     user
							   :database database
							   :put      put))
      (debug-message (_ "return stream") :put put)
      ;; --batch option separates cells with Tab, ends line with Newline
      (regexp:with-loop-split (row s (string #\Tab))
        ;; row is bound to a list of substrings split from one line of s
	(debug-message-value (_ "row") row :put put)
	(loop
	  with new-row = nil
	  for cell    in row
	  as  elem    = (if (or (regexp:match ":" cell) ; do not read packages
				(>= (length cell) 32))  ; md5sum is not number
			    cell
			  (with-input-from-string (s cell) (read s nil 'eof)))
	  do (setq new-row (cond ((numberp elem) (push elem new-row))
				 ((symbolp elem) (push cell new-row))
				 (t              (push cell new-row))))
	     (debug-message-value (_ "new-row") new-row :put put)
	  finally (setq table-list (push (nreverse new-row) table-list)))))
    (debug-message-value (_ "table-list") table-list :put put)
    (setq result (nreverse table-list))
    (debug-message-value (_ "return-list-of-lists") result :put put)
    result))

(defun sql-alter-table-file-0.5 (&key (put nil))
  "Altering table `wpmirror.file' for `wp-mirror-0.5' schema"
  (let ((message         (_ "Updating `wpmirror.file' to wp-mirror-0.5 schema"))
	(alter-table-0.5 (concatenate 'string
        "ALTER TABLE file "
	"MODIFY type ENUM('database','checksum','dump','xml','xchunk','ichunk','images','error') "
	" NOT NULL DEFAULT 'error';")))
    (put-message-start message :put put)
    (query-database-and-return-stream alter-table-0.5
				      :user     *db-wikiadmin-user*
				      :database *db-wpmirror-name*
				      :put      put)
    (put-message-done message :put put)))

(defun sql-alter-table-file (&key (put nil))
  "Altering table `wpmirror.file' to update its schema"
  ;; add more lines here if further changes to `wpmirror.file' schema
  (sql-alter-table-file-0.5 :put put)
  )

(defun sql-create-database (database-name &key (put nil))
  "Creating a database"
  (let ((create-database (format nil "CREATE DATABASE `~a`;" database-name)))
    (query-database-and-return-stream create-database
				      :user     *db-wikiadmin-user*
				      :database nil
				      :put      put)))

(defun sql-create-table-file (&key (put nil))
  "Creating table `file'"
  (let ((create-table    (concatenate 'string
        "CREATE TABLE IF NOT EXISTS file ("
	"timestamp TIMESTAMP DEFAULT CURRENT_TIMESTAMP "
	"ON UPDATE CURRENT_TIMESTAMP, "
	"language  VARCHAR(20) NOT NULL DEFAULT 'error', " ; e.g. `simple'
	"date      VARCHAR(8)  NOT NULL DEFAULT 'error', " ; e.g. `20111007'
	"name      VARCHAR(80) NOT NULL DEFAULT 'error', " 
	"size      BIGINT UNSIGNED NOT NULL DEFAULT 0, "   ; [0-18e18]
	"md5sum    VARCHAR(32) NOT NULL DEFAULT 'error', "
	"type      ENUM('database','checksum','dump','xml','xchunk','ichunk','images','error') "
	"NOT NULL DEFAULT 'error', "
	"state     ENUM('start','created','valid','pending','done','fail','error') "
        "NOT NULL DEFAULT 'error', "
	"page      INTEGER UNSIGNED NOT NULL DEFAULT 0, "
	"pages     INTEGER UNSIGNED NOT NULL DEFAULT 0, " ; pages
	"images    INTEGER UNSIGNED NOT NULL DEFAULT 0, " ; images
	"updates   INTEGER UNSIGNED NOT NULL DEFAULT 0, " ; 
	"semaphore INTEGER UNSIGNED NOT NULL DEFAULT 1, " ; [0-1]
	"PRIMARY KEY (name)) "
        "ENGINE=InnoDB;")))                               ; ACID compliant
    (query-database-and-return-stream create-table 
				      :user     *db-wikiadmin-user*
				      :database *db-wpmirror-name*
				      :put      put)))

(defun sql-create-table-image (&key (put nil))
  "Creating table `image'"
  (let ((create-table    (concatenate 'string
        "CREATE TABLE IF NOT EXISTS image ("
	"timestamp  TIMESTAMP DEFAULT CURRENT_TIMESTAMP "
	"ON UPDATE  CURRENT_TIMESTAMP, "
	"img_md5sum VARCHAR(32) NOT NULL DEFAULT 'error', "
	"img_state  ENUM('created','valid','error') "
        "NOT NULL DEFAULT 'error', "
	"PRIMARY KEY (img_md5sum)) "
        "ENGINE=InnoDB;")))                               ; ACID compliant
    (query-database-and-return-stream create-table
				      :user     *db-wikiadmin-user*
				      :database *db-wpmirror-name*
				      :put      put)))

(defun sql-create-user-debian    (password &key (put nil))
  "Creating DBMS account for `debian-sys-maint', no privileges granted"
  (let ((create-user (concatenate 'string
				  "CREATE USER 'debian-sys-maint'@'localhost' "
				  (format nil "IDENTIFIED BY '~a';" password))))
    (query-database-and-return-stream create-user
				      :user     *db-root-user*
				      :database *db-type*
				      :put      put)))


(defun sql-create-user (user password &key (put nil))
  "Creating DBMS accounts with no privileges granted"
  (let ((create-user (concatenate 'string
			  (format nil "CREATE USER '~a'@'localhost' " user)
			  (format nil "IDENTIFIED BY '~a';" password))))
    (query-database-and-return-stream create-user
				      :user     *db-debian-user*
				      :database *db-type*
				      :put      put)))

(defun sql-drop-user (user &key (put nil))
  "Dropping DBMS user account"
  (let ((drop-user (format nil "DROP USER '~a'@'localhost';" user)))
    (query-database-and-return-stream drop-user
				      :user     *db-debian-user*
				      :database *db-type*
				      :put      put)))

(defun sql-delete-language (language &key (put nil))
  "Deleting all rows of given language from `wpmirror.file' table"
  (let ((delete-from (format nil "DELETE FROM `~a`.file WHERE language='~a';" 
			     *db-wpmirror-name* language)))
    (query-database-and-return-list-of-strings delete-from
					       :user     *db-debian-user*
					       :put      put)))

(defun sql-delete-language-old (language &key (put nil))
  "Deleting all rows of given language from `wpmirror.file' table"
  (let ((delete-from (format nil (concatenate 'string
            (format nil "DELETE FROM `~a`.file WHERE language='~a' " 
		    *db-wpmirror-name* language)
	    "AND type NOT IN ('database', 'checksum');"))))
    (query-database-and-return-list-of-strings delete-from :put put)))

(defun sql-delete-from-objectcache (language &key (put nil))
  "Deleting a few rows from `objectcache' table in `xxwiki'"
  ;; we should routinely delete a few rows from the `objectcache'
  ;; table to avoid an accumulation of millions of lines, followed by
  ;; an hour long delay (which causes other queries to time out) when
  ;; MediaWiki purges the cache.
  (put-message-start (_ "delete from `objectcache' limit ") :put put)
  (let* ((object-count (sql-select-count-xxwiki-objectcache language :put put))
	 (delete-from  (format nil "DELETE FROM `~a`.objectcache LIMIT ~d;"
			       (language-code-to-database-name language
							       :put put)
			       *mirror-objectcache-delete-limit*))
	 (threshold-p  (> object-count *mirror-objectcache-threshold*))
	 (result       nil))
    (if threshold-p
	(put-message-value (_ "objectcache above threshold")
			   (format nil "~d > ~d, deleting ~d"
				   object-count 
				   *mirror-objectcache-threshold*
				   *mirror-objectcache-delete-limit*)
			   :put put)
      (put-message-value (_ "objectcache below threshold")
			   (format nil "~d <= ~d, skipping" 
				   object-count 
				   *mirror-objectcache-threshold*)
			   :put put))
    (when threshold-p
	(setq result (query-database-and-return-list-of-strings 
		      delete-from
		      :database nil
		      :put put)))
    (put-message-value-done (_ "delete from `objectcache' limit ") result
			    :put put)
    result))

(defun sql-delete-interwiki-link (database-name language-code &key (put nil))
  "Deleting interwiki link (interlanguage link actually)"
  (let ((delete (format nil "DELETE FROM `~a`.interwiki WHERE iw_prefix='~a';"
			database-name language-code)))
    (query-database-and-return-stream delete
				      :user     *db-wikiadmin-user*
				      :database nil
				      :put      put)))

(defun sql-drop-database (database-name &key (put nil))
  "Dropping database"
  (let ((drop-database  (format nil "DROP DATABASE `~a`;" database-name)))
    (query-database-and-return-stream drop-database
				      :user *db-wikiadmin-user*
				      :database nil
				      :put put)))

(defun sql-drop-database-language-code (language-code &key (put nil))
  "Dropping `xxwiki' database"
  (let ((database-name  (language-code-to-database-name language-code
							:put put)))
    (sql-drop-database database-name :put put)))

(defun sql-drop-database-table (database-name table-name &key (put nil))
  "Dropping table in database"
  (let ((drop-table  (format nil "DROP TABLE ~a;" table-name)))
    (query-database-and-return-list-of-strings drop-table
					       :user *db-wikiadmin-user*
					       :database database-name
					       :put put)))

(defun sql-encrypt (str &key (put nil))
  "Encrypting a string"
  (let* ((select (format nil "SELECT ENCRYPT('~a');" str))
	 (result (query-database-and-return-list-of-strings select
						    :user *db-debian-user*
						    :put put)))
    (first result)))

(defun sql-encrypt-current-timestamp (&key (put nil))
  "Encrypting current timestamp (to generate a pseudo-random password)"
  (let* ((select "SELECT ENCRYPT(CURRENT_TIMESTAMP());")
	 (result (query-database-and-return-list-of-strings select
						    :user *db-debian-user*
						    :put put)))
    (first result)))

(defun sql-grant-global-to-wikiadmin (&key (put nil))
  "Granting some global privileges to `wikiadmin'"
  ;;
  ;; Design note:
  ;;
  ;; Monitor mode needs the global PROCESS privilege to run
  ;;
  ;; mysql-shell> SHOW ENGINE INNODB STATUS\G
  ;;
  (let ((grant (format nil "GRANT PROCESS ON *.* TO '~a'@'~a';"
		       *db-wikiadmin-user* *db-server*)))
    (query-database-and-return-stream grant
				      :user     *db-debian-user*
				      :database nil
				      :put      put)))

(defun sql-grant-all-on-database-to-wikiadmin (database-name &key (put nil))
  "Granting permissions on given database to *db-wikiadmin-user*"
  ;; error if db-name contains a `-' (e.g. `zh-classical')
  (let ((grant (format nil "GRANT ALL PRIVILEGES ON `~a`.* TO '~a'@'~a';"
		       database-name *db-wikiadmin-user* *db-server*)))
    (query-database-and-return-stream grant
				      :user     *db-debian-user*
				      :database nil
				      :put      put)))

(defun sql-grant-siud-on-database-to-wikiuser (database-name &key (put nil))
  "Granting permissions on given database to *db-wikiuser-user*"
  ;; error if db-name contains a `-' (e.g. `zh-classical')
  (let ((grant (concatenate 'string "GRANT SELECT,INSERT,UPDATE,DELETE "
			    (format nil "ON `~a`.* " database-name)
			    (format nil "TO '~a'@'~a';" 
				    *db-wikiuser-user* *db-server*))))
    (query-database-and-return-stream grant
				      :user     *db-debian-user*
				      :database nil
				      :put      put)))

(defun sql-insert-file (dump-language dump-date file-name file-type
				 &key (size nil) (md5sum nil)
				 (page nil) (pages nil) (images nil) 
				 (put nil))
  "Inserting new file"
  (let* ((insert-values (concatenate 'string
	   (format nil "INSERT INTO `~a`.file " *db-wpmirror-name*)
	   "(language,date,name,size,md5sum,type,state,page,pages,images,updates,semaphore) "
	   "VALUES "
	   (format nil "('~a','~a','~a',~d,'~a','~a','~a',~d,~d,~d,~d,~d) "
		   dump-language
		   dump-date
		   file-name
		   (if (null size) 0 size)               ; size
		   (if (null md5sum) "0" md5sum)         ; md5sum
		   file-type                             ;
		   "start"                               ;
		   (if (null page) 0 page)               ; page start
		   (if (null pages) 0 pages)             ; pages
		   (if (null images) 0 images)           ; images
		   0                                     ; updates
		   1)                                    ; semaphore
	   "ON DUPLICATE KEY UPDATE "
	   (format nil "type='~a', " file-type)
	   (format nil "state='start';")))
	 (result (query-database-and-return-stream insert-values
						   :user *db-wikiadmin-user*
						   :put  put)))
    (put-message insert-values :put put)
    t))

(defun sql-insert-image (md5sum &key (state nil) (put nil))
  "Inserting new image"
  (let* ((insert-values (concatenate 'string
	   (format nil "INSERT INTO `~a`.image " *db-wpmirror-name*)
	   "(img_md5sum,img_state) "
	   "VALUES "
	   (format nil "('~a','~a');" 
		   md5sum
		   (if (null state) "created" state))))
	 (result (query-database-and-return-stream insert-values :put put)))
    (put-message insert-values :put put)
    t))

(defun sql-insert-interwiki-link (database-name language-code &key (put nil))
  "Inserting interwiki link (interlanguage link actually)"
  (let ((insert (concatenate 'string
		   (format nil "REPLACE INTO `~a`.interwiki " database-name)
		   "(iw_prefix, iw_url, iw_local, iw_trans, iw_api, iw_wikiid) "
		   (format nil "VALUES ('~a','~a',~d,~d,'~a','~a');"
			   language-code
			   (format nil "http://~a.~a/index.php/$1"
				   language-code *virtual-host-name*)
			   1
			   0
			   (format nil "http://~a.~a/api.php" 
				   language-code *virtual-host-name*)
			   database-name))))
    (query-database-and-return-stream insert
				      :user     *db-wikiadmin-user*
				      :database nil
				      :put      put)))

(defun sql-revoke-all-on-database-to-wikiadmin (database-name &key (put nil))
  "Revoking permissions to *db-wiki-admin* for given database"
  (let ((revoke (format nil "REVOKE ALL PRIVILEGES ON `~a`.* FROM '~a'@'~a';"
			database-name *db-wikiadmin-user* *db-server*)))
    (query-database-and-return-stream revoke
				      :user     *db-debian-user*
				      :database nil
				      :put      put)))

(defun sql-revoke-siud-on-database-to-wikiuser (database-name &key (put nil))
  "Revoking permissions on given database to *db-wikiuser-user*"
  ;; error if db-name contains a `-' (e.g. `zh-classical')
  (let ((revoke (format nil "REVOKE SELECT,INSERT,UPDATE,DELETE ON `~a`.* FROM '~a'@'~a';" 
		       database-name *db-wikiuser-user* *db-server*)))
    (query-database-and-return-stream revoke
				      :user     *db-wikiadmin-user*
				      :database nil
				      :put      put)))

(defun sql-select-image (md5sum &key (put nil))
  "Determining if image-file is in `wpmirror.image'"
  (let* ((select (format nil "SELECT COUNT(*) FROM `~a`.image WHERE img_md5sum='~a';"
			 *db-wpmirror-name* md5sum))
	 (result (sql-select-numeric select :put put)))
    ;; return t if found, nil otherwise
    (> result 0)))

(defun sql-select-image-state (md5sum &key (put nil))
  "Determining if image-file is in `wpmirror.image'"
  (let* ((select (format nil "SELECT img_state FROM `~a`.image WHERE img_md5sum='~a';"
			 *db-wpmirror-name* md5sum)))
    (put-message select :put put)
    (first
     (first
      (query-database-and-return-list-of-lists select :put put)))))

(defun sql-select-languages (&key (put nil))
  "Determining languages seen in `wpmirror.file'"
  (let* ((select (format nil "SELECT language FROM `~a`.file GROUP BY language;"
			 *db-wpmirror-name*))
	 (result (query-database-and-return-list-of-strings select :put put)))
    (debug-message select)
    (debug-message result)
    result))

(defun sql-select-mysql-user (&key (put nil))
  "Selecting DBMS user accounts"
  (let* ((select "SELECT host,user,password FROM mysql.user;")
	 (result (query-database-and-return-list-of-lists select
							  :user *db-debian-user*
							  :put  put)))
    (put-message-start select :put put)
    (when result
      (dolist (row result nil)
	(put-message (format nil "~{~a ~}" row) :put put)))
    (put-message-done select :put put)))

(defun sql-select-numeric (select &key (put nil))
  "Returning a numeric value from a database query"
  (let ((result (query-database-and-return-list-of-lists select :put  put)))
    (put-message select :put put)
    (put-message result :put put)
    (cond ((null result) 0)
	  ((and (stringp (first (first result)))
		(string= (first (first result)) "NULL")) 0)
	  (t (first (first result))))))

(defun sql-select-count (language &key (type nil) (state nil)
				  (semaphore nil) (put nil))
  "Counting records in `wpmirror.file'"
  ;; we centralize code common to most `SELECT COUNT(*)' queries
  (let ((select (concatenate 'string
			     (format nil "SELECT COUNT(*) FROM `~a`.file "
				     *db-wpmirror-name*)
			     (format nil "WHERE language='~a' " language)
			     (unless (null type)
			       (format nil "AND type='~a' " type))
			     (unless (null state)
			       (format nil "AND state='~a' " state))
			     (unless (null semaphore)
			       (format nil "AND semaphore='~d' " semaphore))
			     ";")))
     (sql-select-numeric select :put put)))

(defun sql-select-count-file (language &key (put nil))
  "Counting records in `wpmirror.file' of given language"
  (sql-select-count language :put put))

(defun sql-select-count-ichunk (language &key (put nil))
  "Counting `ichunk's given language"
  (sql-select-count language :type "ichunk" :put put))

(defun sql-select-count-ichunk-done (language &key (put nil))
  "Counting `ichunk's that are done"
  (sql-select-count language :type "ichunk" :state "done" :put put))

(defun sql-select-count-ichunk-fail (language &key (put nil))
  "Counting `ichunk's that have failed"
  (sql-select-count language :type "ichunk" :state "fail" :put put))

(defun sql-select-count-ichunk-remaining (language &key (put nil))
  "Counting `ichunk's remaining to be processed"
  (- (sql-select-count-ichunk      language :put nil)
     (sql-select-count-ichunk-done language :put nil)
     (sql-select-count-ichunk-fail language :put nil)))

(defun sql-select-count-ichunk-running (language &key (put nil))
  "Counting `ichunks' that are currently being processed"
  (sql-select-count language :type "ichunk" :semaphore 0 :put put))

(defun sql-select-count-image-valid (&key (put nil))
  "Counting validated image-files in `wpmirror.image'"
  (let ((select (format nil "SELECT COUNT(*) FROM `~a`.image WHERE img_state='valid';"
			*db-wpmirror-name*)))
    (sql-select-numeric select :put put)))

(defun sql-select-count-images (language &key (put nil))
  "Counting `images' records"
  (sql-select-count language :type "images" :put put))

(defun sql-select-count-images-start (language &key (put nil))
  "Counting `images' records of state `start'"
  (sql-select-count language :type "images" :state "start" :put put))

(defun sql-select-count-images-created (language &key (put nil))
  "Counting `images' records of state `created'"
  (sql-select-count language :type "images" :state "created" :put put))

(defun sql-select-count-images-valid (language &key (put nil))
  "Counting `images' records of state `valid'"
  (sql-select-count language :type "images" :state "valid" :put put))

(defun sql-select-count-tables (database &key (put nil))
  "Counting tables in a database"
  (let* ((select (format nil (concatenate 'string
		   "SELECT COUNT(*) FROM information_schema.tables "
		   (format nil "WHERE table_schema='~a';" database))))
	 (result (query-database-and-return-list-of-lists select
							  :user *db-wikiadmin-user*
							  :put put)))
    (put-message select :put put)
    (if (null result)
	0
      (first (first result)))))

(defun sql-select-count-xxwiki-objectcache (language-code &key (put nil))
  "Counting `objectcache' records from `xxwiki'"
  (let ((select (format nil "SELECT COUNT(*) FROM `~a`.objectcache;"
			(language-code-to-database-name language-code
							:put put))))
    (sql-select-numeric select :put put)))

(defun sql-select-count-xxwiki-image (language &key (put nil))
  "Counting `image' records from `xxwiki'"
  (let ((select (format nil "SELECT COUNT(*) FROM `~a`.image;"
			(language-code-to-database-name language :put put))))
    (sql-select-numeric select :put put)))

(defun sql-select-count-xxwiki-page (language &key (put nil))
  "Counting `image' records from `xxwiki'"
  (let ((select (format nil "SELECT COUNT(*) FROM `~a`.page WHERE page_namespace=0;"
			(language-code-to-database-name language :put put))))
    (sql-select-numeric select :put put)))

(defun sql-select-count-xchunk (language &key (put nil))
  "Counting xchunks"
  (sql-select-count language :type "xchunk" :put put))

(defun sql-select-count-xchunk-done (language &key (put nil))
  "Counting xchunks that are done"
  (sql-select-count language :type "xchunk" :state "done" :put put))

(defun sql-select-count-xchunk-fail (language &key (put nil))
  "Counting xchunks that have failed"
  (sql-select-count language :type "xchunk" :state "fail" :put put))

(defun sql-select-count-xchunk-remaining (language &key (put nil))
  "Counting `xchunk's remaining to be processed"
  (- (sql-select-count-xchunk      language :put nil)
     (sql-select-count-xchunk-done language :put nil)
     (sql-select-count-xchunk-fail language :put nil)))

(defun sql-select-count-xchunk-running (language &key (put nil))
  "Counting xchunks that are currently being processed"
  (sql-select-count language :type "xchunk" :semaphore 0 :put put))

(defun sql-select-dump-name (language &key (put nil))
  "Selecting name of dump file of given language"
  (let* ((select (format nil "SELECT name FROM `~a`.file WHERE type='dump' and language='~a';"
			 *db-wpmirror-name* language))
	 (result (query-database-and-return-list-of-lists select :put put)))
    (put-message select :put put)
    (if (null result)
	nil
      (first (first result)))))

(defun sql-select-dump-xml-name (language &key (put nil))
  "Selecting name of the `xml' file of a given language"
  (let* ((select (format nil "SELECT name FROM `~a`.file WHERE type='xml' and language='~a';" 
			 *db-wpmirror-name* language))
	 (result (query-database-and-return-list-of-lists select :put put)))
    (put-message select :put put)
    (if (null result)
	nil
      (first (first result)))))

(defun sql-select-file (file-name &key (put nil))
  "Selecting file, returning list of lists"
  (let ((select (format nil "SELECT language,date,name,size,md5sum,type,state,page,pages,images,updates,semaphore FROM `~a`.file WHERE name='~a';"
			*db-wpmirror-name* file-name)))
    (put-message select :put put)
    (query-database-and-return-list-of-lists select :put put)))

(defun sql-select-file-all (&key (put nil))
  "Selecting all, returning list of strings"
  (let ((select (format nil "SELECT * FROM `~a`.file;" *db-wpmirror-name*)))
    (put-message select :put put)
    (query-database-and-return-list-of-strings select :put put)))
  
(defun sql-select-file-date (file-name &key (put nil))
  "Selecting file language, return string or nil"
  (let ((select (format nil "SELECT date FROM `~a`.file WHERE name='~a';"
			*db-wpmirror-name* file-name)))
    (first
     (query-database-and-return-list-of-strings select :put put))))
  
(defun sql-select-file-images (file-name &key (put nil))
  "Selecting images in file, return number or nil"
  (let ((select (format nil "SELECT images FROM `~a`.file WHERE name='~a';"
			*db-wpmirror-name* file-name)))
    (sql-select-numeric select :put put)))
  
(defun sql-select-file-language (file-name &key (put nil))
  "Selecting file language, return string or nil"
  (let ((select (format nil "SELECT language FROM `~a`.file WHERE name='~a';"
			*db-wpmirror-name* file-name)))
    (first
     (query-database-and-return-list-of-strings select :put put))))
  
(defun sql-select-file-md5sum (file-name &key (put nil))
  "Selecting file md5sum, return string or nil"
  (let ((select (format nil "SELECT md5sum FROM `~a`.file WHERE name='~a';"
			*db-wpmirror-name* file-name)))
    (first
     (query-database-and-return-list-of-strings select :put put))))
  
(defun sql-select-file-page (file-name &key (put nil))
  "Selecting file page, return number or nil"
  (let ((select (format nil "SELECT page FROM `~a`.file WHERE name='~a';"
			*db-wpmirror-name* file-name)))
    (sql-select-numeric select :put put)))
  
(defun sql-select-file-pages (file-name &key (put nil))
  "Selecting file pages in file, return number or nil"
  (let ((select (format nil "SELECT pages FROM `~a`.file WHERE name='~a';"
			*db-wpmirror-name* file-name)))
    (sql-select-numeric select :put put)))
  
(defun sql-select-file-semaphore (file-name &key (put nil))
  "Selecting file semaphore, return 0, 1, or nil"
  (let ((select (format nil "SELECT semaphore FROM `~a`.file WHERE name='~a';"
			*db-wpmirror-name* file-name)))
    (sql-select-numeric select :put put)))

(defun sql-select-file-size (file-name &key (put nil))
  "Selecting file size, return number or nil"
  (let ((select (format nil "SELECT size FROM `~a`.file WHERE name='~a';"
			*db-wpmirror-name* file-name)))
    (sql-select-numeric select :put put)))

(defun sql-select-file-state (file-name &key (put nil))
  "Selecting file state, return string or nil"
  (let ((select (format nil "SELECT state FROM `~a`.file WHERE name='~a';"
			*db-wpmirror-name* file-name)))
    (first
     (first
      (query-database-and-return-list-of-lists select :put put)))))

(defun sql-select-file-type (file-name &key (put nil))
  "Selecting file type, return string or nil"
  (let ((select (format nil "SELECT type FROM `~a`.file WHERE name='~a';"
			*db-wpmirror-name* file-name)))
    (first
     (first
      (query-database-and-return-list-of-lists select :put put)))))

(defun sql-select-file-updates (file-name &key (put nil))
  "Selecting file updates, return number or nil"
  (let ((select (format nil "SELECT updates FROM `~a`.file WHERE name='~a';"
			*db-wpmirror-name* file-name)))
    (sql-select-numeric select :put put)))

(defun sql-select-sum-images-images (language &key (put nil))
  "Counting images in `images' records"
  (let ((select (format nil "SELECT SUM(images) FROM `~a`.file WHERE language='~a' AND type='images';" 
			*db-wpmirror-name* language)))
    (sql-select-numeric select :put put)))

(defun sql-select-sum-images-pages (language &key (put nil))
  "Counting validated images in `images' records"
  (let ((select (format nil "SELECT SUM(pages) FROM `~a`.file WHERE language='~a' AND type='images';" 
			*db-wpmirror-name* language)))
    (sql-select-numeric select :put put)))

(defun sql-select-time-zone-name (time-zone-name &key (put nil))
  "Selecting `time_zone_id' from `mysql.time_zone_name' table for given zone"
  (let* ((select (concatenate 'string
			      "SELECT name FROM mysql.time_zone_name "
			      (format nil "WHERE name='~a';" time-zone-name)))
	 (result (query-database-and-return-list-of-strings select
						    :user *db-debian-user*
						    :put put))
	 (name-found (first result)))
    (put-message-value (_"time-zone-name found") name-found :put put)
    name-found))

(defun sql-select-next-file-type-state (file-type file-state &key (put nil))
  "Selecting next file of given type and state, return filename"
  ;; 
  ;; Design note:
  ;;
  ;; Grabbing the lock by one mirror must be Isolated (the `I' in
  ;; `ACID') from attempts to grab the lock by other mirrors.  
  ;;
  ;; For this reason we use `SELECT ... FOR UPDATE' followed by
  ;; `UPDATE ...'.  The first identifies the qualifying row and
  ;; request InnoDB to establish an `IX' lock on that row, while the
  ;; second grabs the lock.
  ;;
  ;; Wrapping the transaction with `BEGIN TRANSACTION' and `COMMIT'
  ;; makes the entire transaction Atomic (the `A' in `ACID')
  ;;
  (let* ((message-select   (_ "sql-select-next-file-type-state select"))
	 (message-result   (_ "sql-select-next-file-type-state result"))
	 (message-state    (_ "sql-select-next-file-type-state state "))
	 (message-file     (_ "sql-select-next-file-type-state file  "))
	 (file-type-state (format nil "~{~a ~}" (list file-type file-state)))
	 (select (concatenate 'string
	       "START TRANSACTION; "                 ; <--- atomic
	       "SELECT @name := name AS name,@semaphore := semaphore AS semaphore "
	       (format nil "FROM `~a`.file " *db-wpmirror-name*)
	       (format nil "WHERE type='~a' " file-type)
	       (format nil "AND state='~a' " file-state)
	       "AND semaphore=1 "                    ; <--- free
	       "ORDER BY name,state "
	       "LIMIT 1 "
	       "FOR UPDATE; "                        ; <--- atomic
	       (format nil "UPDATE `~a`.file SET semaphore=0 " 
		       *db-wpmirror-name*)          ; <--- locked
	       "WHERE name=@name; "
	       "COMMIT;"))                           ; <--- atomic
	 (result (query-database-and-return-list-of-lists select :put put))
	 (file-name (first (first result))))
    (debug-message-value message-state  file-type-state)
    (debug-message-value message-select select)
    (debug-message-value message-result result)
    (debug-message-value message-file   file-name)
    (unless (null result) file-name)))

(defun sql-select-next-file (&key (put nil))
  "Selecting next file for processing by the finite state machine"
  ;; Each file has a semaphore to assure Isolation (the `I' in ACID)
  (let ((message    (_ "selecting next file for processing"))
	(file-name  nil))
    (put-timestamp-message-start message :put put)
    ;; select highest priority file
    (setq file-name
	  (loop 
	     with file-name = nil
	     for (type state concurrency non-image) in *type-state-priority*
	       when (and (or (eql *main-mode* :first-mirror) concurrency)
			 (or *mirror-image-download-p* non-image))
	       do (setq file-name (sql-select-next-file-type-state 
				   type state :put put))
	     thereis file-name))
    ;; notify of grab lock
    (if (null file-name)
	(progn
	  (put-message (_ "no file ready for processing") :put put)
	  (sleep 1))
      (progn
	(put-message (format nil "~a" file-name) :put put)
	(put-message-value (_ "grab lock: set semaphore to")
			   (sql-select-file-semaphore file-name) :put put)
	(put-timestamp-message-done message :put put)))
    file-name))

(defun sql-select-ichunk-wip (language &key (put nil))
  "Selecting ichunk in progess (if any)"
  (let* ((select (format nil "SELECT name FROM `~a`.file WHERE language='~a' AND type='ichunk' AND state='valid';" 
			 *db-wpmirror-name* language))
	 (result (query-database-and-return-list-of-strings select :put put)))
    (if (null result)
	(put-message-value (_ "no ichunks in progress") result :put put)
      (put-message-value (_ "ichunks in progress") (cdr result) :put put))
    (cdr result)))

(defun sql-select-xchunk-wip (language &key (put nil))
  "Selecting xchunk in progess (if any)"
  (let* ((select (format nil "SELECT name FROM `~a`.file WHERE language='~a' AND type='xchunk' AND state='valid';" 
			 *db-wpmirror-name* language))
	 (result (query-database-and-return-list-of-strings select :put put)))
    (if (null result)
	(put-message-value (_ "no xchunks in progress") result :put put)
      (put-message-value (_ "xchunks in progress") (cdr result) :put put))
    (cdr result)))

(defun sql-show-databases (&key (user *db-wikiadmin-user*) 
				(put  nil))
  "Showing all databases"
  (let* ((show-databases "SHOW DATABASES;")
	 (result         (query-database-and-return-list-of-strings
			  show-databases
			  :user     user
			  :database nil
			  :put      put)))
    (put-message result :put put)
    result))

(defun sql-show-databases-wpmirror (&key (user *db-wikiadmin-user*)
				    (put nil))
  "Showing all databases related to wp-mirror"
  (let* ((show-databases-1 "SHOW DATABASES LIKE '%wiki%';")
	 (show-databases-2 "SHOW DATABASES LIKE 'wpmirror';")
	 (result           (append 
			    (query-database-and-return-list-of-strings
			     show-databases-1
			     :user     user
			     :database nil
			     :put      put)
			    (query-database-and-return-list-of-strings 
			     show-databases-2
			     :user     user
			     :database nil
			     :put      put))))
    (put-message result :put put)
    result))

(defun sql-show-grants (user &key (put nil))
  "Showing grants for given user"
  (let ((show-grants (format nil "SHOW GRANTS FOR `~a`@`~a`;" 
			     user *db-server*)))
    (query-database-and-return-list-of-strings show-grants
					       :user     *db-debian-user*
					       :database nil
					       :put      put)))

(defun sql-show-tables (db-name &key (user *db-wikiuser-user*)
				(put  nil))
  "Showing tables in *db-wpmirror-name*"
  (let ((show-tables  (format nil "SHOW TABLES FROM `~a`;" db-name)))
    (query-database-and-return-list-of-strings show-tables
					       :user     user
					       :database nil
					       :put      put)))

(defun sql-show-table-status (db-name table-name &key (user *db-wikiuser-user*)
				      (put nil))
  "Showing table status"
  (let ((show-table-status
	 (format nil "SHOW TABLE STATUS FROM `~a` LIKE '~a';" 
		 db-name table-name)))
    (query-database-and-return-list-of-lists show-table-status
					     :user user
					     :put  put)))

(defun sql-show-table-status-data-free (&key (db-name *db-wpmirror-name*)
					     (table-name "file") 
					     (put nil))
  "Showing data_free as given by table status"
  (let ((result (sql-show-table-status db-name table-name :put put)))
    (if (null result)
	0
      (tenth (first result)))))

(defun sql-show-datadir (&key (put nil))
  "Showing data directory"
  (sql-show-variable "datadir" :put put))

(defun sql-show-innodb-buffer-pool-pages-dirty (&key (put nil))
  "Showing number of buffer pool pages that are modified"
  (sql-show-status "Innodb_buffer_pool_pages_dirty" :put put))

(defun sql-show-innodb-buffer-pool-pages-insert (&key (put nil))
  "Showing number of buffer pool pages used for insert buffer"
  ;;
  ;; Design note:
  ;;
  ;; Ideally, we would call
  ;;
  ;;   (sql-show-status "Innodb_buffer_pool_pages_insert" :put put))
  ;;
  ;; However, as of MySQL 5.5, InnoDB does not expose any status
  ;; variables regarding the insert buffer.  So for now we use a shell
  ;; script that queries:
  ;;
  ;;   SHOW ENGINE INNODB STATUS\G
  ;;
  ;; and then `grep's the output for the line containing `Ibuf: size'
  ;; which looks like:
  ;;
  ;;   Ibuf: size 1, free list len 0, seg size 2, 1 merges
  ;;
  ;; and then reads the third token; which, in this case, is `1'.
  ;;
  (let* ((pattern  "Ibuf: size")
	 (l        (shell-show-engine-innodb-status-grep pattern)))
    (if (or (null l) (zerop (length l)))
	0
      (with-input-from-string (s (third (regexp:regexp-split " " l)))
			      (read s)))))

(defun sql-show-innodb-buffer-pool-pages-misc (&key (put nil))
  "Showing number of buffer pool pages used for miscellaneous purposes"
  (sql-show-status "Innodb_buffer_pool_pages_misc" :put put))

(defun sql-show-innodb-buffer-pool-pages-total (&key (put nil))
  "Showing number of buffer pool pages"
  (sql-show-status "Innodb_buffer_pool_pages_total" :put put))

(defun sql-show-innodb-data-file-path (&key (put nil))
  "Showing filepath to InnoDB table space"
  (put-message (_ "Showing InnoDB data file path") :put put)
  (let* ((datadir   (sql-show-datadir :put put))
	 (result    (sql-show-variable "innodb_data_file_path" :put put))
	 (file-path nil)
	 (size      nil)
	 (rawp      nil))
    (if (null result)
	(put-message-value-fail-and-abort
	 (_ "show innodb_data_file_path returned") result)
      (progn
	(setq file-path result)
	(put-message-value (_ "innodb_data_file_path") file-path :put put)))
    (put-message-value (_ "innodb_data_file_path") file-path :put put)
    ;; this file path may contain a colon ':' such as 
    ;;   'ibdata1:10M:autoextend'              # this is the default
    ;;   '/dev/mapper/vg0-ibdata0:300000Mraw'  # this is a raw partition 
    ;; HDD manufacturers define 'K' as 1000, 'M' as 1000^2, and 'G' as 1000^3
    ;; MySQL however defines    'K' as 1024, 'M' as 1024^2, and 'G' as 1024^3
    (setq size      
	  (string-right-trim
	   "RAWraw" (second (regexp:regexp-split ":" file-path))))
    (setq size
	  (cond ((regexp:match "K" size)
		 (* (read-from-string (string-right-trim "K" size)) 1024))
		((regexp:match "M" size)
		 (* (read-from-string (string-right-trim "M" size)) 1024 1024))
		((regexp:match "G" size)
		 (* (read-from-string (string-right-trim "G" size)) 
		    1024 1024 1024))
		(t (read-from-string size))))
    (setq rawp      
	  (or (regexp:match 
	       "raw" (second (regexp:regexp-split ":" file-path)))
	      (regexp:match 
	       "RAW" (second (regexp:regexp-split ":" file-path)))))
    (setq file-path (first  (regexp:regexp-split ":" file-path)))
    (put-message-value (_ "innodb_data_file_path trimmed") file-path :put put)
    (put-message-value (_ "innodb_data_file_path size")    size      :put put)
    (put-message-value (_ "innodb_data_file_path rawp")    rawp      :put put)
    ;; the default case only gives the file-name, so merge directory path
    (let ((path-dir (pathname-directory (parse-namestring file-path))))
      (when (null path-dir)
	(put-message-value (_ "datadir") datadir :put put)
	(setq file-path (merge-pathnames file-path datadir))
	(put-message-value (_ "table space path") file-path :put put)))
    (values file-path size rawp)))

(defun sql-show-innodb-file-format (&key (put nil))
  "Showing InnoDB file format (`Antelope' or `Barracuda')"
  (sql-show-variable "innodb_file_format" :put put))

(defun sql-show-innodb-data-home-dir (&key (put nil))
  "Showing InnoDB data home directory"
  (sql-show-variable "innodb_data_home_dir" :put put))

(defun sql-show-status (str &key (put nil))
  "Showing MySQL status `str'"
  (let ((result (query-database-and-return-list-of-lists
		 (format nil "SHOW STATUS LIKE '~a';" str)
		 :user     *db-wikiadmin-user*
		 :database nil
		 :put put)))
    (if (null result)
	0
      (second (first result)))))

(defun sql-show-variable (str &key (put nil))
  "Showing MySQL variable `str'"
  (let ((result (query-database-and-return-list-of-lists
		 (format nil "SHOW VARIABLES LIKE '~a';" str)
		 :user     *db-wikiadmin-user*
		 :database nil
		 :put put)))
    (if (null result)
	0
      (second (first result)))))

(defun sql-update-fail-to-pending (&key (put nil))
  "Updating all failed `xchunk's and `ichunk's"
  (put-message-start (_ "updating state from `fail' to `pending'") :put put)
  (put-message (_ "to give failed chunks a second chance") :put put)
  (let ((update (format nil "UPDATE `~a`.file SET state='pending' WHERE state='fail' AND (type='xchunk' OR type='ichunk');" 
			*db-wpmirror-name*)))
    (put-message update :put put)
    (query-database-and-return-stream update :put put)
    (put-message-done (_ "updating state from `fail' to `pending'") :put put)
    t))

(defun sql-update-fail-to-start (&key (put nil))
  (put-message-start (_ "updating state from `fail' to `start'") :put put)
  (put-message (_ "to give failed dump a second chance") :put put)
  (let ((update (format nil "UPDATE `~a`.file SET state='start' WHERE state='fail' AND type='dump';"
			*db-wpmirror-name*)))
    (put-message update :put put)
    (query-database-and-return-stream update :put put)
    (put-message-done (_ "updating state from `fail' to `start'") :put put)
    t))

(defun sql-update-images-to-pending (&key (put nil))
  "Updating all `images' to `pending'"
  (put-message-start (_ "updating state from `done' to `pending'") :put put)
  (put-message (_ "to have `chown' run again") :put put)
  ;; do just once (not once for each language)
  (let ((update (format nil "UPDATE `~a`.file SET state='pending' WHERE language='~a' AND state='done' AND type='images';"
			*db-wpmirror-name*
			(first (mirror-language-list :put put)))))
    (put-message update :put put)
    (query-database-and-return-stream update :put put)
    (put-message-done (_ "updating state from `done' to `pending'") :put put)
    t))

(defun sql-update-file (file-name &key (size nil) (md5sum nil) (state nil)
				  (page nil) (pages nil) (images nil)
				  (semaphore nil)
				  (put nil))
  "Updating file size, md5sum, state, page, pages, images, updates"
  (let* ((old-state   (sql-select-file-state   file-name :put put))
	 (old-updates (sql-select-file-updates file-name :put put))
	 (new-updates (1+ old-updates))
	 (update (concatenate 'string
	           (format nil "UPDATE `~a`.file SET " *db-wpmirror-name*)
		   (unless (null size)     (format nil "size=~d,"       size))
		   (unless (null md5sum)   (format nil "md5sum='~a'," md5sum))
		   (unless (null state)    (format nil "state='~a',"   state))
		   (unless (null page)     (format nil "page=~d,"       page))
		   (unless (null pages)    (format nil "pages=~d,"     pages))
		   (unless (null images)   (format nil "images=~d,"   images))
		   (unless (null semaphore)
		     (format nil "semaphore=~d," semaphore))
		   (format nil "updates=~d " new-updates)
		   (format nil "WHERE name='~a';" file-name))))
    (put-message-value (_ "old-state")         old-state   :put put)
    (put-message-value (_ "old-updates")       old-updates :put put)
    (put-message-value (_ "sql-update-string") update      :put put)
    (query-database-and-return-stream update :put put)
    (put-message-value (_ "new file record") 
		       (sql-select-file file-name :put nil)
		       :put put)
    t))

(defun sql-update-file-images (file-name images &key (put nil))
  "Updating images per file"
  (sql-update-file file-name :images images :put put))

(defun sql-update-file-md5sum (file-name md5sum &key (put nil))
  "Updating file md5sum"
  (sql-update-file file-name :md5sum md5sum :put put))

(defun sql-update-file-page (file-name page &key (put nil))
  "Updating file start page"
  (sql-update-file file-name :page page :put put))

(defun sql-update-file-pages (file-name pages &key (put nil))
  "Updating pages per file"
  (sql-update-file file-name :pages pages :put put))

(defun sql-update-file-semaphore (file-name semaphore &key (put nil))
  "Updating semaphore"
  ;; Only use this function for releasing the lock (setting to
  ;; semaphore to 1).  Grabbing the lock must be isolated from other
  ;; transactions (see sql-select-next-file-type for details)
  (let ((old-semaphore (sql-select-file-semaphore file-name :put put)))
    (sql-update-file file-name :semaphore semaphore :put put)
    ;(sleep 1)
    (let ((new-semaphore (sql-select-file-semaphore file-name :put put)))
      (if (eql new-semaphore 0)
	  (put-message-value (_ "grab lock - update semaphore") 
			     (format nil "~d-->~d" old-semaphore new-semaphore)
			     :put put)
	(put-message-value (_ "release lock - update semaphore") 
			   (format nil "~d-->~d" old-semaphore new-semaphore)
			   :put put)))))

(defun sql-update-file-semaphore-all (semaphore &key (put nil))
  "Updating all semaphores"
  (let ((update (format nil "UPDATE `~a`.file SET semaphore=~d;"
			*db-wpmirror-name* semaphore)))
    (put-message-value (_ "update") update :put put)
    (query-database-and-return-stream update :put put)
    t))

(defun sql-update-file-size (file-name &key (put nil))
  "Updating file size"
  (let ((record-size (sql-select-file-size file-name :put put))
	(actual-size nil))
    (put-message-value (_ "record size") record-size :put put)
    (if (ext:probe-pathname file-name)
	(progn
	  (setq actual-size (posix:file-size file-name))
	  (put-message-value (_ "actual size") actual-size :put put)
	  (unless (eql record-size actual-size)
	    (sql-update-file file-name :size actual-size :put put)))
      (put-message-value-fail-and-abort (_ "file not found") file-name)))
  t)

(defun sql-update-file-state (file-name state &key (put nil))
  "Updating file state"
  (sql-update-file file-name :state state :put put))

  
;;;;--------------------------------------------------------------------------+
;;;; Www utilities:                                                           |
;;;;--------------------------------------------------------------------------+


(defun download-http-head-and-file (url file-name &key (put nil))
  "Downloading file and its HTTP head from URL"
  (and (null (shell-curl-http-head url file-name :put put))
       (null (shell-curl-http-file url file-name :put put))))

(defun download-file (url file-name &key (put nil))
  "Downloading file and its HTTP head from given URL.
Cache header to avoid retrieving existing file."
  (let ((file-head (format nil "~a.head" file-name))
	(tmp-head  (format nil "~a.head" *whereis-file-tmp-http*))
	(wip-file  (format nil "~a.wip" file-name)))
    (if (and (ext:probe-pathname file-name)
	     (ext:probe-pathname file-head)
	     (not (ext:probe-pathname wip-file)))
	;; if both exist (and no `.wip' as evidence of partial download), 
	;; then download tmp-head, and compare ETags
	(when (null (shell-curl-http-head url *whereis-file-tmp-http* :put put))
	  (let* ((file-head-etag (grep-first file-head "ETag"))
		 (tmp-head-etag  (grep-first tmp-head  "ETag")))
	    ;; if ETags do not match, then download file and its head
	    (put-message-value (_ "ETag from head of server file")  
			       tmp-head-etag :put put)
	    (put-message-value (_ "ETag from head of local file") 
			       file-head-etag :put put)
	    (if (equal file-head-etag tmp-head-etag)
		(put-message (_ "server file no newer than local file---not retrieving") :put put)
	      (download-http-head-and-file url file-name :put put))))
      ;; If file or head missing, then download both
      (progn
	(put-message (_ "file or head missing---try to retrieve both") :put put)
	(download-http-head-and-file url file-name :put put)))))


;;;;--------------------------------------------------------------------------+
;;;; Error messages detailed:                                                 |
;;;;--------------------------------------------------------------------------+


(defparameter *error-messages* (make-hash-table :test 'equal))


;;;;--------------------------------------------------------------------------+
;;;; Assert prerequisite hardware:                                            |
;;;;--------------------------------------------------------------------------+


(defun assert-disk-space-if-large-wikipedia-p (&key (put nil))
  "Asserting adequate free space on partition"
  (let* ((partition-images-free (system-partition-free-images :put put))
	 (space-have (format-integer-for-human partition-images-free))
         (space-warn (format-integer-for-human
		      *system-partition-free-images-start-min*))
	 (space-p      (>= partition-images-free
			   *system-partition-free-images-start-min*))
	 (language-code-list (mirror-language-list :put put))
	 (large-wiki-p (and (intersection language-code-list
					  *wikimedia-large-wiki-list*
					  :test #'string=)
			    t)))
    (put-message-value (_ "detecting") space-have :put put)
    ;; assert if we have one of the largest wiki's
    (put-message-value (_ "our wiki languages are") language-code-list :put put)
    (put-message-value (_ "large wiki's are") *wikimedia-large-wiki-list*
		       :put put)
    (put-message-value (_ "are any of our wiki's large") large-wiki-p :put put)
    (when large-wiki-p
      (put-flag-message :info
	      (format nil "disk space below threshold for top ten (~a < ~a)" 
		      space-have space-warn))
      (put-flag-message :done (_ "done")))
    (or (not large-wiki-p) space-p)))

(defun warn-if-disk-space-low-p (&key (put nil))
  "Warning if low free space on partition"
  (let* ((partition-images-free (system-partition-free-images :put put))
	 (space-have   (format-integer-for-human partition-images-free))
         (space-warn   (format-integer-for-human
			*system-partition-free-images-warn*))
	 (space-p      (>= partition-images-free
			*system-partition-free-images-warn*)))
    (put-message-value (_ "detecting")        space-have :put put)
    (put-message-value (_ "warning if below") space-warn :put put)
    (when (null space-p)
      (put-flag-message :info
			(format nil "disk space below threshold (~a < ~a)" 
				space-have space-warn))
      (put-flag-message :done (_ "done")))
    space-p))

(defun system-partition-free-images (&key (put nil))
  "Determining free space [bytes] on HDD partition that stores images"
  (system-partition-free *whereis-directory-images* :put put))

(defun system-partition-free-innodb (&key (put nil)) 
  "Determining free space [bytes] on HDD partition that stores InnoDB table space"
  (multiple-value-bind (path size rawp) 
      (sql-show-innodb-data-file-path :put put)
    (if rawp
	(sql-show-table-status-data-free :put put)
      (system-partition-free path :put put))))

(defun system-partition-free (path &key (put nil))
  "Determining free space [bytes] on HDD partition that holds `path'"
  (let* ((fs             (posix:stat-vfs path))           ; this is a struct
	 (bfree          (stat-vfs-bfree  fs))            ; accessor
	 (frsize         (stat-vfs-frsize fs))            ; accessor
	 (partition-free (* bfree frsize)))
    (debug-message-value (_ "free blocks") (stat-vfs-bfree  fs))
    (debug-message-value (_ "block size")  (stat-vfs-frsize fs))
    partition-free))

(defun system-partition-size-images (&key (put nil))
  "Determining size [bytes] of HDD partition that stores images"
  (system-partition-size *whereis-directory-images* :put put))

(defun system-partition-size-innodb (&key (put nil))
  "Determining size [bytes] of HDD partition that stores InnoDB table space"
  (multiple-value-bind (path size rawp) 
      (sql-show-innodb-data-file-path :put put)
    (if rawp
	size
      (system-partition-size path :put put))))

(defun system-partition-size (path &key (put nil))
  "Determining size [bytes] of HDD partition that stores `path'"
  (let* ((fs             (posix:stat-vfs path))           ; this is a struct
	 (blocks         (stat-vfs-blocks fs))            ; accessor
	 (frsize         (stat-vfs-frsize fs))            ; accessor
	 (partition-size (* blocks frsize)))
    (debug-message-value (_ "blocks")      (stat-vfs-blocks fs))
    (debug-message-value (_ "block size")  (stat-vfs-frsize fs))
    partition-size))

(defun assert-hdd-write-cache-disabled-p (&key (put nil))
  "Asserting IDE/SATA drive write cache is disabled"
  ;;
  ;; Design note:
  ;;
  ;; 1) Hard disks have a write cache in order to schedule the
  ;; movements of the read-write heads more efficiently, using
  ;; something like the `elevator algorythm'.
  ;;
  ;; Disabling disk write cache is however important for an ACID
  ;; compliant DBMS.  The issue here is Durability (the `D' in
  ;; `ACID').  Transactions that `commit' must actually be written to
  ;; disk, and not remain in cache, where they may be lost during
  ;; system failure.
  ;;
  ;; 2) In step 1 below, MySQL is queried as to its InnoDB data file
  ;; path.  This means that the hard drive is identified after MySQL
  ;; credentials have been asserted.  Simply put, the H/W feature is
  ;; asserted after the S/W feature.  An alternative design would have
  ;; been to parse the MySQL configuration file for information
  ;; regarding the InnoDB data file path.  Then the H/W feature could
  ;; have been asserted before MySQL was.  With `GNU/Linux', the
  ;; config files are `/etc/mysql/*cnf' and `/etc/mysql/conf.d/*.cnf'.
  ;;
  ;; 3) In step 6 below, `sysfs' is used to identify the hard drive
  ;; underlying the InnoDB data file.  `sysfs' exists to make system
  ;; resource discovery easier.  A prudent disk storage solution
  ;; features [LVM2 over LUKS over RAID over whole disks].  Without
  ;; `sysfs' this function could have been a chore to write.
  ;;
  (let ((table-space-path           nil)
	(partition-or-file          nil)
	(absolute-path              nil)
	(device-id                  nil)
	(device-id-major            nil)
	(device-id-minor            nil)
	(device-id-major-minor      nil)
	(block-device-init          nil)
	(sys-pathname-init          nil)
	(device-name                nil)
	(dev-path                   (make-pathname :directory "dev"))
	(dir-list-init              nil)
	(physical-block-device-list nil)
	(write-cache-disabled-p     t))
    ;; 1) Get path to table space, usually `/var/lib/mysql/ibdata1'
    (setq table-space-path (sql-show-innodb-data-file-path :put put))
    (put-message-value (_ "table space path") table-space-path :put put)

    ;; 2) path may be a symbolic link---chase it
    (setq absolute-path (ext:probe-pathname table-space-path))
    (put-message-value (_ "absolute path") absolute-path :put put)

    ;; 3) Innodb data file path will be either a file or a partition
    (if (zerop (posix:file-size absolute-path))
	(setq partition-or-file :partition)
      (setq  partition-or-file :file))
    (put-message-value (_ "path is partition or file") partition-or-file
		       :put put)
    
    (case partition-or-file
      (:file 
       ;; 4a) find device ID of device containing file (but it may be mapped)
       (setq device-id (file-stat-dev (posix:file-stat absolute-path)))
       (debug-message-value (_ "device-id") device-id)
       (multiple-value-setq (device-id-major device-id-minor)
	 (floor device-id 256))
       (setq device-id-major-minor (format nil "~d:~d"
					   device-id-major device-id-minor))
       (put-message-value (_ "device-id-major-minor") device-id-major-minor
			  :put put)
       ;; 4b) set initial block-device and dir-list
       (setq block-device-init device-id-major-minor)
       (setq dir-list-init (append '(:absolute "sys" "dev" "block")
				   (list block-device-init))))
      (:partition
       ;; 5a) assert `/dev/foo', block device is `foo' (but it may be mapped)
       (debug-message-value (_ "dev directory")  (pathname-directory dev-path))
       (debug-message-value (_ "pathname-directory data")
			    (pathname-directory absolute-path))
       (debug-message-value (_ "pathname-name data")
			    (pathname-name absolute-path))
       (unless (equal (pathname-directory absolute-path)
		      (pathname-directory dev-path))
	 (put-message-value-fail-and-abort (_ "should be block device")
					   absolute-path))
       ;; 5b) set initial block-device and dir-list
       (setq block-device-init (pathname-name absolute-path))
       (put-message-value (_ "block-device-init") block-device-init :put put)
       (setq dir-list-init (append '(:absolute "sys" "block")
				   (list block-device-init)))))
    (debug-message-value (_ "dir-list-init") dir-list-init)
    (setq sys-pathname-init (make-pathname :directory dir-list-init))
    (put-message-value (_ "sys-pathname-init") sys-pathname-init :put put)

    ;; 6) now go to `/sys', find block device, and chase slaves
    (setq physical-block-device-list
	  (sys-pathname-slaves sys-pathname-init :put put))
    (put-message-value (_ "physical-block-device-list")
		       physical-block-device-list :put put)

    ;; 7) now find the device names of the disk(s) underlying the
    ;; block device(s),
    ;;
    ;; Design note:
    ;;
    ;; Physical block devices look like:
    ;;
    ;; /sys/devices/pci0000:00/0000:00:1f.2/host0/target0:0:0/0:0:0:0/block/sda/sda6/
    ;;
    ;; So we take the item just after `block' in the directory
    ;; pathname (`sda' in this case) and format it like `/dev/sda'
    ;;
    (let* ((device-name-list
	    (mapcar #'(lambda (hdd)
			(format nil "/dev/~a"
				(second (member "block" 
						(pathname-directory hdd)
						:test #'string=))))
		    physical-block-device-list))
	   (dummy0  (debug-message-value (_ "device-name-list") device-name-list
				       :put put))
	   (write-cache-disabled-p-list
	    (mapcar #'(lambda (device-name)
			(shell-hdparm-write-cache-disabled-p device-name
							     :put put))
		    device-name-list)))
      (defparameter *system-disk-name-list* device-name-list)
      (put-message-value (_ "device-name-list") device-name-list :put put)
      (put-message-value (_ "write-cache-disabled-p-list") 
			 write-cache-disabled-p-list :put put)
      (setq write-cache-disabled-p
	    (notany #'null write-cache-disabled-p-list)))
    (put-message-value (_ "are all write caches disabled") 
		       write-cache-disabled-p :put put)
    ;;(die) ; uncomment when debug
    write-cache-disabled-p))

(defun sys-pathname-slaves (sys-pathname &key (put nil))
  "Determining `/sys' pathnames of slaves below given `/sys' pathname"
  ;;
  ;; Design note:
  ;;
  ;; A given block device may be mapped to several slaves (e.g. RAID).
  ;; This means `/sys/block/<dev>/slaves/<dev>/slaves/...'  has the
  ;; topology of a tree.  The leaves of the tree (i.e. devices with no
  ;; slaves) are the physical block devices.
  ;;
  ;; This function assumes that the given `sys-block-pathname' starts
  ;; somewhere within the `/sys' filesystem, perhaps `/sys/block' or
  ;; `/sys/dev/block'.  The function chases slaves from there by tree
  ;; search.  This tree search is done using recursion.  The function
  ;; returns a list of sys-block-pathnames to physical block devices.
  ;;
  (put-message (_ "-----sys-pathname-slaves-----")   :put put)
  (put-message-value (_ "sys-pathname") sys-pathname :put put)
  (let* ((dir-list              (pathname-directory sys-pathname))
	 (block-device          (first (last dir-list)))
	 (dir-list-slaves       (append dir-list (list "slaves")))
	 (sys-pathname-slaves   (make-pathname :directory dir-list-slaves))
	 (sys-pathname-slaves-p (directory sys-pathname-slaves))
	 (dir-list-slaves-wild  (append dir-list (list "slaves" :wild)))
	 (sys-pathname-slaves-wild
	                        (make-pathname :directory dir-list-slaves-wild))
	 (sys-pathname-slaves-found 
	                        (directory sys-pathname-slaves-wild)))
    (unless (and (eql     (first  dir-list) :absolute)
		 (string= (second dir-list) "sys"))
      (put-message-fail-and-abort (_ "invalid sys-pathname")))
    (debug-message-value   (_ "block-device")         block-device)
    (debug-message-value (_ "dir-list")               dir-list)
    (debug-message-value (_ "dir-list-slaves")        dir-list-slaves)
    (debug-message-value (_ "sys-pathname-slaves")    sys-pathname-slaves)
    (debug-message-value (_ "sys-pathname-slaves-p")  sys-pathname-slaves-p)
    (if (null sys-pathname-slaves-p)
	(progn
	  ;; no slaves directory was found
	  (put-message-value (_ "found physical device")
			     (directory sys-pathname) :put put)
	  (put-message (_ "-----sys-pathname-slaves-----done-----") :put put)
	  (directory sys-pathname))
      (progn
	(debug-message-value (_ "dir-list-slaves-wild") dir-list-slaves-wild)
	(debug-message-value (_ "sys-pathname-slaves-wild")
			     sys-pathname-slaves-wild)
	(put-message-value (_ "sys-pathname-slaves-found")
			     sys-pathname-slaves-found :put put)
	(if (null sys-pathname-slaves-found)
	    (progn
	      ;; empty slaves directory was found
	      (put-message-value (_ "found physical device") sys-pathname
				 :put put)
	      (put-message (_ "-----sys-pathname-slaves-----done-----") 
			   :put put)
	      (list sys-pathname))
	  (progn
	    ;; non-empty slaves directory was found
	    (put-message (_ "found slaves - recursing") :put put)
	    ;; gather the leaves of the tree
	    (reduce #'append
		    (mapcar #'(lambda (slave) (sys-pathname-slaves  slave
								    :put put))
			    sys-pathname-slaves-found))))))))

(defun assert-internet-access-to-wikimedia-site-p (&key (put nil))
  "Asserting internet access to wikimedia site"
  ;; Bug: actually this will not detect a failure if: 
  ;; a) your traffic is going through a caching web proxy that, 
  ;; b) has already cached a copy of the page, and 
  ;; c) then suffers a failure lying between the proxy and wikimedia
  (put-message-value (_ "fetching HTTP-header from") *wikimedia-site* :put put)
  (let ((result (shell-curl-http-head *wikimedia-site* *whereis-file-tmp-http*
				      :put put))
	(tmp-head (format nil "~a.head" *whereis-file-tmp-http*)))
    (and (null result)
	 (assert-file-exists-p tmp-head :put put))))

(defun assert-physical-memory-if-large-wikipedia-p (&key (put nil))
  "Asserting adequate physical memory"
  (let* ((phys-mem     (os:physical-memory))
	 (language-code-list (mirror-language-list :put put))
	 (large-wiki-p (and (intersection language-code-list
					  *wikimedia-large-wiki-list*
					  :test #'string=)
			   t)))
    (put-message-value (_ "detecting")
		       (format-integer-for-human phys-mem) :put put)
    ;; assert if we have one of the largest wiki's
    (put-message-value (_ "our wiki languages are") language-code-list
		       :put put)
    (put-message-value (_ "large wiki's are") *wikimedia-large-wiki-list*
		       :put put)
    (put-message-value (_ "are any of our wiki's large") large-wiki-p :put put)
    (if large-wiki-p
	(progn
	  (put-message-value (_ "requiring minimum")
			     (format-integer-for-human
			      *system-physical-memory-min*) :put put)
	  (>= phys-mem *system-physical-memory-min*))
      t)))

(defun count-cpu (&key (put nil))
  "Counting CPU's"
  ;; number of cpu's useful for determining how many mirror processes
  ;; to run concurrently
  (let* ((directory-list '(:absolute "sys" "devices" "system" "cpu"))
	 (file-name      "present")
	 (cpu-pathname   (make-pathname :directory directory-list
					:name file-name))
	 (cpu-line       nil)
	 (cpus           nil))
    ;; "present" will contain a line like "0-1" (meaning two cpus)
    (with-open-file (f cpu-pathname)
      (setq cpu-line (read-line f)))     ; "0-1"
    (with-input-from-string (s (first    ; "1"
				(last
				 (regexp:regexp-split "-" cpu-line))))
      (setq cpus (1+ (read s))))         ; 2
    (if (and (numberp cpus) (>= cpus *system-cpu-min*))
	(put-message-value-done (_ "number of CPU's found") cpus :put put)
      (progn
	(put-message-value (_ "number of CPU's found") cpus :put put)
	(setq cpus *system-cpu-min*)
	(put-message-value-done (_ "set CPU's to") cpus :put put)))
    (setq *system-cpu* cpus)
    cpus))


;;;;--------------------------------------------------------------------------+
;;;; Assert prerequisite software:                                            |
;;;;--------------------------------------------------------------------------+


(defun assert-clisp-features-p (&key (put nil))
  "Asserting clisp features"
  (assert-test-set-p
   #'(lambda (x) (member x *features*))
   '(:asdf2 :asdf :clisp :clx :common-lisp-controller :gettext
     :i18n :loop :regexp :screen :syscalls)
   :put put))

(defun assert-concurrency-limit-xchunk-p (&key (put nil))
  "Limiting concurrency of importing `x-chunk's"
  (let ((innodb-file-format (sql-show-innodb-file-format :put put)))
    (put-message-value (_ "innodb-file-format") innodb-file-format :put put)
    (setq *mirror-concurrency-limit-xchunk* 
	  (cond ((string= innodb-file-format "Antelope")
		 (put-message (_ "Antelope handles concurrency well") :put put)
		 (put-message-value (_ "choose lesser of number of CPUs or")
				    *system-cpu-max* :put put)
		 (min *system-cpu* *system-cpu-max*))
		((string= innodb-file-format "Barracuda")
		 (put-message (_ "Barracuda does not handle concurrency well - frequent deadlocks") :put put)
		 *system-cpu-min*)
		(t nil)))
    (put-message-value (_ "*mirror-concurrency-limit-xchunk*")
		       *mirror-concurrency-limit-xchunk* :put put))
  *mirror-concurrency-limit-xchunk*)

(defun assert-configuration-files-or-restore-default (&key (put nil))
  "Asserting (or restoring default) configuration files"
  (loop
   initially
     (put-message-start (_ "cd to restore directory")               :put put)
     (ext:cd *wpmirror-directory-restore*)
     (put-message-value-done (_ "cd to restore directory") (ext:cd) :put put)
   with count-mysql   = 0
   with count-restore = 0
   for (dir-symbol file-symbol) in *wpmirror-config-restore-list* 
   as dir-path     = (symbol-value dir-symbol)
   as file-name    = (symbol-value file-symbol)
   as dummy        = (put-message-value dir-path file-name :put put)
   as path-name    = (merge-pathnames (parse-namestring dir-path)
				      (parse-namestring file-name))
   ;; some file in `restore' directory may be compressed
   ;; in which case, we copy and decompress
   as file-name-gz = (format nil "~a.gz" file-name)            
   as path-name-gz = (merge-pathnames (parse-namestring dir-path)
				      (parse-namestring file-name-gz))
   when (null (file-exists-p path-name :put put))
   do (cond ((file-exists-p file-name :put put)
	     (put-flag-message-value :info (_ "restoring default") path-name)
	     (shell-copy-file file-name dir-path :put put))
	    ((file-exists-p file-name-gz :put put)
	     (put-flag-message-value :info (_ "restoring default") path-name)
	     (shell-copy-file file-name-gz dir-path :put put)
	     (ext:cd dir-path)
	     (shell-gunzip file-name :put put)
	     (ext:cd *wpmirror-directory-restore*))
	    (t
	     (put-flag-message-value :info (_ "not restoring") path-name)))
   ;; if `LocalSettings*' restored, then chown and chmod
   and do (when (and (string= dir-path *whereis-directory-mediawiki-config*)
		     (file-exists-p path-name :put put))
	    (shell-chown path-name :owner "www-data:www-data" :put put)
	    (shell-chmod path-name :permissions "600"         :put put))
   ;; if `favicon.ico' restored, then chown
   and do (when (and (string= dir-path *whereis-directory-mediawiki*)
		     (file-exists-p path-name :put put))
	    (shell-chown path-name :owner "www-data:www-data" :put put))
   ;; if `wp-mirror.cnf' restored, then see finally clause
   and do (when (string= dir-path *whereis-directory-mysql-config-conf.d*)
	    (incf count-mysql))
   and do (incf count-restore)
   finally
     (put-message-start (_ "cd to working directory")               :put put)
     (ext:cd *whereis-directory-working*)
     (put-message-value-done (_ "cd to working directory") (ext:cd) :put put)
     ;; if `wp-mirror.cnf' was restored, then restart `mysqld'
     (when (> count-mysql 0)   (shell-mysql-restart :put put))
     (when (> count-restore 0) (put-flag-message :info (_ "done"))))
  ;;
  ;; Design note:
  ;;
  ;; To get math equations formatted with `TeX', we need to create a link:
  ;;   `/var/lib/mediawiki/extensions/Math/math/texvc' -> `/usr/bin/texvc'
  ;;
  (let* ((path-name (merge-pathnames
		     (parse-namestring
		      *whereis-directory-mediawiki-extensions-math*)
		     (parse-namestring "texvc")))
	 (command-string (format nil "ln -s ~a ~a" *whereis-texvc* path-name)))
    (when (null (probe-file path-name))
      (when (null (ext:probe-directory
		   *whereis-directory-mediawiki-extensions-math*))
	(ext:make-directory *whereis-directory-mediawiki-extensions-math*))
      (shell-command command-string :put put)))
  t)


;;;;--------------------------------------------------------------------------+
;;;; Assert DBMS credentials:                                                 |
;;;;--------------------------------------------------------------------------+

(defun assert-dbms-accounts-p (&key (put nil))
  "Asserting DBMS user account `wikiadmin'"
  (let* ((select-admin (concatenate 'string
			   "SELECT host,user,password FROM mysql.user "
			   (format nil "WHERE user='~a';" *db-wikiadmin-user*)))
	 (select-user  (concatenate 'string
			   "SELECT host,user,password FROM mysql.user "
			   (format nil "WHERE user='~a';" *db-wikiuser-user*)))
	 (result-admin (query-database-and-return-list-of-lists 
			select-admin :user *db-debian-user* :put  put))
	 (result-user  (query-database-and-return-list-of-lists 
			select-user  :user *db-debian-user* :put  put)))
    (sql-select-mysql-user :put put)
    (and result-admin result-user)))

(defun create-dbms-accounts (&key (put nil))
  "Creating DBMS user accounts for `wikiadmin' `wikiuser'"
  (let ((password nil))
    ;; wikiadmin
    (setq password     (sql-encrypt-current-timestamp :put put))
    (put-message-value (_ "wikiadmin password") password :put put)
    (defparameter      *db-wikiadmin-password*  password)
    (sql-create-user   *db-wikiadmin-user* *db-wikiadmin-password* :put put)
    (sleep 1) ; otherwise same timestamp --> same passwords
    ;; wikiuser
    (setq password     (sql-encrypt-current-timestamp :put put))
    (put-message-value (_ "wikiuser password")  password :put put)
    (defparameter      *db-wikiuser-password*   password)
    (sql-create-user   *db-wikiuser-user*  *db-wikiuser-password* :put put)
    (sleep 1)))

(defun assert-dbms-accounts-or-create-p (&key (put nil))
  "Asserting or creating DBMS user accounts `wikiadmin' and `wikiuser'"
  (let ((path-name (merge-pathnames
		    (parse-namestring *whereis-directory-mediawiki-config*)
		    (parse-namestring
		     *whereis-file-mediawiki-config-localsettings-account*))))
    (if (and (assert-dbms-accounts-p  :put put)
	     (file-exists-p path-name :put put))
	(put-message (_ "DBMS accounts for 'wikiadmin' and 'wikiuser' exist") 
		     :put put)
      (progn
	(assert-dbms-drop-accounts-p                 :put put)
	(create-dbms-accounts                        :put put)
	(create-file-mediawiki-localsettings-account :put put)))
    (and (assert-dbms-accounts-p  :put put)
	 (file-exists-p path-name :put put))))

(defun assert-dbms-connect-with-credentials-debian-p (&key (put nil))
  "Asserting access to DBMS with debian credentials"
  (not (null (sql-show-databases :user *db-debian-user* :put put))))

(defun assert-dbms-connect-with-credentials-root-p (&key (put nil))
  "Asserting access to DBMS with root credentials"
  (not (null (sql-show-databases :user *db-root-user* :put put))))

(defun assert-dbms-connect-with-credentials-wikiadmin-p (&key (put nil))
  "Asserting access to DBMS with wikiadmin credentials"
  (not (null (sql-show-databases :user *db-wikiadmin-user* :put put))))

(defun assert-dbms-connect-with-credentials-wikiuser-p (&key (put nil))
  "Asserting access to DBMS with wikiuser credentials"
  (not (null (sql-show-databases :user *db-wikiuser-user* :put put))))

(defun assert-dbms-credentials-debian-p (&key (put nil))
  "Asserting DBMS credentials for `debian-sys-maint' account"
  (not (or (null *db-debian-user*)
	   (null *db-debian-password*))))

(defun scrape-dbms-credentials-debian (&key (put nil))
  "Scraping DBMS credentials for `debian-sys-maint' account"
  (let ((path-name (merge-pathnames
		    (parse-namestring *whereis-directory-mysql-config*)
		    (parse-namestring *whereis-file-mysql-config-debian*))))
    (when (assert-file-exists-p path-name :put put)
      (defparameter *db-debian-user*     
	(parse-file-key path-name "user"     :put put))
      (defparameter *db-debian-password* 
	(parse-file-key path-name "password" :put put)))))

(defun assert-dbms-credentials-debian-or-scrape-p (&key (put nil))
  "Asserting or scraping DBMS credentials for `debian-sys-maint' account"
  (when (null (assert-dbms-credentials-debian-p :put put))
    (scrape-dbms-credentials-debian :put put))
  (assert-dbms-credentials-debian-p :put put))
      
(defun assert-dbms-credentials-p (&key (put nil))
  "Asserting DBMS credentials for `wikiadmin' and `wikiuser'"
  (not (or (null *db-wikiadmin-user*)
	   (null *db-wikiadmin-password*)
	   (null *db-wikiuser-user*)
	   (null *db-wikiuser-password*))))

(defun scrape-dbms-credentials (&key (put nil))
  "Scraping DBMS credentials for `wikiuser' account"
  (let ((path-name (merge-pathnames
		(parse-namestring *whereis-directory-mediawiki-config*)
		(parse-namestring
		 *whereis-file-mediawiki-config-localsettings-account*)))
	(result    nil))
    (when (assert-file-exists-p path-name :put put)
      (setq result (parse-file-key path-name "$wgDBadminuser"     :put put))
      (when result (defparameter *db-wikiadmin-user*     result))
      (setq result (parse-file-key path-name "$wgDBadminpassword" :put put))
      (when result (defparameter *db-wikiadmin-password* result))
      (setq result (parse-file-key path-name "$wgDBuser"          :put put))
      (when result (defparameter *db-wikiuser-user*     result))
      (setq result (parse-file-key path-name "$wgDBpassword"      :put put))
      (when result (defparameter *db-wikiuser-password* result)))))

(defun assert-dbms-credentials-or-scrape-p (&key (put nil))
  "Scraping DBMS credentials for `wikiadmin' and `wikiuser' accounts"
  (when (null (assert-dbms-credentials-p :put put))
    (scrape-dbms-credentials :put put))
  (assert-dbms-credentials-p :put put))

(defun assert-dbms-drop-accounts-p (&key (put nil))
  "Dropping DBMS `wikiadmin' and `wikiuser' accounts"
  (let* ((accounts-p       (assert-dbms-accounts-p :put put))
	 (grant-admin-list (when accounts-p
			     (sql-show-grants *db-wikiadmin-user* :put put)))
	 (grant-user-list  (when accounts-p
			     (sql-show-grants *db-wikiuser-user*  :put put))))
    (when accounts-p
      (dolist (grant grant-admin-list nil)
	(put-message grant :put put))
      (dolist (grant grant-user-list  nil)
	(put-message grant :put put))
      (when (not (null grant-admin-list))
	(sql-drop-user *db-wikiadmin-user* :put put))
      (when (not (null grant-user-list))
	(sql-drop-user *db-wikiuser-user* :put put)))
  (null (assert-dbms-accounts-p :put put))))

(defun assert-dbms-grant-for-wikiadmin-p (&key (put nil))
  "Granting DBMS privileges on databases to `wikiadmin'"
  (sql-grant-global-to-wikiadmin                             :put put)
  (sql-grant-all-on-database-to-wikiadmin *db-name*          :put put)
  (sql-grant-all-on-database-to-wikiadmin *db-wpmirror-name* :put put)
  t)

(defun assert-dbms-grant-for-wikiuser-p (&key (put nil))
  "Granting DBMS privileges on databases to `wikiuser'"
  (sql-grant-siud-on-database-to-wikiuser *db-name* :put put)
  (sql-grant-siud-on-database-to-wikiuser *db-wpmirror-name* :put put)
  t)

(defun assert-dbms-mysql-p (&key (put nil))
  "Asserting DBMS MySQL"
  (string= *db-type* "mysql"))

(defun assert-dbms-mysql-config-debian-p (&key (put nil))
  "Asserting mysql debian.cnf"
  (let ((path-my (merge-pathnames
		  (parse-namestring *whereis-directory-mysql-config*)
		  (parse-namestring *whereis-file-mysql-config-debian*))))
    (assert-file-exists-p path-my :put put)))

(defun assert-dbms-time-zone-name-p (&key (time-zone "UTC") (put nil))
  "Asserting DBMS time zone tables have a given timezone"
  (let ((message (_ "looking for time-zone"))
	(result  nil))
    (put-message-value message time-zone :put put)
    (setq result (sql-select-time-zone-name time-zone :put put))
    (if (null result)
	(put-message-value message time-zone)
      (put-message-value-done message time-zone :put put))
    result))

(defun assert-dbms-time-zone-or-load (&key (put nil))
  "Asserting or loading time zone data in DBMS"
  (let ((found-p (assert-dbms-time-zone-name-p :put put))
	(message (_ "loading time zone info")))
    (when (null found-p)
      (put-message-start message :put put)
      (shell-mysql-tzinfo-to-sql-load-database :put put)
      (sleep-until 2 #'assert-dbms-time-zone-name-p :put put)
      (put-flag-message :info (_ "loaded time-zone tables"))
      (put-flag-message :done (_ "done")))
    (or found-p (assert-dbms-time-zone-name-p :put put))))

(defun create-file-mediawiki-localsettings-account (&key (put nil))
  "Creating /etc/mediawiki/LocalSettings_account.php'"
  (let ((path-name (merge-pathnames
   (parse-namestring *whereis-directory-mediawiki-config*)
   (parse-namestring *whereis-file-mediawiki-config-localsettings-account*)))
	(file-permissions     "600"))
    (put-message-value (_ "creating file") path-name :put put)
    (with-open-file (s path-name 
		       :direction :output
		       :if-exists :supersede)
		    (format s "<?php~%")
		    (format s "# Automatically generated by WP-MIRROR for MediaWiki scripts. DO NOT TOUCH!~%")
		    (format s "$wgDBadminuser      = '~a';~%"
			    *db-wikiadmin-user*)
		    (format s "$wgDBadminpassword  = '~a';~%"
			    *db-wikiadmin-password*)
		    (format s "$wgDBuser           = '~a';~%"
			    *db-wikiuser-user*)
		    (format s "$wgDBpassword       = '~a';~%"
			    *db-wikiuser-password*))
    ;; chown www-data:wwwdata
    (shell-chown path-name :put put)
    ;; chmod 600
    (shell-chmod path-name :permissions file-permissions :put put)))

(defun assert-mediawiki-favicon-p (&key (put nil))
  "Asserting mediawiki `favicon.ico' settings"
  (let* ((path-name (merge-pathnames
	     (parse-namestring *whereis-directory-mediawiki*)
	     (parse-namestring *whereis-file-mediawiki-favicon*)))
	 (result    (assert-file-exists-p path-name :put put)))
    (unless (null result)
      ;; chown www-data:wwwdata
      (shell-chown path-name :put put))
    result))

(defun assert-mediawiki-logo-p (&key (put nil))
  "Asserting mediawiki `wp-mirror.png' settings"
  (let* ((path-name (merge-pathnames
	     (parse-namestring *whereis-directory-mediawiki*)
	     (parse-namestring *whereis-file-mediawiki-logo*)))
	 (result    (assert-file-exists-p path-name :put put)))
    (unless (null result)
      ;; chown www-data:wwwdata
      (shell-chown path-name :put put))
    result))

(defun assert-mediawiki-localsettings-p (&key (put nil))
  "Asserting mediawiki local settings"
  (let* ((path-name (merge-pathnames
	     (parse-namestring *whereis-directory-mediawiki-config*)
	     (parse-namestring *whereis-file-mediawiki-config-localsettings*)))
	 (result    (assert-file-exists-p path-name :put put))
	 (file-permissions     "600"))
    (unless (null result)
      ;; chown www-data:wwwdata
      (shell-chown path-name :put put)
      ;; chmod 600
      (shell-chmod path-name :permissions file-permissions :put put))
    result))

(defun assert-mediawiki-localsettings-account-p (&key (put nil))
  "Asserting mediawiki local settings for database accounts"
  (let* ((path-name (merge-pathnames
		(parse-namestring *whereis-directory-mediawiki-config*)
		(parse-namestring
		 *whereis-file-mediawiki-config-localsettings-account*)))
	 (result    (assert-file-exists-p path-name :put put))
	 (file-permissions     "600"))
    (unless (null result)
      ;; chown www-data:wwwdata
      (shell-chown path-name :put put)
      ;; chmod 600
      (shell-chmod path-name :permissions file-permissions :put put))
    result))

(defun assert-mediawiki-localsettings-wpmirror-p (&key (put nil))
  "Asserting mediawiki local settings for wp-mirror"
  (let* ((path-name (merge-pathnames
		(parse-namestring *whereis-directory-mediawiki-config*)
		(parse-namestring
		 *whereis-file-mediawiki-config-localsettings-wpmirror*)))
	 (result    (assert-file-exists-p path-name :put put))
	 (file-permissions     "600"))
    (unless (null result)
      ;; chown www-data:wwwdata
      (shell-chown path-name :put put)
      ;; chmod 600
      (shell-chmod path-name :permissions file-permissions :put put))
    result))

(defun parse-file-mediawiki-config-localsettings (key &key (put nil))
  "Parsing mediawiki local settings for `key = value'"
  (let* ((path-name-1 (merge-pathnames
		       (parse-namestring *whereis-directory-mediawiki-config*)
		       (parse-namestring
			*whereis-file-mediawiki-config-localsettings*)))
	 (path-name-2 (merge-pathnames
		       (parse-namestring *whereis-directory-mediawiki-config*)
		       (parse-namestring
		       *whereis-file-mediawiki-config-localsettings-account*)))
	 (path-name-3 (merge-pathnames
		       (parse-namestring *whereis-directory-mediawiki-config*)
		       (parse-namestring
		       *whereis-file-mediawiki-config-localsettings-wpmirror*)))
	(path-name-list (list path-name-1 path-name-2 path-name-3))
	(temp           nil)
	(result         nil))
    (dolist (path-name path-name-list result)
	(setq temp (parse-file-key path-name key :put put))
	(when temp (setq result temp)))))

(defun parse-file-mysql-config-debian (key &key (put nil))
  "Parsing '/etc/mysql/debian.cnf' for `key = value'"
  (let ((path-name (merge-pathnames
		    (parse-namestring *whereis-directory-mysql-config*)
		    (parse-namestring *whereis-file-mysql-config-debian*))))
    (parse-file-key path-name key :put put)))

(defun parse-file-key (path-name key &key (put nil))
  "Parsing path-name for `key = value'"
  (let* ((l        (grep-last-non-comment path-name key))
	 (m        (format nil "probing ~a" key))
	 (rhs      nil)
	 (val-type nil)
	 (val      nil))
    (if (null l)
	(put-message m)
      (progn
	(setq rhs (second (regexp:regexp-split "=" l))) ; right-hand side
	;; rhs can be `"foo";' (where foo is a string)
	;; rhs can be  `foo;'  (where foo is `true', `false', numeric)
	(if (regexp:match "['\"]" rhs)
	    (setq val-type :string)
	  (setq val-type :other))
	(case val-type
	      (:string (setq val (second (regexp:regexp-split "['\"]" rhs))))
	      (:other  (setq val (string-trim
				  `(#\Space #\Tab #\Newline)
				  (first  (regexp:regexp-split ";"  rhs))))))
	(if (regexp:match "password" key)
	    (put-message-value m "********" :put put)
	  (put-message-value m val :put put))))
    val))

(defun put-dbms-credentials (&key (put nil))
  "Putting DBMS credentials"
  (put-message-value "*db-type*"   *db-type*     :put put)
  (put-message-value "*db-server*" *db-server*   :put put)
  (put-message-value "*db-name*"   *db-name*     :put put)
  (put-message-value "*db-debian-user*"
		     *db-debian-user*    :put put)
  (put-message-value "*db-debian-password*"
		     (if *db-debian-password*    "********" nil) :put put)
  (put-message-value "*db-root-user*"
		     *db-root-user*      :put put)
  (put-message-value "*db-root-password*"
		     (if *db-root-password*      "********" nil) :put put)
  (put-message-value "*db-wikiadmin-user*"
		     *db-wikiadmin-user* :put put)
  (put-message-value "*db-wikiadmin-password*"
		     (if *db-wikiadmin-password* "********" nil) :put put)
  (put-message-value "*db-wikiuser-user*"
		     *db-wikiuser-user*  :put put)
  (put-message-value "*db-wikiuser-password*"
		     (if *db-wikiuser-password*  "********" nil) :put put))

(defun warn-if-dbms-root-account-has-no-password (&key (put nil))
  "Warning if DBMS root account has no password"
  (let* ((select (concatenate 'string
			      "SELECT host,user,password FROM mysql.user "
			      "WHERE user='root' AND password='';"))
	 (result (query-database-and-return-list-of-lists select
					    :user *db-debian-user*
					    :put  put)))
    (debug-message result)
    (null result)))

(defun warn-if-dbms-has-anonymous-user-account (&key (put nil))
  "Warning if DBMS has any anonymous user accounts"
  (let* ((select "SELECT host,user,password FROM mysql.user WHERE user='';")
	 (result (query-database-and-return-list-of-lists select
					    :user *db-debian-user*
					    :put  put)))
    (debug-message result)
    (null result)))

(defun warn-if-dbms-has-root-accounts-accessible-from-outside-localhost (&key (put nil))
  "Warning if DBMS has root accounts accessible from outside localhost"
  (let* ((select (concatenate 'string
			      "SELECT host,user,password FROM mysql.user "
			      "WHERE user='root' AND host NOT IN "
			      "('localhost', '127.0.0.1', '::1');"))
	 (result (query-database-and-return-list-of-lists select
					    :user *db-debian-user*
					    :put  put)))
    (debug-message result)
    (null result)))

(defun warn-if-dbms-has-test-database (&key (put nil))
  "Warning if DBMS has test database"
  (let* ((show   "SHOW DATABASES LIKE 'test%';")
	 (result (query-database-and-return-list-of-strings show
					    :user *db-debian-user*
					    :put  put)))
    (debug-message-value show result)
    (null result)))


;;;;--------------------------------------------------------------------------+
;;;; Assert more prerequisite software:                                       |
;;;;--------------------------------------------------------------------------+

(defun assert-database-p (database-name &key (put nil))
  "Asserting database-name exists"
  (let ((database-list (sql-show-databases :user *db-debian-user* :put put)))
    (put-message-value (_ "database list") database-list :put put)
    (dolist (l database-list nil)
      (when (string= database-name l)
	(put-message-value-done (_ "found database") database-name :put put)
	(return t))))) ; break do loop

(defun assert-database-dump-p (database-name &key (put nil))
  "Asserting database dump file"
  (let* ((message      (_ "database dump-file"))
	 (file-name    (database-name-to-database-dump-file-name database-name
								 :put put))
	 (path-name    (merge-pathnames 
			(parse-namestring *whereis-directory-working*)
			(parse-namestring file-name)))
	 (result       (file-exists-p path-name :put put)))
    (if result
	(put-message-value-done message database-name :put put)
      (put-message-value-fail message database-name :put put))
    result))

(defun assert-database-tables-p (database-name &key (put nil))
  "Asserting database exists and has at least one table"
  (let* ((exist-p (assert-database-p database-name :put put))
	 (count   0))
    (if (null exist-p)
	(put-message-value-fail (_ "database not found") database-name :put put)
      (setq count (sql-select-count-tables database-name :put put)))
    (put-message-value (_ "number of tables") count :put put)
    (> count 0)))

(defun assert-database-template-p (&key (put nil))
  "Asserting database template for `wikidb' and `xxwiki'"
  (let ((template-path  (merge-pathnames
			 (parse-namestring
			  *whereis-directory-mediawiki-maintenance*)
			 (parse-namestring
			  *whereis-file-mediawiki-farm-database*))))
    (file-exists-p template-path :put put)))

(defun assert-database-template-or-dump (&key (put nil))
  "Asserting or creating a `mysqldump' of `wikidb'"
  (let* ((message-found  (_ "database template found"))
	 (message-dump   (_ "dumping database to make template"))
	 (file-path      (merge-pathnames
			  (parse-namestring
			   *whereis-directory-mediawiki-maintenance*)
			  (parse-namestring
			   *whereis-file-mediawiki-farm-database*)))
	 (result         nil))
    (if (file-exists-p file-path :put put)
	(put-message message-found :put put)
      (progn
	(put-message-start message-dump :put put)
	(setq result (shell-mysqldump-to-file *db-name* file-path :put put))
	(if (null result)
	    (put-message-done message-dump :put put)
	  (put-message-fail-and-abort message-dump))))
    (file-exists-p file-path :put put)))

(defun assert-database-wikidb-p (&key (put nil))
  "Asserting database `wikidb'"
  (assert-database-p *db-name* :put put))

(defun assert-database-wikidb-or-load (&key (put nil))
  "Asserting or loading `wikidb' database"
  (let ((template-path (merge-pathnames
		   (parse-namestring *whereis-directory-mediawiki-maintenance*)
		   (parse-namestring *whereis-file-mediawiki-farm-database*))))
    (put-message template-path :put put)
    (assert-database-or-load *db-name* template-path :put put)))

(defun assert-database-template-and-wikidb-p (&key (put nil))
  "Asserting database template and `wikidb' or die"
  (let ((template-p  (assert-database-template-p :put put))
	(wikidb-p    (assert-database-wikidb-p :put put))
	(result      nil))
    (put-message-value "template-p" template-p :put put)
    (put-message-value "wikidb-p"   wikidb-p   :put put)
    ;; missing wikidb, load wikidb from template
    (when (and template-p (null wikidb-p))
      (assert-database-wikidb-or-load :put put)
      (setq result t))
    ;; missing template, dump wikidb to template
    (when (and (null template-p) wikidb-p)
      (assert-database-template-or-dump :put put)
      (setq result t))
    ;; have both, do nothing
    (when (and template-p wikidb-p)
      (setq result t))
    result))

(defun assert-database-wpmirror-p (&key (put t))
  "Asserting database `wpmirror'"
  (assert-database-p *db-wpmirror-name* :put put))

(defun assert-database-wpmirror-or-create-p (&key (put nil))
  "Asserting database `wpmirror' or create it"
  ;; assert or create database
  (if (assert-database-wpmirror-p :put put)
      (put-message-value (_ "database found") *db-wpmirror-name* :put put)
    (progn
      ;; grant
      (put-message-start (_ "granting privileges to *db-wikiuser-user*") :put put)
      (sql-grant-siud-on-database-to-wikiuser *db-wpmirror-name* :put put)
      (put-message-value-done (_ "granting privileges to *db-wikiuser-user*")
			      *db-wikiuser-user* :put put)
      ;; database
      (put-message-start (_ "creating *db-wpmirror-name*") :put put)
      (sql-create-database *db-wpmirror-name* :put put)
      (put-message-value-done (_ "creating *db-wpmirror-name*") 
			      *db-wpmirror-name* :put put)))
  ;; assert or create tables
  (let ((tables-found (sql-select-count-tables *db-wpmirror-name* :put put))
	(tables-expected *db-wpmirror-tables*))
    (if (= tables-found tables-expected)
	(progn
	  (put-message-value (_ "tables found") tables-found :put put)
	  ;; table `file'
	  (put-message-start (_ "updating schema table `file'") :put put)
	  (sql-alter-table-file :put put)
	  (put-message-value-done (_ "updating schema table") "file" :put put))
      (progn
	;; table `file'
	(put-message-start (_ "creating table `file'") :put put)
	(sql-create-table-file :put put)
	(put-message-value-done (_ "creating table") "file" :put put)
	;; table `image'
	(put-message-start (_ "creating table `image'") :put put)
	(sql-create-table-image :put put)
	(put-message-value-done (_ "creating table") "image" :put put))))
  (sleep 1)
  (assert-database-wpmirror-table-p :put put))

(defun assert-database-wpmirror-table-p (&key (put t))
  "Asserting database `wpmirror' has a table"
  (let* ((result (sql-select-count-tables *db-wpmirror-name* :put put)))
    (put-message-value (_ "number of tables") result :put put)
    (= result *db-wpmirror-tables*)))

(defun assert-database-or-load (database-name template-path &key (put nil))
  "Asserting or loading a database from a dump file"
  ;; assert or create database
  (if (assert-database-p database-name :put put)
      (put-message-value (_ "database found") database-name :put put)
    (progn
      ;; grant
      (put-message-value (_ "granting privileges for") database-name :put put)
      (sql-grant-siud-on-database-to-wikiuser database-name :put put)
      (put-message-value-done (_ "granting privileges to") *db-wikiuser-user*
			      :put put)
      ;; database
      (put-message-value (_ "creating database") database-name :put put)
      (sql-create-database database-name :put put)
      (put-message-value-done (_ "creating database") database-name :put put)))
  ;; assert or create tables
  (let ((tables-found (sql-select-count-tables database-name :put put)))
    (if (> tables-found 0)
	(put-message-value (_ "tables found") tables-found :put put)
      (progn
	;; tables 
	(put-message-start (_ "loading database template") :put put)
	(if (file-exists-p template-path :put put)
	    (progn
	      (put-message-value (_ "database template found") template-path
				 :put put)
	      (shell-mysql-load-database database-name template-path :put put)
	      (sleep 1)
	      (put-message-value-done (_ "loading database template for")
				      database-name :put put))
	  (put-message-value-fail (_ "database template not found") 
				  template-path)))))
  ;; assert tables
  (assert-database-tables-p database-name :put put))

(defun assert-directory-exists-or-create-p (directory-name &key (put nil))
  "Asserting directory exists, or create it"
  (put-message-value (_ "probing directory") directory-name :put put)
  (or (ext:probe-pathname directory-name)
      (progn
	(put-message-start (_ "creating directory") :put put)
	(ext:make-directory directory-name)
	(ext:probe-pathname directory-name))))

(defun assert-directory-exists-p (directory-name &key (put nil))
  "Asserting directory exists or die"
  (put-message-value (_ "probing directory") directory-name)
  (ext:probe-pathname directory-name))

(defun file-exists-p (file-name &key (put nil))
  "Determining if file exists and not empty."
  (and (ext:probe-pathname file-name)
       (let ((fsize (posix:file-size file-name)))
	 (put-message-value (_ "File size")
			    (format-integer-for-human fsize) :put put)
	 (> fsize 0))))

(defun assert-file-exists-or-create-p (file-name printable-object &key (put t))
  "Asserting file exists, or create it"
  (put-message-value (_ "probing file") file-name :put put)
  (or (ext:probe-pathname file-name)
    (progn
      (put-message-value (_ "creating file") file-name :put put)
      (with-open-file (f file-name
			 :direction :output
			 :if-exists :supersede)
	(prin1 printable-object f))
      (ext:probe-pathname file-name))))

(defun assert-file-exists-p (file-name &key (put t))
  "Asserting file exists and not empty, or die."
  (put-message-value (_ "probing") file-name :put put)
  (file-exists-p file-name :put put))

(defun assert-images-directory-or-create-p (&key (put t))
  "Asserting or creating mediawiki images directory tree (../images/0/00/)"
  (assert-images-directory-tree-or-create-p *whereis-directory-images*
					    :put put))

(defun assert-images-bad-directory-or-create-p (&key (put t))
  "Asserting or creating a bad-images directory tree (../images/bad-images/0/00/)"
  (assert-images-directory-tree-or-create-p *whereis-directory-images-bad*
					    :put put))

(defun assert-images-math-directory-or-create-p (&key (put t))
  "Asserting or creating a math directory tree (../images/math/0/0/0/)"
  (let ((path-name *whereis-directory-images-math*))
    (when (not (ext:probe-directory path-name))
      (put-message (_ "directory tree not found---creating") :put put)
      (ext:make-directory path-name)
      (shell-chown path-name :put put))
    (loop 
      with hex = (the simple-string "0123456789abcdef") 
      for c1 of-type standard-char across hex
      as  d1 of-type pathname = (make-pathname
				 :directory (append
					     (pathname-directory path-name) 
					     (list (format nil "~a" c1))))
      do
        (when (not (ext:probe-directory d1))
	  (ext:make-directory d1)
	  (shell-chown d1 :put put)
	  (loop
	    for c2 of-type standard-char across hex
	    as  d2 of-type pathname = (make-pathname
			         :directory (append
					     (pathname-directory d1)
					     (list (format nil "~a" c2))))
	    do
	      (when (not (ext:probe-directory d2))
		(ext:make-directory d2)
		(shell-chown d2 :put put)
		(loop
		   for c3 of-type standard-char across hex
		   as  d3 of-type pathname = (make-pathname
				:directory (append
					    (pathname-directory d2)
					    (list (format nil "~a" c3))))
		   do
		   (when (not (ext:probe-directory d3))
		     (ext:make-directory d3)
		     (shell-chown d3 :put put)))))))
    (ext:probe-directory path-name)))

(defun assert-images-thumb-directory-or-create-p (&key (put t))
  "Asserting or creating a thumb directory tree (../images/thumb/0/00/"
  (assert-images-directory-tree-or-create-p *whereis-directory-images-thumb*
					    :put put))

(defun assert-images-tmp-directory-or-create-p (&key (put t))
  "Asserting or creating a tmp (../images/tmp/"
  (let ((path-name *whereis-directory-images-tmp*))
    (assert-directory-exists-or-create-p path-name :put put)
    (shell-chown path-name :put put)
    (ext:probe-directory path-name)))

(defun assert-images-directory-tree-or-create-p (path-name &key (put t))
  "Asserting or creating a sub directory (bad-images, thumb) under images"
  (let ((path-name (parse-namestring path-name)))
    (when (not (ext:probe-directory path-name))
      (put-message (_ "directory tree not found---creating") :put put)
      (ext:make-directory path-name)
      (shell-chown path-name :put put))
    (loop 
      with hex = (the simple-string "0123456789abcdef") 
      for c1 of-type standard-char across hex
      as  d1 of-type pathname = (make-pathname
				 :directory (append
					     (pathname-directory path-name) 
					     (list (format nil "~a" c1))))
      do
        (when (not (ext:probe-directory d1))
	  (ext:make-directory d1)
	  (shell-chown d1 :put put)
	  (loop
	    for c2 of-type standard-char across hex
	    as  d2 of-type pathname = (make-pathname
			         :directory (append
					     (pathname-directory d1)
					     (list (format nil "~a~a" c1 c2))))
	    do
	      (when (not (ext:probe-directory d2))
		(ext:make-directory d2)
		(shell-chown d2 :put put)))))
    (ext:probe-directory path-name)))

(defun assert-mediawiki-dbms-credentials-p (&key (put nil))
  "Asserting mediawiki has DBMS credentials"
  (defparameter *db-type*               
    (parse-file-mediawiki-config-localsettings "$wgDBtype"          :put put))
  (defparameter *db-server* 
    (parse-file-mediawiki-config-localsettings "$wgDBserver"        :put put))
  (defparameter *db-name*
    (parse-file-mediawiki-config-localsettings "$wgDBname"          :put put))
  (defparameter *db-wikiadmin-user*
    (parse-file-mediawiki-config-localsettings "$wgDBadminuser"     :put put))
  (defparameter *db-wikiadmin-password* 
    (parse-file-mediawiki-config-localsettings "$wgDBadminpassword" :put put))
  (defparameter *db-wikiuser-user* 
    (parse-file-mediawiki-config-localsettings "$wgDBuser"          :put put))
  (defparameter *db-wikiuser-password*  
    (parse-file-mediawiki-config-localsettings "$wgDBpassword"      :put put))
  (notany #'null (list *db-type* *db-server* *db-name*
		       *db-wikiadmin-user* *db-wikiadmin-password*
		       *db-wikiuser-user*  *db-wikiuser-password*)))

(defun assert-mediawiki-localsettings-image-p (&key (put nil))
  "Asserting mediawiki localsettings for image file conversion"
  (let ((val-imagemagick (parse-file-mediawiki-config-localsettings
			  "$wgUseImageMagick"       :put put))
	(val-custom      (parse-file-mediawiki-config-localsettings
			  "$wgCustomConvertCommand" :put put))
	(val-svg         (parse-file-mediawiki-config-localsettings
			  "$wgSVGConverter"         :put put)))
    (and 
        val-imagemagick
	val-custom
	val-svg
	(regexp:match *mediawiki-wgUseImageMagick*       val-imagemagick)
	(regexp:match *mediawiki-wgCustomConvertCommand* val-custom)
	(regexp:match *mediawiki-wgSVGConverter*         val-svg))))

(defun assert-mediawiki-localsettings-tidy-p (&key (put nil))
  "Asserting mediawiki localsettings for `tidy'"
  (let ((val (parse-file-mediawiki-config-localsettings
	      "$wgUseTidy" :put put)))
    (and val
	 (regexp:match "true" val))))

(defun assert-test-set-p (test set &key (put nil))
  "Asserting all elements in set pass test"
  (let ((all-pass t))
    (dolist (element set all-pass)
      (let ((result (funcall test element)))
	(setq all-pass (and result all-pass))
	(if result
	    (put-symbol-value-pass element :put put)
	  (put-symbol-value-fail element))))))

(defun assert-utilities-p (&key (put nil))
  "Asserting prerequisite utilities"
  (assert-test-set-p
   #'(lambda (x) (ext:probe-pathname (symbol-value x)))
   '(*whereis-bunzip2* *whereis-cat* *whereis-chown* *whereis-chmod*
                       *whereis-cp* *whereis-cron* *whereis-curl*
		       *whereis-env*
		       *whereis-gm* *whereis-grep* *whereis-gunzip*
		       *whereis-hdparm*
		       *whereis-md5sum*
		       *whereis-mv* *whereis-mysql* *whereis-mysqldump*
		       *whereis-openssl*
		       *whereis-php*
                       *whereis-rm* *whereis-rsvg*
		       *whereis-texvc*
		       *whereis-wget*)
   :put put)
  (when *mirror-image-use-wikix-p*
    (assert-test-set-p
     #'(lambda (x) (ext:probe-pathname (symbol-value x)))
     '(*whereis-wikix*)
     :put put))
  (assert-test-set-p
   #'(lambda (x) (ext:probe-pathname
		  (merge-pathnames
		   (parse-namestring *whereis-directory-mediawiki-config*)
		   (parse-namestring (symbol-value x)))))
   '(*whereis-file-mediawiki-config-localsettings*
     *whereis-file-mediawiki-config-localsettings-account*
     *whereis-file-mediawiki-config-localsettings-wpmirror*)
   :put put)
  (assert-test-set-p
   #'(lambda (x) (ext:probe-pathname
		  (merge-pathnames
		   (parse-namestring *whereis-directory-mediawiki-maintenance*)
		   (parse-namestring (symbol-value x)))))
   '(*whereis-file-mediawiki-farm-importdump*
     *whereis-file-mediawiki-farm-rebuildimages*
     *whereis-file-mediawiki-farm-update*
     *whereis-file-mediawiki-importdump*
     *whereis-file-mediawiki-rebuildimages*
     *whereis-file-mediawiki-update*)
   :put put))

(defun assert-virtual-host-p (&key (put nil))
  "Asserting virtual host `wpmirror.site'"
  (and
   (assert-file-exists-p
    (merge-pathnames
     (parse-namestring *whereis-directory-apache-sites-available*)
     (parse-namestring *whereis-file-virtual-host*))
    :put put)
   (assert-file-exists-p
    (merge-pathnames
     (parse-namestring *whereis-directory-apache-sites-enabled*)
     (parse-namestring *whereis-file-virtual-host*))
    :put put)
   (shell-apache2ctl-grep-vhost-p :put put)))

(defun assert-virtual-host-name-resolution-p (&key (put nil))
  "Asserting name resolution for `wpmirror.site'"
  (let ((lines   (shell-cat-grep *whereis-file-etc-hosts* *virtual-host-name*
				 :put put))
	(missing nil)
	(results nil))
    ;; 1) determining which virtual-hosts we can resolve
    (put-message-start (format nil "grepping ~a" *whereis-file-etc-hosts*)
		       :put put)
    (loop
      for line in lines
      do (put-message line :put put))
    (put-message-done (format nil "grepping ~a" *whereis-file-etc-hosts*) 
		      :put put)
    ;; 2) determining which virtual-hosts we need
    (put-message-start (_ "matching languages") :put put)
    (setq missing
	  (loop
	    with languages = (append *mirror-languages* (list "www"))
	    for lang    in languages
	    as  message =  (format nil "~a.~a" lang *virtual-host-name*)
	    as  result  =  (loop
			     for line in lines
			     thereis (regexp:match lang line))
	    do (if result
		   (put-message-value (_ "found") message :put put)
		 (put-message-value (_ "need") message :put put))
	    when (null result) collect lang))
    (put-message-done (_ "matching languages") :put put)
    ;; 3) autoconfiguring any virtual-hosts that are missing
    (unless (null missing)
      (put-message-start (format nil "appending virtual host names to ~a" 
				 *whereis-file-etc-hosts*) :put put)
      (with-open-file (s *whereis-file-etc-hosts*
			 :direction :output
			 :if-exists :append
			 :if-does-not-exist :error)
	(loop
	  for lang in missing
	  as  line =  (format nil "::1 ~a.~a" lang *virtual-host-name*)
	  do (format s "~a~%" line)
	  (put-message line :put put)
	  (put-flag-message :info
			    (format nil "appending virtual host name '~a' to ~a"
				    line *whereis-file-etc-hosts*))))
      (put-flag-message :done (_ "done")))
    t))

(defun assert-working-directory-or-create-p (&key (put t))
  "Asserting working directory, or create it"
  (when (null (ext:probe-directory *whereis-directory-working*))
    (put-message (_ "working directory not found---creating") :put put)
    (ext:make-directory *whereis-directory-working*))
  (ext:cd *whereis-directory-working*)
  (ext:probe-directory *whereis-directory-working*))

(defun release-all-file-semaphores (&key (put nil))
  "Releasing all file semaphores"
  (put-message-start (_ "releasing all file semaphores"))
  (sql-update-file-semaphore-all 1 :put nil)
  (put-message-done (_ "releasing all file semaphores"))
  t)

(defun initialize-system (&key (put nil))
  "Initializing system"
  ;;(log-test-set '(process-command-line-arguments-or-die)            :put put)
  (process-command-line-arguments-or-die            :put put)
  ;; this message comes after, to avoid cluttering `--help' and `--version'
  (put-message (_ "-----initializing-begin-----------------------") :put t)
  ;; set *main-mode*
  (set-main-mode-of-operation :put put)
  (case *main-mode*
	(:add
	 (log-test-set        '(log-start)                            :put put)
	 (log-test-set-or-die '(assert-clisp-features-p
				assert-utilities-p)                   :put put)
	 (log-test-set-or-die '(assert-images-directory-or-create-p
				assert-working-directory-or-create-p)
			      :put put)
	 (log-test-set-or-die '(assert-dbms-mysql-p
				assert-dbms-mysql-config-debian-p)    :put put)
	 (log-test-set-or-die '(assert-dbms-credentials-debian-or-scrape-p
				assert-dbms-connect-with-credentials-debian-p)
			      :put put)
	 (log-test-set-or-die '(assert-dbms-time-zone-or-load)        :put put)
	 )
	((:delete :drop :dump :restore-default :update)
	 (log-test-set        '(log-start)                            :put put)
	 (log-test-set-or-die '(assert-dbms-mysql-p
				assert-dbms-mysql-config-debian-p)    :put put)
	 (log-test-set-or-die '(assert-dbms-credentials-debian-or-scrape-p
				assert-dbms-connect-with-credentials-debian-p)
			      :put put)
	 )
	((:first-mirror)
	 (log-test-set        '(log-start)                            :put put)
	 (log-test-set-or-die '(assert-clisp-features-p
				assert-utilities-p)                   :put put)
	 (log-test-set-or-die '(assert-images-directory-or-create-p
				assert-images-bad-directory-or-create-p
				assert-images-math-directory-or-create-p
				assert-images-thumb-directory-or-create-p
				assert-images-tmp-directory-or-create-p
				assert-working-directory-or-create-p) :put put)
	 (log-test-set-or-die '(assert-dbms-mysql-p
				assert-dbms-mysql-config-debian-p)    :put put)
	 (log-test-set-or-die '(assert-dbms-credentials-debian-or-scrape-p
				assert-dbms-connect-with-credentials-debian-p)
			      :put put)
	 (log-test-set-or-die '(assert-dbms-time-zone-or-load)        :put put)
	 (log-test-set-or-die '(assert-configuration-files-or-restore-default
				process-configuration-files-or-die
				put-parameters)
			      :put put)
	 )
	((:next-mirror)
	 (log-test-set        '(log-start)                            :put put)
	 (log-test-set-or-die '(assert-working-directory-or-create-p) :put put)
	 (log-test-set-or-die '(assert-dbms-credentials-debian-or-scrape-p
				assert-dbms-connect-with-credentials-debian-p)
			      :put put)
	 (log-test-set-or-die '(process-configuration-files-or-die)
			      :put put)
	 )
	((:monitor)
	 (log-test-set-or-die '(assert-clisp-features-p
				assert-utilities-p)                   :put put)
	 (log-test-set-or-die '(assert-images-directory-or-create-p
				assert-working-directory-or-create-p)
			      :put put)
	 (log-test-set-or-die '(assert-dbms-mysql-p
				assert-dbms-mysql-config-debian-p)    :put put)
	 (log-test-set-or-die '(assert-dbms-credentials-debian-or-scrape-p
				assert-dbms-connect-with-credentials-debian-p)
			      :put put)
	 (log-test-set-or-die '(assert-dbms-time-zone-or-load)        :put put)
	 (log-test-set-or-die '(assert-configuration-files-or-restore-default
				process-configuration-files-or-die)
			      :put put)
	 )
	(otherwise nil))
  (put-message (_ "-----initializing-done------------------------") :put put)
  t)

(defun assert-prerequisite-software (&key (put nil))
  "Asserting prerequisite software"
  (put-message (_ "-----asserting-prerequisite-software-begin----") :put t)
  (log-test-set-or-die '(assert-dbms-accounts-or-create-p
			 assert-dbms-credentials-or-scrape-p
			 assert-dbms-connect-with-credentials-wikiadmin-p
			 assert-dbms-grant-for-wikiadmin-p)
			 :put put)
  (case *main-mode*
	((:delete :drop :dump :restore-default :update)
	 )
	((:first-mirror)
	 (log-test-set-or-die '(assert-dbms-connect-with-credentials-wikiuser-p
				assert-dbms-grant-for-wikiuser-p)      :put put)
	 ;; warn if dbms insecure
	 (log-test-set        '(warn-if-dbms-root-account-has-no-password 
				warn-if-dbms-has-anonymous-user-account 
				warn-if-dbms-has-root-accounts-accessible-from-outside-localhost
				warn-if-dbms-has-test-database)        :put put)
	 (log-test-set-or-die '(assert-database-wpmirror-or-create-p
				assert-database-template-and-wikidb-p) :put put)
	 (log-test-set-or-die '(assert-mediawiki-localsettings-p
				assert-mediawiki-localsettings-account-p
				assert-mediawiki-localsettings-wpmirror-p)
			      :put put)
	 (log-test-set-or-die '(assert-mediawiki-localsettings-image-p
				assert-mediawiki-localsettings-tidy-p) :put put)
	 (log-test-set-or-die '(assert-mediawiki-favicon-p
				assert-mediawiki-logo-p)               :put put)
	 (log-test-set-or-die '(assert-mediawiki-dbms-credentials-p)   :put put)
	 (log-test-set-or-die '(assert-concurrency-limit-xchunk-p)     :put put)
	 (log-test-set-or-die '(assert-virtual-host-p
				assert-virtual-host-name-resolution-p) :put put)
	 (log-test-set        '(warn-if-detect-proxy)                  :put put)
	 )
	((:next-mirror)
	 (log-test-set-or-die '(assert-dbms-connect-with-credentials-wikiuser-p
				assert-dbms-grant-for-wikiuser-p)      :put put)
	 (log-test-set-or-die '(assert-mediawiki-dbms-credentials-p)   :put put)
	 )
	((:add :monitor)
	 (log-test-set-or-die '(assert-dbms-connect-with-credentials-wikiuser-p
				assert-dbms-grant-for-wikiuser-p)     :put put)
	 (log-test-set-or-die '(assert-database-wpmirror-or-create-p) :put put)
	 )
	(otherwise nil))
  (put-message (_ "-----asserting-prerequisite-software-done-----") :put put)
  t)

(defun assert-prerequisite-hardware (&key (put nil))
  "Asserting prerequisite hardware"
  (put-message (_ "-----asserting-prerequisite-hardware-begin----") :put t)
  (case *main-mode*
	((:add :delete :drop :dump :restore-default :update)
	 )
	((:first-mirror :next-mirror)
	 (log-test-set-or-die '(count-cpu
				assert-disk-space-if-large-wikipedia-p
				assert-physical-memory-if-large-wikipedia-p
				assert-partition-free-images
				assert-hdd-write-cache-disabled-p)
			      :put put)
	 (log-test-set        '(warn-if-disk-space-low-p
				assert-internet-access-to-wikimedia-site-p)
			      :put put)
	 )
	((:monitor)
	 )
	(otherwise nil))
  (put-message (_ "-----asserting-prerequisite-hardware-done-----") :put put)
  t)

(defun finalize-system (&key (put nil))
  "Finalizing system"
  (put-message (_ "-----finalizing-begin-------------------------") :put t)
  (case *main-mode*
	((:add :delete :drop :dump :restore-default :update)
	 (log-test-set '(log-stop) :put put)
	 )
	(:first-mirror
	 (log-test-set '(clear-pidfile
			 log-stop) :put put)
	 )
	(:next-mirror
	 )
	(:monitor
	 )
	(otherwise nil))
  (put-message (_ "-----finalizing-done--------------------------") :put put)
  t)


;;;;--------------------------------------------------------------------------+
;;;; Warn user if proxy detected:                                             |
;;;;--------------------------------------------------------------------------+


(defun warn-if-detect-proxy (&key (put t))
  "Warning if see mention of web proxy"
  (let* ((pattern        "[pP][rR][oO][xX][yY]")
	 (message        (_ "proxy config in"))
	 (result-bashrc (grep-first-non-comment *whereis-bashrc* pattern
						:put put))
	 (result-curlrc (grep-first-non-comment *whereis-curlrc* pattern
						:put put))
	 (result-wgetrc (grep-first-non-comment *whereis-wgetrc* pattern
						:put put)))
    (unless (null result-bashrc)
      (put-message-value (format nil "~a ~a" message *whereis-bashrc*) 
			 result-bashrc :put put))
    (unless (null result-curlrc)
      (put-message-value (format nil "~a ~a" message *whereis-curlrc*) 
			 result-curlrc :put put))
    (unless (null result-wgetrc)
      (put-message-value (format nil "~a ~a" message *whereis-wgetrc*)
			 result-wgetrc :put put))
    (if (or result-bashrc result-curlrc result-wgetrc)
	(put-message-done (_ "might have found proxy") :put put)
      (put-message-done (_ "no proxy found") :put put))
    (not (or result-bashrc result-curlrc result-wgetrc))))


;;;;--------------------------------------------------------------------------+
;;;; Determine mode of operation (mirror v. monitor):                         |
;;;;   Mode can be forced by use of command-line options                      |
;;;;     `--gui'     forces monitor mode, first attempting to use GUI         |
;;;;     `--mirror'  forces mirror mode (overriding any monitor options)      |
;;;;     `--monitor' forces monitor mode, first attempting to use GUI         |
;;;;     `--screen'  forces monitor mode, first attempting to use SCREEN      |
;;;;     `--text'    forces monitor mode, using normal text output            |
;;;;  If no options are given, then the PID file routines (below) come into   |
;;;;  play.  1) If no PID file is found, then `wp-mirror' goes into mirror    |
;;;;  mode. 2) If PID file is found (and not stale), then `wp-mirror' goes    |
;;;;  into monitor mode, first attempting to use GUI.                         |
;;;;  This use of PID files is somewhat different from the way daemons (when  |
;;;;  invoked by `/etc/init.d/*') use them.                                   |
;;;;--------------------------------------------------------------------------+


(defun pidfile-exists-p (&key (put nil))
  "Determining if PID file exists.  If so return PID, else NIL"
  (debug-message-start (_ "probing for PID file") :put put)
  (with-open-file (s *whereis-file-pid* :direction :input
		     :if-does-not-exist nil)
    (if s
	(let ((pid (read s)))
	  (debug-message-value (_ "reading value") pid)
	  pid)                  ; return pid
      (progn
	(debug-message (_ "PID file not found"))
	nil))))                	; return nil

(defun proc-exists-p (pid &key (put nil))
  "Determining existence of process with given PID"
  (put-message-value (_ "probing for process with PID") pid :put put)
  (let ((proc-p  (ext:probe-pathname (format nil "/proc/~d" pid))))
    (if proc-p
	(put-message-value-done (_ "probing for process with PID") pid :put put)
      (put-message-value-fail (_ "probing for process with PID") pid :put put))
    proc-p))

(defun first-mirror-exists-p (&key (put nil))
  "Determining if first instance of `wp-mirror' in mirror mode is running."
  (let* ((pid            (pidfile-exists-p :put put))
	 (first-mirror-p (and pid (proc-exists-p pid :put put))))
    (unless first-mirror-p
      (if pid
	  (put-message (_ "PID file found stale - first mirror was probably killed") :put put)
	(put-message (_ "PID file not found") :put put)))
    first-mirror-p))

;;(defun i-am-the-first-mirror-p (&key (put nil))
;;  "Determining if this process set the PID file"
;;  (eql (os:process-id) (pidfile-exists-p :put put)))

(defun mirror-language-list (&key (put nil))
  "Return a list of all language-codes known to the mirror"
  (union *mirror-languages*
	 (when (assert-database-wpmirror-p :put put)
	   (sql-select-languages :put put))
	 :test #'string=))

(defun set-pidfile (&key (put nil))
  "Creating PID file"
  (let ((pid (os:process-id)))
    (put-message-value (_ "my process id is") pid :put put)
    (with-open-file (s *whereis-file-pid* 
		       :direction :output
		       :if-exists :supersede)
    (if s
	(progn
	  (format s "~d~%" pid)
	  (put-message-value-done (_ "creating PID file") *whereis-file-pid*
				  :put put))
      (put-message-fail-and-die (_ "unable to create PID file"))))
    t))

(defun clear-pidfile (&key (put nil))
  "Removing PID file"
  (put-message-start (_ "probing for PID file") :put put)
  (if (ext:probe-pathname *whereis-file-pid*)
      (progn
	(delete-file *whereis-file-pid*)
	(put-message-value-done (_ "removing PID file") *whereis-file-pid*
				:put put))
    (put-message-done (_ "PID file not found") :put put))
  t)

(defun main-mode-mirror-or-monitor (&key (put nil))
  "Determining if this instance of wp-mirror should run in mirror mode
or in monitor mode"
  ;;
  ;; Design note:
  ;;
  ;; Command-line options should override default PID logic.
  ;;
  ;;      mirror cmd-line                         set
  ;; Case exists options   action                 *main-mode*
  ;; ----+------+---------+----------------------+-------------
  ;;  1)    no     none    clear-pid              :first-mirror
  ;;                       set-pidfile
  ;;                       release-all-semaphores
  ;;  2)    no   --mirror  clear-pid              :first-mirror
  ;;                       set-pidfile 
  ;;                       release-all-semaphores
  ;;  3)    no   --monitor   none                 :monitor
  ;;  4)   yes     none      none                 :monitor
  ;;  5)   yes   --mirror    none                 :next-mirror
  ;;  6)   yes   --monitor   none                 :monitor
  ;;
  ;; The options `--gui', `--screen', and `--text', imply `--monitor'.
  ;;
  (put-message-start (_ "determining mode of operation") :put put)
  (let* ((mirror-exists-p (first-mirror-exists-p :put put)))
    (put-message-value (_ "does mirror process exist") mirror-exists-p :put put)
    (put-message-value (_ "does command line force mirror") 
		       *cmd-force-mirror* :put put)
    (put-message-value (_ "does command line force monitor") 
		       *cmd-force-monitor* :put put)
    ;; logic goes here (the six cases mentioned above)
    (setq *main-mode*
	  (cond ((and (not mirror-exists-p)
		      (not *cmd-force-mirror*)
		      (not *cmd-force-monitor*)) :first-mirror)
		((and (not mirror-exists-p)
		      *cmd-force-mirror*)        :first-mirror)
		((and (not mirror-exists-p)
		      (not *cmd-force-mirror*)
		      *cmd-force-monitor*)       :monitor)
		((and mirror-exists-p
		      (not *cmd-force-mirror*)
		      (not *cmd-force-monitor*)) :monitor)
		((and mirror-exists-p
		      *cmd-force-mirror*)        :next-mirror)
		((and mirror-exists-p
		      (not *cmd-force-mirror*)
		      *cmd-force-monitor*)       :monitor)
		(t nil)))
    ;; action goes here
    (case *main-mode*
      (:first-mirror
       (put-message-start (_ "clearing stale PID file") :put put)
       (log-test-set '(clear-pidfile
		       set-pidfile)
		     :put put))
      (:next-mirror
       )
      (:monitor
       )
      (otherwise nil))
    *main-mode*))

(defun set-main-mode-of-operation(&key (put nil))
  "Setting main mode of operation"
  (when (null *main-mode*)
      (setq *main-mode* (main-mode-mirror-or-monitor :put put)))
  (put-flag-message-value :info (_ "set mode of operation to") 
			  (string-upcase *main-mode*))
  *main-mode*)


;;;;--------------------------------------------------------------------------+
;;;; Status utilities:                                                        |
;;;;   State is stored for each file to enable resuming after outage          |
;;;;                                                                          |
;;;;   Globals:                                                               |
;;;;     *input*               - list of valid inputs to finite-state-machine |
;;;;     *state*               - list of valid states                         |
;;;;     *state-transition*    - list of valid state transitions              |
;;;;     *type*                - list of valid file types                     |
;;;;     *type-state-function* - list of functions to be run when a file of a |
;;;;                             given type and state receives the "start"    |
;;;;                             input                                        |
;;;;     *type-state-priority* - prioritized list of type and state (files are|
;;;;                             processed in order of priority)              |
;;;;   Functions:                                                             |
;;;;     fsm-abort             - abort wp-mirror (due to error)               |
;;;;     fsm-boot              - initialize wp-mirror (load `database's)      |
;;;;     fsm-database-checksum - create `checksum' record                     |
;;;;     fsm-database-create   - assert or create `database'                  |
;;;;     fsm-database-grant    - grant privileges on `database' to `wikiadmin'|
;;;;     fsm-database-interwiki- insert interlanguage links into `interwiki'  |
;;;;     fsm-file-count        - count pages and images in `xml' file         |
;;;;     fsm-file-decompress   - decompress `bz2' file                        |
;;;;     fsm-file-download     - download checksum or dump file from wikimedia|
;;;;     fsm-file-import       - import `xchunk' into `xxwiki'                |
;;;;     fsm-file-md5sum       - compute `md5sum' checksum                    |
;;;;     fsm-file-parse        - extract `dump' file name from `checksum' file|
;;;;     fsm-file-remove       - remove unneeded `xml', `xchunk's, `ichunk's  |
;;;;     fsm-file-shell        - run `ichunk' shell script to download images |
;;;;     fsm-file-split        - split decompressed dump into `xchunk's       |
;;;;     fsm-file-validate     - compare `md5sum' with known value            |
;;;;     fsm-file-wikix        - generate `ichunk' with `wikix'               |
;;;;     fsm-images-chown      - chown images to `www-data:www-data'          |
;;;;     fsm-images-count      - count downloaded images                      |
;;;;     fsm-images-rebuild    - import image metadata into `xxwiki'          |
;;;;     fsm-images-validate   - find corrupt files w/ `gm identify -verbose' |
;;;;     fsm-no-op             - do nothing                                   |
;;;;     fsm-process-file      - invokes appropriate fsm-*                    |
;;;;     fsm-transition        - initial state + valid input --> final state  |
;;;;   Design note:                                                           |
;;;;     finite-state-machine  - graphical representation                     |
;;;;                                                                          |
;;;;    __           __         __           __        __              __     |
;;;;   |  |s        |  |s      |  |s        |  |s     |  |s           |  |sdf |
;;;;   v  |         v  |       v  |         v  |      v  |            v  |    |
;;;; +-----+ d  +-------+ d  +-----+ d  +-------+ d  +----+ d        +-----+  |
;;;; |start|--->|created|--->|valid|--->|pending|--->|done|--------->|error|  |
;;;; +-----+    +-------+    +-----+    +-------+    +----+    __    +-----+  |
;;;;     \            \          \            \         \     |  |s    /      |
;;;;      \f           \f         \f           \f        \f   v  |    /df     |
;;;;       \            \          \            \         \  +----+  /        |
;;;;        ------------------------------------------------>|fail|--         |
;;;;                                                         +----+           |
;;;;                                                                          |
;;;;--------------------------------------------------------------------------+


(defconstant  *input*
  '("start" "done" "fail"))
(defconstant  *state*
  '("start" "created" "valid" "pending" "done" "fail" "error"))
(defconstant  *state-transition*
  ;;  state  +  input ->state
  '(("start"   "start" "start"  ) ;resume
    ("start"   "done"  "created")
    ("start"   "fail"  "fail"   )
    ("created" "start" "created") ;resume
    ("created" "done"  "valid"  )
    ("created" "fail"  "fail"   )
    ("valid"   "start" "valid"  ) ;resume
    ("valid"   "done"  "pending")
    ("valid"   "fail"  "fail"   )
    ("pending" "start" "pending") ;resume
    ("pending" "done"  "done"   )
    ("pending" "fail"  "fail"   )
    ("done"    "start" "done"   ) ;resume
    ("done"    "done"  "done"   ) ;no-op
    ("done"    "fail"  "error"  ) ;should not happen
    ("fail"    "start" "fail"   ) ;resume
    ("fail"    "done"  "fail"   ) ;no-op
    ("fail"    "fail"  "error"  ) ;should not happen
    ("error"   "start" "error"  ) ;should not happen
    ("error"   "done"  "error"  ) ;should not happen
    ("error"   "fail"  "error"  ) ;should not happen
    ))
(defconstant  *type*
  '("database" "checksum" "dump" "xml" "xchunk" "ichunk" "images" "error"))
(defconstant  *fsm-function*
  '(fsm-abort
    fsm-boot
    fsm-database-checksum
    fsm-database-create
    fsm-database-grant
    fsm-database-interwiki
    fsm-file-count
    fsm-file-decompress 
    fsm-file-download 
    fsm-file-import
    fsm-file-md5sum
    fsm-file-parse
    fsm-file-shell
    fsm-file-remove
    fsm-file-split
    fsm-file-validate
    fsm-file-wikix
    fsm-images-chown
    fsm-images-count
    fsm-images-rebuild
    fsm-images-validate
    fsm-no-op
))
(defconstant *fsm-function-needs-delete-objectcache*
  '(fsm-file-import
    fsm-image-rebuild))

(defconstant *fsm-function-needs-internet-access*
  '(fsm-file-download
    fsm-file-shell
))
(defconstant *fsm-function-needs-partition-free-images*
  '(fsm-file-decompress
    fsm-file-download
    fsm-file-shell
    fsm-file-split
    fsm-file-wikix
))
(defconstant *fsm-function-needs-partition-free-innodb*
  '(fsm-file-import
    fsm-file-rebuild
))
(defconstant  *type-state-function*
  ;;  type     state     function
  '(("database" "start"   fsm-database-grant    ) ; OK
    ("database" "created" fsm-database-create   ) ; OK
    ("database" "valid"   fsm-database-checksum ) ; OK
    ("database" "pending" fsm-database-interwiki)
    ("database" "done"    fsm-no-op             )
    ("database" "fail"    fsm-abort             )
    ("database" "error"   fsm-abort             )
    ("checksum" "start"   fsm-file-download     ) ; OK
    ("checksum" "created" fsm-file-md5sum       ) ; OK
    ("checksum" "valid"   fsm-file-parse        ) ; OK
    ("checksum" "pending" fsm-no-op             )
    ("checksum" "done"    fsm-no-op             )
    ("checksum" "fail"    fsm-abort             )
    ("checksum" "error"   fsm-abort             )
    ("dump"     "start"   fsm-file-download     )
    ("dump"     "created" fsm-file-validate     ) ; OK
    ("dump"     "valid"   fsm-file-decompress   ) ; OK
    ("dump"     "pending" fsm-no-op             )
    ("dump"     "done"    fsm-no-op             )
    ("dump"     "fail"    fsm-abort             )
    ("dump"     "error"   fsm-abort             )
    ("xml"      "start"   fsm-file-count        ) ; OK
    ("xml"      "created" fsm-file-md5sum       )
    ("xml"      "valid"   fsm-file-split        ) ; OK
    ("xml"      "pending" fsm-file-remove       )
    ("xml"      "done"    fsm-no-op             )
    ("xml"      "fail"    fsm-abort             )
    ("xml"      "error"   fsm-abort             )
    ("xchunk"   "start"   fsm-file-count        )
    ("xchunk"   "created" fsm-file-md5sum       )
    ("xchunk"   "valid"   fsm-file-wikix        ) ; OK
    ("xchunk"   "pending" fsm-file-import       ) ; OK
    ("xchunk"   "done"    fsm-no-op             )
    ("xchunk"   "fail"    fsm-no-op             )
    ("xchunk"   "error"   fsm-abort             )
    ("ichunk"   "start"   fsm-file-count        )
    ("ichunk"   "created" fsm-file-md5sum       )
    ("ichunk"   "valid"   fsm-file-shell        ) ; OK
    ("ichunk"   "pending" fsm-file-remove       )
    ("ichunk"   "done"    fsm-no-op             )
    ("ichunk"   "fail"    fsm-no-op             )
    ("ichunk"   "error"   fsm-abort             )
    ("images"   "start"   fsm-images-count      ) ; OK
    ("images"   "created" fsm-images-validate   ) ; OK
    ("images"   "valid"   fsm-images-rebuild    ) ; OK
    ("images"   "pending" fsm-images-chown      ) ; OK
    ("images"   "done"    fsm-no-op             )
    ("images"   "fail"    fsm-no-op             )
    ("images"   "error"   fsm-abort             )
    ("error"    "start"   fsm-abort             )
    ("error"    "created" fsm-abort             )
    ("error"    "valid"   fsm-abort             )
    ("error"    "pending" fsm-abort             )
    ("error"    "done"    fsm-abort             )
    ("error"    "fail"    fsm-abort             )
    ("error"    "error"   fsm-abort             )))
(defconstant  *type-state-priority*
  ;;  type       state    concurrent non-image
  '(("database" "pending" t          t          ) ; fsm-database-interwiki
    ("database" "valid"   t          t          ) ; fsm-database-checksum
    ("database" "created" t          t          ) ; fsm-database-create
    ("database" "start"   t          t          ) ; fsm-database-grant
    ("checksum" "pending" t          t          ) ; fsm-no-op
    ("checksum" "valid"   t          t          ) ; fsm-file-parse
    ("checksum" "created" t          t          ) ; fsm-file-md5sum
    ("checksum" "start"   nil        t          ) ; fsm-file-download
    ("dump"     "pending" t          t          ) ; fsm-no-op
    ("dump"     "valid"   t          t          ) ; fsm-file-decompress
    ("dump"     "created" t          t          ) ; fsm-file-validate
    ("dump"     "start"   nil        t          ) ; fsm-file-download
    ("xml"      "pending" t          t          ) ; fsm-file-remove
    ("xml"      "valid"   t          t          ) ; fsm-file-split
    ("xml"      "created" t          t          ) ; fsm-file-md5sum
    ("xml"      "start"   t          t          ) ; fsm-file-count
    ("xchunk"   "valid"   t          t          ) ; fsm-file-wikix
    ("xchunk"   "created" t          t          ) ; fsm-file-md5sum
    ("xchunk"   "start"   t          t          ) ; fsm-file-count
    ("ichunk"   "pending" t          t          ) ; fsm-file-remove
    ("ichunk"   "valid"   nil        nil        ) ; fsm-file-shell
    ("ichunk"   "created" t          nil        ) ; fsm-file-md5sum
    ("ichunk"   "start"   t          nil        ) ; fsm-file-count
    ("images"   "start"   t          nil        ) ; fsm-images-count
    ("images"   "created" t          nil        ) ; fsm-images-validate
    ("images"   "valid"   nil        nil        ) ; fsm-images-rebuild
    ("xchunk"   "pending" t          t          ) ; fsm-file-import
    ("images"   "pending" t          nil        ) ; fsm-images-chown
    )) 


(defun fsm-abort (file-name &key (put nil))
  "Aborting due to error condition"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |don't care          none
  ;; state-in  |see *type-state-function*
  ;; state-out |no change
  ;; Return: nil means fail, t means done.
  (let ((file-type  (sql-select-file-type  file-name))
	(file-state (sql-select-file-state file-name)))
    (put-message-start (_ "ERROR: in finite state machine"))
    (put-message (_ "this should never happen"))
    (put-message-value (_ "file-name")  file-name)
    (put-message-value (_ "file-type")  file-type)
    (put-message-value (_ "file-state") file-state)
    (put-message-value-fail-and-abort (_ "ERROR:") file-name))
  nil) ; always mark as `fail'

(defun fsm-boot (&key (put nil))
  "Booting the finite state machine"
  ;; Prime the pump
  ;; Insert database      names into empty `wpmirror.file'
  ;; Insert checksum file names into empty `wpmirror.file'
  (put-message-start (_ "booting finite state machine") :put put)
  (let ((results (loop
		   for language-code in (mirror-language-list :put put)
		   as  database-name = (language-code-to-database-name
					language-code)
		   as  result        = (sql-insert-file language-code "latest" 
							database-name "database"
							:put put)
		   collect result
		   when (eq put t)
		     do (put-message-value (_ "record after insert") 
			 (sql-select-file database-name :put put) :put put))))
    (put-message-done (_ "booting finite state machine") :put put)
    (not (member nil results))))

(defun fsm-database-checksum (database-name &key (put nil))
  "Insert record for checksum file"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |database            checksum
  ;; state-in  |valid               n/a
  ;; state-out |pending             start
  ;; Return: nil means fail, t means done.
  (let* ((language-code (sql-select-file-language database-name :put put))
	 (file-name     (language-code-to-checksums-file-name language-code
							      :put put))
	 (result        (sql-insert-file language-code "latest" 
					 file-name "checksum"
					 :put put)))
    (when (eq put t)
      (put-message-value (_ "record after insert") 
			 (sql-select-file database-name :put put) :put put)))
  t)

(defun fsm-database-create (database-name &key (put nil))
  "Assert or create database using database template file `database_farm.sql'"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |database            n/a
  ;; state-in  |start               n/a
  ;; state-out |created             n/a
  ;; Return: nil means fail, t means done.
  (let ((path-name (merge-pathnames
		    (parse-namestring *whereis-directory-mediawiki-maintenance*)
		    (parse-namestring *whereis-file-mediawiki-farm-database*))))
    (assert-database-or-load database-name path-name :put put)))

(defun fsm-database-grant (database-name &key (put nil))
  "Grant privileges on database to users `wikiadmin' and `wikiuser'"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |database            n/a
  ;; state-in  |created             n/a
  ;; state-out |valid               n/a
  ;; Return: nil means fail, t means done.
  (sql-grant-all-on-database-to-wikiadmin database-name :put put)
  (sql-grant-siud-on-database-to-wikiuser database-name :put put)
  t)

(defun fsm-database-interwiki (database-name &key (put nil))
  "Insert records into `interwiki' database table for interlanguage links"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |database            n/a
  ;; state-in  |pending             n/a
  ;; state-out |done                n/a
  ;; Return: nil means fail, t means done.
  (loop
   for language-code in (mirror-language-list :put put)
   do (sql-insert-interwiki-link database-name language-code :put put))
  t)

(defun fsm-file-shell (ichunk-name &key (put nil))
  "Downloading images for given `ichunk'"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |ichunk              image files
  ;; state-in  |valid               n/a
  ;; state-out |pending
  ;; Return: nil means fail, t means done.
  ;;
  ;; Design note:
  ;;
  ;; `Wikix' has two important bugs.  
  ;;
  ;; 1) If you run `wikix' with the `-i' option, e.g.
  ;;
  ;;   /home/myname# wikix -i /var/lib/mediawiki/images/ \
  ;;                       -o /var/lib/mediawiki/images/ \
  ;;                       < xchunk.xml > ichunk.sh
  ;;
  ;; you will get a script that does not run as expected.  For example,
  ;;
  ;;   /home/myname# ./ichunk.sh
  ;;
  ;; will correctly check under `/var/lib/mediawiki/images/' for
  ;; existing images.  If a particular image is not found, `curl' will
  ;; be invoked and the image downloaded to the directory where
  ;; `ichunk.sh' is running (`/home/myname/' in this example).  The
  ;; script will then incorrectly check again under
  ;; `/var/lib/mediawiki/images/' to see if the image was successfully
  ;; downloaded, and therefore not find it.  The script will then fail
  ;; to store the image there, and will then mis-report the image in
  ;; `failed.log' rather than `download.log'.
  ;; 
  ;; If you want the scripts to run correctly, then 
  ;;   a) run `wikix' without the "-i" option, for example
  ;;
  ;;   /home/myname# wikix < xchunk.xml > ichunk.sh
  ;;    
  ;;   b) move (or copy) `ichunk.sh' to `/var/lib/mediawiki/images/', and then
  ;;   c) run the script in that directory, like so
  ;;
  ;;   /var/lib/mediawiki/images# ./ichunk.sh
  ;;
  ;; It is for this reason, that `fsm-file-shell' (this function)
  ;;   o changes directory to the working directory (if not already there),
  ;;   o copies the ichunk to the images directory, 
  ;;   o changes directory to the images directory,
  ;;   o probes that the copy is there,
  ;;   o runs the copied script (which downloads and stores the images),
  ;;   o removes the copy, and finally
  ;;   o changes directory back to the working directory,
  ;;
  ;; 2) `Wikix' generates image download scripts that incorrectly
  ;; handle the case of partial downloads.
  ;; `http://upload.wikimedia.org/' has thousands of images that
  ;; download only part way, and then block.  Without a timeout,
  ;; mirror building will never finish.  
  ;;
  ;; The built-in alternative to `wikix' generates shell scripts where
  ;; every call to `curl' includes a `-m 1000' (--max-time 1000) option.
  ;;
  (let ((r nil))
    (debug-message ichunk-name)
    (ext:cd *whereis-directory-working*)
    (debug-message-value (_ "copying") ichunk-name)
    (setq r (null (shell-copy-file ichunk-name *whereis-directory-images*
				   :put put)))
    (when r
      (debug-message-value (_ "changing to images directory")
			   *whereis-directory-images*)
      (ext:cd *whereis-directory-images*)
      (debug-message-value (_ "current working directory is now") (ext:cd))
      (debug-message-value (_ "probing") (ext:probe-pathname ichunk-name))
      (setq r (ext:probe-pathname ichunk-name))
      (when r
	(debug-message-value (_ "running") (ext:probe-pathname ichunk-name))
	(setq r (null (shell-script ichunk-name :put put)))
	(when r
	  (debug-message-value (_ "removing") (ext:probe-pathname ichunk-name))
	  (shell-remove-file ichunk-name :put put))))
    (debug-message-value (_ "changing to working directory")
			 *whereis-directory-working*)
    (ext:cd *whereis-directory-working*)
    (debug-message-value (_ "current working directory is now") (ext:cd))
    (put-flag-message :info (_ "done"))
    r))

(defun fsm-file-count (file-name &key (put nil))
  "Counting pages `<page>' and images `[[File' and `[[Image:'  in `.xml' files"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |any except dump     none
  ;; state-in  |start
  ;; state-out |created
  ;; Return: nil means fail, t means done.
  (sql-update-file-size file-name :put put)
  (let ((file-type   (sql-select-file-type file-name :put put))
	(page-count  0)
	(image-count 0))
    (when (or (string= file-type "xml") (string= file-type "xchunk"))
      (setq page-count (shell-count-page  file-name :put put))
      (sql-update-file file-name :pages page-count :put put))
    (setq image-count (shell-count-image file-name :put put))
    (sql-update-file file-name :images image-count :put put)))

(defun fsm-file-decompress (file-name &key (put nil))
  "Decompressing dump file to yield XML file"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |dump                dump-xml
  ;; state-in  |valid               n/a
  ;; state-out |pending             start
  ;; Return: nil means fail, t means done.
  ;; 1) decompress dump-file
  (let* ((language  (sql-select-file-language file-name :put put))
	 (file-date (sql-select-file-date     file-name :put put))
	 (xml-name  (string-right-trim ".bz2" file-name)))
    (put-message-value (_ "file dump xml name") xml-name :put put)
    (put-message-start (_ "decompressing") :put put)
    (shell-bunzip2 file-name :put put)
    (put-message-done (_ "decompressing") :put put)
    ;; 2) create record for decompressed file
    (when (ext:probe-pathname xml-name)
      (sql-insert-file language file-date xml-name "xml" :put put)
      (sql-update-file-size xml-name :put put))))

(defun fsm-file-download (file-name &key (put nil))
  "Downloading checksum or dump file"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |checksum or dump    checksum or dump
  ;; state-in  |start
  ;; state-out |created
  ;; Return: nil means fail, t means done.
  ;;
  ;; Names:
  ;; site     is   `http://dumps.wikimedia.org/'
  ;; checksum like `xxwiki/yyyymmdd/xxwiki-latest-md5sums.txt'
  ;; dump     like `xxwiki/yyyymmdd/xxwiki-yyyymmdd-pages-articles.xml.bz2'
  ;;
  ;; Design note:
  ;;
  ;; Languages such as `zh-classical' (i.e. `xx' contains a dash `-')
  ;; must be carefully handled.  MySQL does not readily accept
  ;; database names containing a dash:
  ;;
  ;; fail:  CREATE DATABASE zh-classicalwiki;
  ;; OK  :  CREATE DATABASE `zh-classicalwiki`;
  ;; OK  :  CREATE DATABASE zh_classicalwiki;
  ;; 
  ;; WikiMedia Foundation usage is:
  ;; 
  ;; underscore : directory     : http://dumps.wikimedia.org/zh_classical/...
  ;;            : checksum file : zh_classicalwiki-latest-md5sums.txt
  ;;            : dump file     : zh_classical-yyyymmdd-pages-articles.xml.bz2
  ;; dash       : URL           : http://zh-classical.wikipedia.org/
  ;;
  ;; Hence: http://dumps.wikimedia.org/zh_classicalwiki/latest/zh_classicalwiki
  ;;
  ;; Wp-mirror attempts to follow WikiMedia Foundation usage:
  ;; languages use dash, while files and databases use underscore.
  ;; See functions `language-to-file-name' and `str_replace'.
  ;;
  (debug-message-value (_ "file-name") file-name)
  (let* ((file-type     (sql-select-file-type     file-name :put put))
	 (language-code (sql-select-file-language file-name :put put))
	 (url       nil)
	 (file-path nil))
    (cond ((string= file-type "checksum")
	   (setq url       (language-code-to-checksums-url  language-code))
	   (setq file-path (language-code-to-checksums-path language-code)))
	  ((string= file-type "dump")
	   (setq url       (language-code-to-dump-url       language-code))
	   (setq file-path (language-code-to-dump-path      language-code)))
	  (t nil))
    (put-message url :put put)
    (when (download-file url file-path :put put))
      (put-message file-path :put put)
      (when (file-exists-p file-path :put put)
	(sql-update-file-size file-name :put put))))

(defun fsm-file-import (xchunk-name &key (put nil))
  "Importing `xchunk' into `xxwiki'"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |xchunk              none
  ;; state-in  |pending             n/a
  ;; state-out |done                
  ;; Return: nil means fail, t means done.
  ;;
  ;; Design note:
  ;;
  ;; Problem.  This import process `importDump.php' has been known to
  ;; loop forever.
  ;;
  ;; Solution.  We impose a timeout.
  ;;
  (debug-message-value (_ "xchunk-name") xchunk-name)
  (let* ((language     (sql-select-file-language xchunk-name :put put))
	 (date         (sql-select-file-date     xchunk-name :put put))
	 (page         (sql-select-file-page     xchunk-name :put put))
	 (pages        (sql-select-file-pages    xchunk-name :put put))
	 (result       nil))
    (put-message xchunk-name :put put)
    (setq result (not (shell-mediawiki-farm-importdump xchunk-name
			    :timeout *mirror-importdump-timeout-sec-max*
			    :put put)))
    (put-flag-message :info (_ "done"))
    result))
				       
(defun fsm-file-md5sum (file-name &key (put nil))
  "Computing md5sum checksum"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |any except dump     n/a
  ;; state-in  |created             
  ;; state-out |valid               
  ;; Return: nil means fail, t means done.
  (debug-message-value (_ "file-name") file-name)
  (let ((result (shell-md5sum file-name :put put)))
    (if (null result)
	(put-message-value-fail-and-abort (_ "md5sum returned") result)
      (progn
	(debug-message-value (_ "md5sum") result)
	(sql-update-file-md5sum file-name result :put put)))))

(defun fsm-file-parse (file-name &key (put nil))
  "Parsing checksums file to identify dump file"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |checksum            dump file
  ;; state-in  |valid               n/a
  ;; state-out |pending
  ;; Return: nil means fail, t means done.
  ;;
  (put-message-value (_ "checksums file") file-name :put put)
  (put-message-start (_ "looking for pattern match") :put put)
  (let* ((language       (sql-select-file-language file-name :put put))
	 (file-template  (first (last (regexp:regexp-split 
				       "/" *wikimedia-dump-template*))))
	 (pattern        (join (cddr (regexp:regexp-split "-" file-template))
			       "-"))
	 (line           (grep-first file-name pattern :put put))
	 (dump-checksum  (first  (regexp:regexp-split " " line)))
	 (dump-name      (first  (last (regexp:regexp-split " " line))))
	 (dump-date      (second (regexp:regexp-split "-" dump-name))))
    (put-message-value (_ "dump file name")     dump-name :put put)
    (put-message-value (_ "dump file checksum") dump-checksum :put put)
    (put-message-done  (_ "looking for pattern match") :put put)
    ;; store status in `wpmirror'
    (put-message-start (_ "probing dump file") :put put)
    (let ((r (sql-select-file dump-name :put put)))
      (put-message-value (_ "sql-select-file dump-name returned") r :put put)
      (when (null r)
	(put-message-start (_ "Seeing newer dump so cleaning any old ones")
			   :put put)
	(dump-file-remove-p language :put put)
	(put-message-done (_ "Seeing newer dump so cleaning any old ones") 
			  :put put)
	(put-message-start (_ "sql-insert-file dump") :put put)
	(sql-insert-file language dump-date dump-name "dump"
			 :md5sum dump-checksum :put put)
	(put-message-done (_ "sql-insert-file dump") :put put)))
    (put-message-done (_ "probing dump file") :put put)
    t))

(defun fsm-file-remove (file-name &key (put nil))
  "Removing (deleting) working files that are no longer needed"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |xml, xchunk, ichunk none
  ;; state-in  |pending
  ;; state-out |done
  ;; Return: nil means fail, t means done.
  (let ((file-type   (sql-select-file-type file-name :put put)))
    (when (or (string= file-type "xml") (string= file-type "xchunk")
	      (string= file-type "ichunk"))
      (not (shell-remove-file file-name :put put)))))


(defun fsm-file-split (file-name &key (put nil))
  "Splitting decompressed dump file into `xchunk's"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |dump-xml            xchunks
  ;; state-in  |valid               n/a
  ;; state-out |pending             start
  ;; Return: nil means fail, t means done.
  (let ((language  (sql-select-file-language file-name :put put))
	(file-date (sql-select-file-date     file-name :put put)))
    (with-open-file (f file-name
		       :direction :input
		       :if-does-not-exist nil)
      ;; generating the set of xchunks
      (loop named xchunk-set
	with mediawiki-tag-start = (read-line f nil nil)
	and mediawiki-tag-end   = "</mediawiki>"
	and file-name-head      = (concatenate 'string
					       (string-right-trim 
						".xml"
						file-name))
	and file-name-tail      = (concatenate 'string
					       (format nil "-c~9,'0d.xml"
						       *xchunk-page-count*))
	and page                = 0
	;; xchunk file-names will look like
	;; `enwiki-yyyymmdd-pages-articles-p000000000-c000001000.xml'
	for file-name-page of-type string
	                 = (format nil "-p~9,'0d" page) ; see <termination>
	as  chunk-name     of-type string
                         = (concatenate 'string         ; below for end-test
					file-name-head
					file-name-page
					file-name-tail)
	;; creating one xchunk
	count chunk-name into chunk-count
	do (debug-message-value (_ "xchunks-generated") chunk-count)
	   ;; 0) create record in `file' table of `wpmirror'
	   (sql-insert-file language file-date chunk-name "xchunk" :page page
			    :put put)
	   ;; 1) create file on disk
	   (with-open-file (x chunk-name
			      :direction :output
			      :if-exists :supersede)
	     (loop named xchunk
	       initially (debug-message-start (_ "creating xchunk"))
	                 (put-message chunk-name :put put)
			 (write-line mediawiki-tag-start x)
	       with xchunk-page-count = 0
	       for l of-type string = (read-line f nil 'eof) 
	                            then (read-line f nil 'eof)
	       when (eq l 'eof) do (return-from xchunk-set t) ; <termination>
	       do   (write-line l x)
	       when (string= l "  </page>") 
	       do   (setq xchunk-page-count (1+ xchunk-page-count))
                    (setq page (1+ page))
	       until (eql xchunk-page-count *xchunk-page-count*)
	       finally   (write-line mediawiki-tag-end   x)
	                 (debug-message-done (_ "creating xchunk")))
	     t)
	   ;; 2) update record in `file' table of `wpmirror'
	   (when (ext:probe-pathname chunk-name)
	     (sql-update-file-size file-name :put put))))))

(defun fsm-file-validate (file-name &key (put nil))
  "Validating downloaded dump file against known checksum"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |dump                none
  ;; state-in  |created             
  ;; state-out |valid               
  ;; Return: nil means fail, t means done.
  (let* ((dump-checksum-should (sql-select-file-md5sum file-name :put put))
	 (dump-checksum-actual (shell-md5sum file-name :put put)))
    ;; 1) update `size' just in case the dump file was manually downloaded
    ;;    into the working directory
    (sql-update-file-size file-name :put put)
    ;; 2) compare known checksum with actual checksum
    (put-message-value (_ "checksum should be") dump-checksum-should :put put)
    (put-message-value (_ "checksum actually is") dump-checksum-actual :put put)
    (equal dump-checksum-should dump-checksum-actual)))

(defun fsm-file-wikix (xchunk-name &key (put nil))
  "Generating `ichunk' for given `xchunk' using `wikix'"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |xchunk              ichunk
  ;; state-in  |valid               n/a
  ;; state-out |pending             start
  ;; Return: nil means fail, t means done.
  (when *mirror-image-download-p*
    (let* ((language     (sql-select-file-language xchunk-name :put put))
	   (date         (sql-select-file-date     xchunk-name :put put))
	   (page         (sql-select-file-page     xchunk-name :put put))
	   (pages        (sql-select-file-pages    xchunk-name :put put))
	   (ichunk-name  (concatenate 'string
				      (string-right-trim ".xml" xchunk-name)
				      ".sh"))
	   (process      nil))
      (debug-message xchunk-name)
      (debug-message ichunk-name)
      ;; generate i-chunk
      (if *mirror-image-use-wikix-p*
	  (progn 
	    ;; deprecated as of WP-MIRROR 0.4 (not POSIX compliant)
	    (shell-wikix     xchunk-name ichunk-name :put put)
	    (ichunk-edit-p   ichunk-name))
	(let ((image-hash-table (make-hash-table :test 'equal))) ; test strings
	  ;; to debug pathological file names, set :put t 
	  (parse-image-file-names xchunk-name image-hash-table :put put)
	  (filter-exist image-hash-table :put put)
	  (put-shell-script ichunk-name language image-hash-table :put put)))
      ;; insert i-chunk state info
      (sql-insert-file language date ichunk-name "ichunk"
		       :page page :pages pages :put put)
      (shell-chmod     ichunk-name :put put)))
  t) ; do not mark `xchunk' as `fail', even if `wikix' fails


(defun known-image-extension-p (string)
  "Matching with known file extensions"
  (or (regexp:match "\\.jpg$"  string :ignore-case t) ; jpg
      (regexp:match "\\.jpeg$" string :ignore-case t) ; jpeg
      (regexp:match "\\.png$"  string :ignore-case t) ; png
      (regexp:match "\\.gif$"  string :ignore-case t) ; gif
      (regexp:match "\\.svg$"  string :ignore-case t) ; svg
      (regexp:match "\\.pdf$"  string :ignore-case t) ; pdf
      (regexp:match "\\.tif$"  string :ignore-case t) ; tif
      (regexp:match "\\.tiff$" string :ignore-case t) ; tiff
      (regexp:match "\\.ogg$"  string :ignore-case t) ; ogg
      (regexp:match "\\.djvu$" string :ignore-case t) ; djvu
      (regexp:match "\\.wav$"  string :ignore-case t) ; wav
      (regexp:match "\\.mid$"  string :ignore-case t) ; mid
      (regexp:match "\\.fig$"  string :ignore-case t) ; fig
      (regexp:match "\\.bmp$"  string :ignore-case t) ; bmp
      (regexp:match "\\.xcf$"  string :ignore-case t) ; xcf
      (regexp:match "\\.dia$"  string :ignore-case t) ; dia
      (regexp:match "\\.mp3$"  string :ignore-case t) ; mp3
      (regexp:match "\\.psp$"  string :ignore-case t) ; psp
      (regexp:match "\\.bz2$"  string :ignore-case t) ; bz2
      (regexp:match "\\.mov$"  string :ignore-case t) ; mov
      ))

(defun normalize-file-name (string-in &key (put nil))
  "Normalizing file name"
  (let* ((dummy1      (debug-message string-in))
	 ;; trim leading and trailing whitespace, use underscore not space
	 (string-mid  (substitute #\_ #\Space (string-trim " " string-in)))
	 (dummy2      (debug-message string-mid))
	 (string-out  (make-array 0
				  :element-type 'character
				  :fill-pointer 0
				  :adjustable t)))
    (loop
       for char across string-mid
       ;; this does not yet handle the AMP commands like
       ;; &lt; &gt; &amp; &copy; &quot; &nbsp; and so on
       when (member char '(#\- #\' #\" #\( #\) #\& #\; #\` #\@ #\* #\$))
         do (vector-push-extend #\\ string-out)
       do (vector-push-extend char string-out))
    string-out))

(defun parse-image-file-names (xchunk-name image-hash-table &key (put nil))
  "Parsing image file names from `x-chunk'"
  ;;
  ;; Design note:
  ;;
  ;; Problem: File names with special characters can open security
  ;; holes.  For example:
  ;;
  ;; shell# curl http://upload.wikimedia.org/wikipedia/en/0/00/foo`bar`baz.jpg
  ;; bash: bar: command not found
  ;;
  ;; Clearly some file names present a security risk.  A malicious
  ;; person could upload image files with names containing shell
  ;; commands (e.g. replace `bar` with any shell command).  As a not
  ;; too hazardous example, try
  ;;
  ;; shell# curl http://upload.wikimedia.org/wikipedia/en/0/00/foo`ls`baz.jpg
  ;;
  ;; which will produce one error message for each file in your
  ;; current working directory.
  ;;
  ;; Now, guess what the next example does.
  ;;
  ;; shell# curl http://upload.wikimedia.org/wikipedia/en/0/00/foo`halt`baz.jpg
  ;;
  ;; Solution: Ignore file names with certain special characters.
  ;; 
  (let ((xchunk-path      (merge-pathnames
			   (parse-namestring *whereis-directory-working*)
			   (parse-namestring xchunk-name))))
    (with-open-file (s xchunk-path :direction :input)
      (loop named many-lines
	for line  = (read-line s nil nil)
	while line
	do
	;(debug-message line)
	(loop named one-line
	  with string = (make-array 0
				    :element-type 'character 
				    :fill-pointer 0
				    :adjustable t)
	  with state = 0                             ; newline
	  for char across line
	  ;;when (= state 34) do (put-message line :put t) end
	  ;;when (and (>= state 34) (<= state 35))
	  ;;  do (put-message char :put t) end
	  do
	  (cond ;; state 0 - ignore leading spaces `<newline> ' 
		((and (eql state  0)(equal char #\Space))     (setq state  0))
                ;; state 0 - look for `<newline>[[' etc. 
		((and (eql state  0)(equal char #\[))         (setq state  2))
		;; state 0 - look for `<newline>{{' etc. 
		((and (eql state  0)(equal char #\{))         (setq state  3))
		;; state 0 - look for `<newline>|File:' etc. in `{{gallery' 
                ((and (eql state  0)(equal char #\|))         (setq state 10))
                ;; state 0 - look for `<newline>File:'  etc. in `<gallery>'
                ;;                                              `<imagemap>'
	        ((and (eql state  0)(member char '(#\F #\f))) (setq state 20))
		((and (eql state  0)(member char '(#\I #\i))) (setq state 30))
		((and (eql state  0)(member char '(#\M #\m))) (setq state 40))
		;; state 0 - look for `=foo.png<EOL>' from `{{infobox'
		((and (eql state  0)(equal char #\=))         (setq state 90))
		(     (eql state  0)                          (setq state  1))

		;; state 1 - look for link `[[' 
		((and (eql state  1)(equal char #\[)) (setq state  2))
		((and (eql state  2)(equal char #\[)) (setq state 10))
		(     (eql state  2)                  (setq state  1))

		;; state 1 - look for template `{{' 
		((and (eql state  1)(equal char #\{)) (setq state  3))
		((and (eql state  3)(equal char #\{)) (setq state 50))
		(     (eql state  3)                  (setq state  1))
		
		;; state 1 - look for `=foo.png<EOL>' from `{{infobox'
		((and (eql state  1)(equal char #\=))         (setq state 90))

		;; state 10 - look for <newline>`|F'
		((and (eql state 10)(equal char #\Space))     (setq state 10))
		((and (eql state 10)(equal char #\[))         (setq state  2))
		((and (eql state 10)(equal char #\{))         (setq state  3))
		((and (eql state 10)(member char '(#\F #\f))) (setq state 20))
		((and (eql state 10)(member char '(#\I #\i))) (setq state 30))
		((and (eql state 10)(member char '(#\M #\m))) (setq state 40))
		(     (eql state 10)                          (setq state  1))

		;; state 20 - look for `File:' or `file:'
		((and (eql state 20)(equal char #\i)) (setq state 21))
		((and (eql state 21)(equal char #\l)) (setq state 22))
		((and (eql state 22)(equal char #\e)) (setq state 23))
		((and (eql state 23)(equal char #\:)) (setq state 90))
		((member state '(20 21 22 23))        (setq state  1))

		;; state 30 - look for `Image:' or `image:'
		((and (eql state 30)(equal char #\m)) (setq state 31))
		((and (eql state 31)(equal char #\a)) (setq state 32))
		((and (eql state 32)(equal char #\g)) (setq state 33))
		((and (eql state 33)(equal char #\e)) (setq state 34))
		((and (eql state 34)(equal char #\:)) (setq state 90))
		;; state 30 - look for `|image=' or `|image2=' w/ whitespace
		;;                                        in `{{multiple image'
		((and (eql state 34)(equal char #\=)) (setq state 90))
		((and (eql state 34)(member char
		    '(#\Space #\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9)))
                                                      (setq state 35))
		((and (eql state 35)(equal char #\Space)) (setq state 35))
		((and (eql state 35)(equal char #\=)) (setq state 90))
		((member state '(30 31 32 33 34 35))  (setq state  1))

		;; state 40 - look for `Media:' or `media:'
		((and (eql state 40)(equal char #\e)) (setq state 41))
		((and (eql state 41)(equal char #\d)) (setq state 42))
		((and (eql state 42)(equal char #\i)) (setq state 43))
		((and (eql state 43)(equal char #\a)) (setq state 44))
		((and (eql state 44)(equal char #\:)) (setq state 90))
		((member state '(40 41 42 43 44))     (setq state  1))

		;; state 50 - look for template `{{wide image|' 
		((and (eql state 50)(equal char #\w)) (setq state 51))
		((and (eql state 51)(equal char #\i)) (setq state 52))
		((and (eql state 52)(equal char #\d)) (setq state 53))
		((and (eql state 53)(equal char #\e)) (setq state 54))
		((and (eql state 54)(equal char #\Space)) (setq state 55))
		((and (eql state 55)(equal char #\Space)) (setq state 55))
		((and (eql state 55)(equal char #\i)) (setq state 56))
		((and (eql state 56)(equal char #\m)) (setq state 57))
		((and (eql state 57)(equal char #\a)) (setq state 58))
		((and (eql state 58)(equal char #\g)) (setq state 59))
		((and (eql state 59)(equal char #\e)) (setq state 60))
		((and (eql state 60)(equal char #\Space)) (setq state 60))
		((and (eql state 60)(equal char #\|)) (setq state 90))
		((member state '(50 51 52 53 54 55 56 57 58 59 60))
			                              (setq state  1))

		;; state 90 - potentially an image file name
		((and (eql state 90)
		      (or (equal char #\:)           ; failing on 2nd colon
			  (member char '(#\< #\>))   ; failing on XML tag
			  (member char '(#\{ #\}))   ; failing on braces
			  (equal char #\/)           ; failing on URL
			  (equal char #\~)           ; failing on tilde
			  (equal char #\%)           ; failing on percent
			  (equal char #\*)           ; failing on asterisk
			  ;; failing on whitespace
			  (member char '(#\Backspace #\Linefeed #\Page
					 #\Return #\Rubout #\Tab))
			  ;; delete this when we decide how to deal 
			  ;; with AMP codes (like &amp;)
			  (equal char #\&)           ; failing on ampersand
			  ;; delete this when we can mix backquote with unicode
			  ;; delete this when we can protect shell from attack
			  (equal char #\`)
			  ;; delete this when proxy does not `302 Redirect'
			  ;(equal char #\')
			  ))
		 (setq state 1) ; reset
		 (setq string (make-array 0
					  :element-type 'character
					  :fill-pointer 0
					  :adjustable t)))
		;; state 90 - probably ran into a link
		((and (eql state 90) (equal char #\[))
		 (setq state 2) ; possible link
		 (setq string (make-array 0
					  :element-type 'character
					  :fill-pointer 0
					  :adjustable t)))
		;; state 90 - probably found valid image file name
		((and (eql state 90) (member char '(#\| #\] #\Newline)))
		 ;(debug-message line)
		 (when (known-image-extension-p string)
		   ;; storing file-name and md5sum in hash table
		   ;; this has the effect of suppressing duplicates
		   (let* ((dummy0      (debug-message line))
			  (dummy1      (debug-message string))
			  (norm-string (normalize-file-name string :put put))
			  (md5sum      (string-to-md5sum norm-string :put put))
			  (d1          (subseq md5sum 0 1))
			  (d2          (subseq md5sum 0 2)))
		     (put-message-value md5sum norm-string :put put)
		     (setf (gethash norm-string image-hash-table)
			   (format nil "~a/~a/" d1 d2))))
		 (setq state 1) ; reset
		 (setq string (make-array 0
				  :element-type 'character
				  :fill-pointer 0
				  :adjustable t)))
		;; state 90 - building image file name char-by-char
		((eql state 90)
		 (vector-push-extend char string)))
	  finally
		;; state 90 - probably found valid image file name
	        ;             usually `=foo.png<EOL>' from `{{infobox'
	    (when (eql state 90)
		 ;(debug-message line)
		 (when (known-image-extension-p string)
		   ;; storing file-name and md5sum in hash table
		   ;; this has the effect of suppressing duplicates
		   (let* ((dummy0      (debug-message line))
			  (dummy1      (debug-message string))
			  (norm-string (normalize-file-name string   :put put))
			  (md5sum      (string-to-md5sum norm-string :put put))
			  (d1          (subseq md5sum 0 1))
			  (d2          (subseq md5sum 0 2)))
		     (put-message-value md5sum norm-string :put put)
		     (setf (gethash norm-string image-hash-table)
			   (format nil "~a/~a/" d1 d2))))))
      ))))

(defun string-to-md5sum (string &key (put nil))
  "Computing md5sum of a string, return as a 32 char hex string"
  (let* ((dummy2      (debug-message string))
	 (no-escape   (remove #\\ string))
	 (dummy3      (debug-message no-escape)))
    ;; compute md5sum
    (cond ((find #\` string)
	   ;; backquote OK, fail on unicode
	   (simple-string-to-md5sum no-escape :put put))
	   ;; unicode OK, fail on backquote
	  (t (shell-printf-md5sum string :put put)))))

(defun simple-string-to-md5sum (string &key (put nil))
  "Computing md5sum of a string, return as a 32 char hex string"
  ;; Handles backquote '`'
  ;; Bug: fail if neither simple-string nor (unsigned-byte 8)
  (put-message-start (format nil "string-to-md5sum ~a" string) :put put)
  (let* ((sequence  (md5:md5sum-sequence string))
	 (ident     (map 'list #'identity sequence))
	 (result    (format nil "~(~{~2,'0x~}~)" ident)))
    (put-message-value (_ "string-to-md5sum") result :put put)
    result))

(defun filter-exist (image-hash-table &key (put nil))
  "Removing from hash-table names of existing image files"
  (loop
    with path = nil
    for key being the hash-keys of image-hash-table
    using (hash-value value)
    do
    (put-message-value value key :put put)
    (setq path (parse-namestring
		(format nil "~a~a~a"
			*whereis-directory-images* 
			value
			(remove #\\ key))))             ; unescaped file name
    (put-message path :put put)
    (unless (or (find #\? key)                          ; skip wildcards
		(find #\* key))
      (when (ext:probe-pathname path)
	(remhash key image-hash-table)))))

(defun put-shell-script (ichunk-name language image-hash-table &key (put nil))
  "Putting shell script for downloading images (like output of `wikix')"
  ;;
  ;; Design note:
  ;;
  ;; We apply several options to `cURL'.  
  ;;
  ;;    curl -f -m 1000 -O --retry 0 <pathname>
  ;;
  ;; The following text is excerpted from its man page:
  ;;
  ;; -f/--fail
  ;;        (HTTP) Fail silently (no output at all) on server errors.
  ;; -m/--max-time <seconds>
  ;;        Maximum time in seconds that you allow the whole operation to
  ;;        take.  This is useful for preventing your batch jobs from 
  ;;        hanging for hours due to slow networks or links going down.
  ;; -O/--remote-name
  ;;        Write output ot a local file named like the remote file we get.
  ;;        (Only the file part of the remote file is used, the path is cut
  ;;        off.)
  ;; --retry <num>
  ;;        If a transient error is returned when curl tries to perform a
  ;;        transfer, it will retry this number of times before giving up.
  ;;
  (let ((ichunk-path (merge-pathnames
		      (parse-namestring *whereis-directory-working*)
		      (parse-namestring ichunk-name))))
    (with-open-file (s ichunk-name
		       :direction :output
		       :if-exists :supersede)
    (loop
      initially
      (format s "#!/bin/sh~%~%")
      (format s "COMMONSPATH=http://upload.wikimedia.org/wikipedia/commons/~%")
      (format s "IMAGEPATH=http://upload.wikimedia.org/wikipedia/~a/~%" 
	      language)
    for key being the hash-keys of image-hash-table
    using (hash-value value)
    as  path = (format nil "~a~a" value key)
    do
    (format s "~%")
    (format s "if [ -e ./~a ]\; then~%" path)
    (format s "        echo ./~a already exists >> exists.log~%" path)
    (format s "else~%")
    (format s "        curl -f -m 1000 -O --retry 0 $COMMONSPATH./~a~%" path)
    (format s "        if [ -e ./~a ]\; then~%" key)
    (format s "                /bin/mkdir -p ./~a~%" value)
    (format s "                /bin/mv ./~a ./~a~%" key value)
    (format s "                echo ./~a downloaded >> download.log~%" path)
    (format s "        else~%")
    (format s "                curl -f -m 1000 -O --retry 0 $IMAGEPATH./~a~%" path)
    (format s "                if [ -e ./~a ]\; then~%" key)
    (format s "                        /bin/mkdir -p ./~a~%" value)
    (format s "                        /bin/mv ./~a ./~a~%" key value)
    (format s "                        echo ./~a downloaded >> download.log~%" path)
    (format s "                else~%")
    (format s "                        echo ./~a failed >> failed.log~%" path)
    (format s "                fi~%")
    (format s "        fi~%")
    (format s "fi~%")
    ))))

(defun fsm-function (file-type file-state &key (put nil))
  "Determining function to use according to *type-state-function* table"
  (debug-message-value (_ "file type, state")
		       (format nil "~a, ~a" file-type file-state))
  (unless (member file-type *type* :test #'string=)
    (put-message-value-fail-and-abort (_ "Error: FSM invalid type") file-type))
  (unless (member file-state *state* :test #'string=)
    (put-message-value-fail-and-abort (_ "Error: FSM invalid state") file-state))
  (dolist (tuple *type-state-function* nil)
    (when (and (string= file-type  (first  tuple))
	       (string= file-state (second tuple)))
      (progn
	(put-message-value (_ "file type, state, and function")
			   (format nil "~a, ~a, ~a" 
				   (first  tuple)      ; type
				   (second tuple)      ; state
				   (third  tuple))     ; function
			   :put put)
	(return (third tuple)))))) ; break loop        ; <--- return function

(defun fsm-images-count (file-name &key (put nil))
  "Counting images in the images directory"
  (put-message-value (_ "counting images") file-name :put put)
  (let* ((name-split     (regexp:regexp-split "-" file-name))
	 (dir-name-image (format nil "~a~a/~a/"
				 *whereis-directory-images*
				 (first (last name-split 2)) 
				 (first (last name-split  ))))
	 (probe          (ext:probe-pathname dir-name-image))
	 (image-count    0))
    (when (not (null probe))
      (setq image-count (shell-count-dir dir-name-image :put put)))
    (sql-update-file file-name :images image-count :put put)
    (put-message-value-done (_ "counting images") image-count :put put))
  t)

(defun fsm-images-chown (file-name &key (put nil))
  "Setting ownership of image, math, and thumb directories"
  ;;
  ;; Design note:
  ;;
  ;; Wikix downloads image files, and stores them with `root:root'
  ;; ownership.  However, when you browse the mirrored Wikipedia
  ;; articles, MediaWiki will not generate thumbs of image files with
  ;; `root:root' ownership, but will instead render gray rectangles.
  ;;
  ;; MediaWiki expects `www-data:www-data' ownership.  So we `chown'.
  ;;
  (put-message-value (_ "chowning images") file-name :put put)
  (let* ((name-split (regexp:regexp-split "-" file-name))
	 (dir-name-image (format nil "~a~a/~a/" 
				 *whereis-directory-images*
				 (first (last name-split 2)) 
				 (first (last name-split  ))))
	 (dir-name-math  (format nil "~amath/~a/~a/"
				 *whereis-directory-images*
				 (first (last name-split 2)) 
				 (subseq (first (last name-split  )) 1 2)))
	 (dir-name-thumb (format nil "~athumb/~a/~a/"
				 *whereis-directory-images*
				 (first (last name-split 2)) 
				 (first (last name-split  ))))
	 (probe-image    (ext:probe-pathname dir-name-image))
	 (probe-math     (ext:probe-pathname dir-name-math ))
	 (probe-thumb    (ext:probe-pathname dir-name-thumb)))
  ;; 1) directories which hold images
    (when (not (null probe-image))
      (shell-chown-dir dir-name-image :put put))
  ;; 2) `math' directory which holds equations formated with TeX 
    (when (not (null probe-math))
      (shell-chown-dir dir-name-math  :put put))
  ;; 3) `thumb' directory which holds resized images
    (when (not (null probe-thumb))
      (shell-chown-dir dir-name-thumb :put put)))
  t)

(defun fsm-images-rebuild (file-name &key (put nil))
  "Running `rebuildImages.php' to load image meta-data into `xxwiki.image'"
  ;;
  ;; Design note:
  ;;
  ;; Problem: It seems that `rebuildImages.php' has a memory leak.
  ;; For every 20,000 new images processed, it captures 1G memory.  
  ;;
  ;; Solution: To deal with this, we impose a time out to kill the
  ;; forked process before it occupies all memory.  Then we start
  ;; over.
  ;;
  ;; Timing:  
  ;;  o Old images are checked at about 150 files/sec.
  ;;  o New image files are processed at about 5 files/sec.
  ;;
  ;; Problem: The `en' wikipedia has about 2 million images (as of
  ;; 2012).  Each time we start over, we must first loop through image
  ;; files that have already been processed, before we can get to the
  ;; new images.
  ;;
  ;; Solution: We start with a modest time out, and gradually increase
  ;; it, up to a cap.  The very last file has no time out and serves
  ;; to finish-up any remaining unprocessed image files.
  ;;
  ;; Assuming
  ;;  o *mirror-rebuildimages-timeout-sec-min* =  6000
  ;;  o *mirror-rebuildimages-timeout-sec-inc* =   200
  ;;  o *mirror-rebuildimages-timeout-sec-max* = 20000  <-- cap for safety
  ;;
  ;; The time out schedule (for initial build) would look like this:
  ;;
  ;;                             Time out  Images  Images 
  ;; File-name                      [sec]     Old     New
  ;; ---------------------------+--------+-------+-------
  ;; xxwiki-yyyymmdd-images-0-00     6000       0   30000
  ;; xxwiki-yyyymmdd-images-0-01     6200   30000   30000
  ;; xxwiki-yyyymmdd-images-0-02     6400   60000   30000
  ;; ...
  ;; xxwiki-yyyymmdd-images-4-45    19800 2070000   30000
  ;; xxwiki-yyyymmdd-images-4-46    20000 2100000   30000
  ;; xxwiki-yyyymmdd-images-4-47    20000 2130000   29000
  ;; ...
  ;; xxwiki-yyyymmdd-images-f-ff      nil
  ;;
  ;; Also, if the number of images to be processed falls below some
  ;; threshold, we do nothing (i.e. just skip ahead to the next file),
  ;; the exception to this rule being the last file `...images-f-ff'
  ;; where we process any and all remaining images.
  ;;
  ;; If we are updating a large existing mirror, then we should
  ;; increase the increment.
  ;;
  (let* ((language      (sql-select-file-language file-name :put put))
	 (down-images   (shell-count-images :put put))
	 (import-images (sql-select-count-xxwiki-image language :put put))
         (valid-images  (sql-select-sum-images-pages language :put put))
	 (diff-images   (- down-images import-images))
	 (threshold     *mirror-rebuildimages-threshold*)
	 (threshold-p   (> diff-images threshold))
	 (index         (file-name-to-index file-name :put put))
	 (timeout       (min
			 (+ *mirror-rebuildimages-timeout-sec-min*
			    (* index *mirror-rebuildimages-timeout-sec-inc*))
			 *mirror-rebuildimages-timeout-sec-max*))
	 (last-file-p   (regexp:match "f-ff" file-name)))
    (put-message-value (_ "images down / import / valid") 
		       (format nil "~d / ~d / ~d" 
			       down-images import-images valid-images)
		       :put put)
    (if threshold-p
	(progn
	  (put-message-value (_ "threshold exceeded - rebuilding")
			     (format nil "d-i=~d > ~d" diff-images threshold)
			     :put put)
	  (put-message-value (_ "time out") 
			     (format nil "~d = min( ~d+(~d*~d), ~d )" 
				     timeout
				     *mirror-rebuildimages-timeout-sec-min*
				     index
				     *mirror-rebuildimages-timeout-sec-inc*
				     *mirror-rebuildimages-timeout-sec-max*)
			     :put put)
	  (shell-mediawiki-farm-rebuildimages file-name :timeout timeout
					      :put put))
      (put-message-value (_ "threshold not exceeded - skipping")
			 (format nil "d-i=~d <= ~d" diff-images threshold)
			 :put put))
    (when last-file-p
      (put-message (_ "rebuilding images - forced") :put put)
      (shell-mediawiki-farm-rebuildimages file-name :timeout nil :put put)))
  t)

(defun file-name-to-index (file-name &key (put nil))
  "Computing number from the last two chars of `xxwiki-yyyymmdd-images-y-yz'"
  (let ((s (first    ; `yz' where {y,z in [0-9a-f]}
	    (last
	     (regexp:regexp-split "-" file-name)))))
    (parse-integer s :radix 16)))

(defun fsm-images-validate (file-name &key (put nil))
  "Validating downloaded images, sequestering corrupt ones"
  ;;
  ;; Design note:
  ;;
  ;; Wikix generates `i-chunk's (image download scripts) that
  ;; incorrectly handle the case of partial downloads.
  ;; `http://upload.wikimedia.org/' has thousands of images that
  ;; download only part way, and then block.  Without a timeout, the
  ;; mirror would never finish.
  ;;
  ;; CURL can be configured with a timeout.  The option is
  ;; `-m/--max-time <seconds>'.  It is best to edit this option into
  ;; the shell scripts.
  ;;
  ;; However, when cURL times out, the `i-chunk's will then store the
  ;; partially downloaded (i.e. corrupt) image files just as if they
  ;; had been downloaded successfully.
  ;;
  ;; We must later hunt for these corrupt files and sequester them.
  ;; For this we use `gm identify -verbose [file]'.  For example,
  ;;
  ;;   /var/lib/mediawiki/images/3/35# gm identify -verbose \
  ;;   Model_of_ancient_Olympia.jpg
  ;;
  ;; When we look at the last line, it reads 
  ;;
  ;;   gm identify: Corrupt JPEG data: premature end of data segment
  ;;  (Model_of_ancient_Olympia.jpg).
  ;;
  ;; We must iterate over all images, sequestering those identified as
  ;; `Corrupt'.
  ;;
  ;; Design note:
  ;;
  ;; We sequester (rather than delete) bad image files, by moving them
  ;; to a `bad-images' directory.  First this lets us examine them
  ;; later.  Second, a programming error during development could
  ;; inadvertently sequester ALL image files.  Sequestration can be
  ;; reversed easily, whereas deletion can only be remedied at a cost
  ;; of redownloading 2T of image files (and say two months delay).
  ;;
  ;; Design note:
  ;;
  ;; We do not check the SVG files with `gm'.  Some of them cause
  ;; `gm' to hang.  Sometimes error messages and a back-trace is 
  ;; emitted, suggesting that malloc is involved.  Example files 
  ;; include: `/var/lib/mediawiki/images/0/00/IFK_Mariehamnin_logo.svg'
  ;; `/var/lib/mediawiki/images/0/00/Arms_of_Infante_Charles_of_Spain_as_King_of_Naples_and_Sicily.svg'
  ;;
  ;; We check some of the SVG files another way.  If an SVG file is
  ;; small and contains the text `302 Redirected', then it is actually
  ;; an error message returned by the proxy, and is sequestered.
  ;;
  ;; Not checking is unsatisfactory.  Someday we should explore ways
  ;; of letting `gm' timeout.
  ;;
  ;;
  (when *mirror-image-validate-p*
    (put-message (_ "validating images") :put put)
    (put-message "-----" :put put) 
    (let* ((language       (sql-select-file-language file-name :put put))
	   (name-split     (regexp:regexp-split "-" file-name))
	   (d1             (first (last name-split 2)))
	   (d2             (first (last name-split  )))
	   (dir-name-image (format nil "~a~a/~a/" 
				   *whereis-directory-images*
				   d1 d2))
	   (probe          (ext:probe-pathname dir-name-image)))
      ;; 0) probe directory 
      (if (null probe)
	  (put-message (_ "directory does not exist") :put put)
	(progn
	  ;; 1) sequester if file-name contains `%' char
	  (put-message-start (_ "looking for invalid names (with `%' char)") 
			     :put put)
	  (loop
	   for file-image of-type pathname in (shell-ls-grep dir-name-image "%"
							     :put put)
	   do (when (regexp:match "%" (file-namestring file-image))
		(put-message file-image :put put)
		(sequester-bad-image (format nil "~a~a" dir-name-image
					     file-image)
				     :put put)))
	  (put-message-done (_ "looking for invalid names (with `%' char)")
			    :put put)
	  ;; 2) sequester if file is actually an error message from proxy,
	  ;;    such files are small (< 500 bytes)
	  (put-message-start (_ "looking for files that are proxy error messages (`302 Redirected')")
			     :put put)
	  (loop
	   for path-image of-type pathname in (shell-find-1k dir-name-image
							     :put put)
	   do (when (and (ext:probe-pathname path-image)
			 (< (posix:file-size path-image) 1000)
			 (not (string= path-image dir-name-image)))
		(put-message path-image :put put)
		(unless (null (shell-cat-grep path-image "302 Redirected"
					      :put put))
		  (sequester-bad-image path-image :put put))))
	  (put-message-done (_ "looking for files that are proxy error messages (`302 Redirected')")
			    :put put)
	  ;; 3) sequester if `gm identify -verbose' calls file corrupt,
	  ;;    but skip validation if already imported into `xxwiki.image'
	  (put-message-start (_ "looking for corrupt image files") :put put)
	  (loop
	   with images-validated = 0
	   for path-image of-type pathname in (directory
					      (format nil "~a*" dir-name-image))
	   as  dummy0      = (debug-message path-image)
	   as  norm-string = (normalize-file-name (file-namestring path-image))
	   as  md5sum      = (string-to-md5sum norm-string :put put)
	   as  dummy3      = (debug-message md5sum)
	   as  md5-d1      = (subseq md5sum 0 1)
	   as  md5-d2      = (subseq md5sum 0 2)
	   as  validp      = (string= (sql-select-image-state md5sum :put put)
				      "valid")
	   as  result = (unless (or 
				 ;; skip if already validated
				 validp
				 ;; skip certain problem cases
				 (string= (pathname-type path-image) "svg")
				 (string= (pathname-type path-image) "SVG")
				 )
			  (shell-gm-identify path-image :put put))
	   do (put-message-value path-image result :put put)
	   ;; result = nil, usually means image file is OK
	   when (and (null result) (null validp))
	   do (progn
		(incf images-validated)
		(sql-update-file file-name
				 :pages images-validated          ; hack
				 :put put)
		(sql-insert-image md5sum :state 'valid' :put put))
	   ;; result = str, usually means either trouble or no decode delegate
	   when result
	   do (put-message result :put put)
	   and do (if (or
		       ;; usually caused by partial download
		       (regexp:match "Corrupt" result)
		       ;; caused by proxy `302 Redirected' error
		       (regexp:match "Improper image header" result) ;GIF,PNG
		       (regexp:match "Not a JPEG file" result)       ;JPG
		       (regexp:match "Not a TIFF or MDI file" result);TIF
		       (regexp:match "Opening and ending tag mismatch"
				     result)                         ;SVG
		       ;; caused by escaping (or not) the file name
		       (not (string= d1 md5-d1))
		       (not (string= d2 md5-d2))
		       )
		      ;; sequester (do not mark as validated)
		      (sequester-bad-image path-image :put put)
		    (progn
		      ;; keep OGG,PDF,WAV files (do not revalidate)
		      ;; `No decode delegate for this image format'
		      (put-message (_ "keeping file") :put put)
		      (incf images-validated)
		      (sql-update-file file-name
				       :pages images-validated          ; hack
				       :put put)
		      (sql-insert-image md5sum :state 'valid' :put put)
		      (put-message "-----" :put put))))
	  (put-message-done (_ "looking for corrupt image files") :put put)
	  ))))
  t)

(defun sequester-bad-image (pathname-old &key (put nil))
  "Moving bad-image to a bad-images directory"
  (let* ((pathname-bad (parse-namestring *whereis-directory-images-bad*))
	 ;; change directory only (keep host, device, name, type, version)
	 (pathname-new (merge-pathnames
			(make-pathname
			 :directory (append
				     (pathname-directory pathname-bad)
				     (last (pathname-directory pathname-old) 
					   2)))
			pathname-old)))  ; use as default
    (if (wild-pathname-p pathname-old)
	;; eliminate this branch when we learn how to deal with wildcards
	(progn
	  (put-message pathname-old :put put)
	  (put-message (_ "pathname contains a wildcard") :put put)
	  (put-message (_ "keeping file") :put put))
      (progn
	(put-message (_ "sequestering file to `bad-images' directory --->")
		     :put put)
	(put-message pathname-new :put put)
	;; we might have already sequestered this file
	(when (probe-file pathname-new)
	  (delete-file pathname-new))
	(rename-file pathname-old pathname-new)))
    (put-message "-----" :put put)
    ))

(defun fsm-no-op (file-name &key (put nil))
  "Doing nothing"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |don't care          none
  ;; state-in  |don't care             
  ;; state-out |no change               
  ;; Return: nil means fail, t means done.
  t) ; never fails

(defun assert-partition-free-images (&key (put nil))
  "Asserting image partition free space exceeds threshold"
  (let* ((partition-free-images (system-partition-free-images :put put))
	 (val (format
	       nil "~a/~a"
	       (format-integer-for-human partition-free-images)
	       (format-integer-for-human *system-partition-free-images-min*)))
	 (freep (> partition-free-images *system-partition-free-images-min*)))
    (if freep
	(put-message-value (_ "images space above threshold") val :put put)
      (put-message-value (_ "images space below threshold") 
			 (format nil "~a, skip image download" val) :put put))
    freep))

(defun assert-partition-free-innodb (&key (put nil))
  "Asserting InnoDB partition free space exceeds threshold"
  (let* ((partition-free-innodb (system-partition-free-innodb :put put))
	 (val (format
	       nil "~a/~a"
	       (format-integer-for-human partition-free-innodb)
	       (format-integer-for-human *system-partition-free-innodb-min*)))
	 (freep (> partition-free-innodb *system-partition-free-innodb-min*)))
    (if freep
	(put-message-value (_ "innodb space above threshold") val :put put)
      (put-message-value (_ "innodb space below threshold") 
			 (format nil "~a, skip page import" val) :put put))
    freep))

(defun fsm-process-file (file-name &key (put nil))
  "Processing file with finite-state-machine"
  (let ((message (_ "processing file with finite-state-machine"))
	(file-type (sql-select-file-type  file-name :put put))
	(old-state (sql-select-file-state file-name :put put))
	(new-state "error")
	(input     "error")
	(fsm-func  nil)
	(partition-free-images-p nil)
	(partition-free-innodb-p nil))
    (put-timestamp-message-start message :put put)
    (debug-message file-name)
    ;; 1) give FSM the `start' input, store new `state', increment `updates'
    (setq input "start")
    (setq new-state (fsm-transition old-state input :put put))
    (sql-update-file-state file-name new-state :put nil)
    (setq old-state new-state)
    ;; 2) process file with appropriate function,
    ;;    but fail if inadequate disk space for given function
    (setq fsm-func (fsm-function file-type new-state :put put))
    (setq partition-free-images-p 
	  (if (member fsm-func *fsm-function-needs-partition-free-images*)
	      (assert-partition-free-images :put put)
	    t))
    (setq partition-free-innodb-p 
	  (if (member fsm-func *fsm-function-needs-partition-free-innodb*)
	      (assert-partition-free-innodb :put put)
	    t))
    (if (and
	 partition-free-images-p
	 partition-free-innodb-p
	 (log-test fsm-func file-name :put put))                  ; <-- apply
	(setq input "done")
      (setq input "fail"))
    ;; 3) give FSM the `done' or `fail' input, and store new state
    (setq new-state (fsm-transition old-state input :put put))
    (unless (string= old-state new-state)
      (sql-update-file-state file-name new-state :put put))
    (put-timestamp-message-done message :put put)
    ;; 4) we should routinely delete some rows from the `objectcache'
    (when (member fsm-func *fsm-function-needs-delete-objectcache*)
      (sql-delete-from-objectcache
       (sql-select-file-language file-name :put put)))
  ))

(defun fsm-transition (state input &key (put nil))
  "Changing file state according to *state-transition* table"
  (debug-message-value (_ "initial state and input")
		       (format nil "~a + ~a" state input))
  (unless (member state *state* :test #'string=)
    (put-message-value-fail-and-abort (_ "Error: FSM invalid state") state))
  (unless (member input *input* :test #'string=)
    (put-message-value-fail-and-abort (_ "Error: FSM invalid input") input))
  (dolist (transition *state-transition* nil)
    (when (and (string= state (first transition))
	       (string= input (second transition)))
      (progn
	(put-message-value (_ "state transition") 
			   (format nil "~a + ~a --> ~a" 
				   (first  transition)
				   (second transition)
				   (third  transition))
			   :put put)
	(return (third transition)))))) ; break loop

(defun fsm-working-chown (file-name &key (put nil))  
  "Setting working directory ownership"
  (null (shell-chown-dir *whereis-directory-working* :put put)))


;;;;--------------------------------------------------------------------------+
;;;; Log utilities:                                                           |
;;;;   Note:  `put-message-*' functions should call `put-and-log-string'.     |
;;;;   Note:  Do not use `dribble', which is intended for logging interactive |
;;;;          sessions, and turns out to be non-portable.                     |
;;;;--------------------------------------------------------------------------+

		     
(defun log-start (&key (put nil))
  "Starting log"
  (put-message-value (_ "creating") *whereis-file-log* :put put)
  (defparameter *log-enabled-p* t)
  (put-timestamp-message-start "Starting log" :put put)
  t)

(defun log-stop (&key (put nil))
  "Stopping log"
  (put-message-value (_ "closing") *whereis-file-log* :put put)
  (put-timestamp-message-done "stopping log" :put put)
  (defparameter *log-enabled-p* nil)
  t)

(defun format-color-flag (flag)
  "Coloring text with ANSI Escape Sequences"
  (let ((clear    (format nil "~c[0m"  #\Escape))
	(up       (format nil "~c[1A"  #\Escape))
	(black    (format nil "~c[30m" #\Escape))
	(red      (format nil "~c[31m" #\Escape))
	(green    (format nil "~c[32m" #\Escape))
	(yellow   (format nil "~c[33m" #\Escape))
	(blue     (format nil "~c[34m" #\Escape))
	(magenta  (format nil "~c[35m" #\Escape))
	(cyan     (format nil "~c[36m" #\Escape))
	(white    (format nil "~c[37m" #\Escape)))
    (case flag
	  (:up    (format nil "~a" up))
	  (:fail  (format nil "[~afail~a]" red    clear))
	  (:info  (format nil "[~ainfo~a]" cyan   clear))
	  (:ok    (format nil "[~a ok ~a]" green  clear))
	  (:start (format nil "[~a....~a]" white  clear))
	  (:warn  (format nil "[~awarn~a]" yellow clear))
	  (otherwise nil)
	  )))

(defun put-flag-message (flag message &key (up nil))
  "Putting colored flag followed by message"
  (let* ((to-log    (concatenate 'string
				 (format-color-flag flag) 
				 message))
	 (to-term   (concatenate 'string
				(unless (null up) (format-color-flag :up))
				to-log)))
    (format *standard-output* "~a~%" to-term)
    (put-message to-log :put nil)))

(defun put-flag-message-value (flag message value &key (up nil))
  "Putting colored flag followed by message and value"
  (let* ((to-log    (concatenate 'string
				 (format-color-flag flag) 
				 message))
	 (to-term   (concatenate 'string
				(unless (null up) (format-color-flag :up))
				to-log)))
    (format *standard-output* "~34a: ~a~%" to-term value)
    (put-message-value to-log value :put nil)))

(defun log-test (f &rest args)
  "Wrapping evaluation of a boolean valued function with logging"
  (let* ((put        nil)
	 (doc        (_ (documentation f 'function)))
	 ;; use (cs-cl:symbol-name f) in modern case-sensitive world
	 (warn-p     (string= "warn" (cs-cl:symbol-name f) :end2 4))
	 (first-args (butlast args 2))
	 (message    (format nil "~a~{ ~a~}" f first-args)))
    (put-timestamp-message-start doc :put put)
    (put-flag-message :start message)
    (let ((res (apply f args)))                                 ; <-- apply 
      (if res
	  (progn
	    (put-timestamp-message-done doc :put put)
	    (put-flag-message :ok message :up t))
	(progn
	  (put-message (gethash f *error-messages*) :put put)
	  (put-timestamp-message-fail doc :put put)
	  (if (null warn-p)
	      (put-flag-message :fail message :up t)
	    (put-flag-message :warn message :up t))))
      res)))

(defun log-test-set (function-list &rest args)
  "Running tests and logging result"
  (let ((put      nil)
	(message  (_ "logging test set"))
	(all-pass t))
    (put-message-value function-list args :put put)
    (put-timestamp-message-start message :put put)
    (dolist (f function-list all-pass)
      (setq all-pass (and (apply #'log-test f args) all-pass))) ; <-- apply
    (if all-pass
	(put-timestamp-message-done message :put put)
      (put-timestamp-message-fail message :put put))
    all-pass))

(defun log-test-set-or-die (function-list &rest args)
  "Running tests, logging results, and die if any fail"
  (unless (apply #'log-test-set function-list args)
    (abort-and-die)))

(defun formatted-timestamp ()
  "Formatting decoded universal time as a human readable string"
  (multiple-value-bind (s m h dd mm yyyy dow dst tz)
      (get-decoded-time)
    (format nil "~4d-~2,'0d-~2,'0d ~2,'0d:~2,'0d:~2,'0d" yyyy mm dd h m s)))

(defun put-timestamp-message-start (message &key (put nil))
  "Logging timestamped message"
  (put-message-start (format nil "~a ~a" (formatted-timestamp) message)
		     :put put))

(defun put-timestamp-message-done (message &key (put nil))
  "Logging timestamped message done"
  (put-message-done (format nil "~a ~a" (formatted-timestamp) message)
		    :put put))

(defun put-timestamp-message-fail (message &key (put nil))
  "Logging timestamped message fail"
  (put-message-fail (format nil "~a ~a" (formatted-timestamp) message) 
		    :put put))

(defun put-timestamp-message-fail-and-abort (message &key (put nil))
  "Logging timestamped message fail and abort"
  (put-message-fail-and-abort 
   (format nil "~a ~a" (formatted-timestamp) message) :put put))

;;;;--------------------------------------------------------------------------+
;;;; Mirror mode:                                                             |
;;;;   0) Initialize parameters (command line, configuration file)            |
;;;;   1) Assert prerequisites                                                |
;;;;   2) Run in mirror mode                                                  |
;;;;   2.1) download checksums and dump file, validate, and decompress        |
;;;;   2.2) split dump into x-chunks                                          |
;;;;   2.3) generate i-chunks                                                 |
;;;;   2.4) run i-chunks (download images)                                    |
;;;;   2.5) validate downloaded image files, sequester corrupt ones           |
;;;;   2.6) import x-chunks (import pages)                                    |
;;;;--------------------------------------------------------------------------+

(defun language-code-normalized (language-code &key (put nil))
  "Replacing dashes with underscores"
  ;; replace `zh-classical' with `zh_classical'
  (substitute #\_ #\- (string-trim " " language-code)))

(defun language-code-to-database-name (language-code &key (put nil))
  "Mapping language code to database name"
  (if (string= language-code "template")
      *db-name*
    (format nil "~awiki" (language-code-normalized language-code :put put))))

(defun database-name-to-database-dump-file-name (database-name &key (put nil))
  "Mapping database name to database dump file name"
  (if (string= database-name *db-name*)
      *whereis-file-mediawiki-farm-database*
    (format nil "~a.sql" database-name)))

(defun language-code-to-database-dump-file-name (language-code &key (put nil))
  "Mapping language code to database dump file name"
  (let (database-name (language-code-to-database-name language-code :put put))
    (database-name-to-database-dump-file-name database-name :put put)))

(defun add-language-and-die (language-code &key (put nil))
  "Add a given language"
  ;; 2) Add mode
  (put-message (_ "-----add-mode-begin---------------------------") :put t)
  (put-message-start (_ "adding language") :put put)
  (log-test-set '(fsm-boot) :put put)
  (put-message-value-done (_ "adding language") language-code :put put)
  (put-message (_ "-----add-mode-done----------------------------") :put nil))

(defun language-code-to-checksums-url (language-code &key (put nil))
  "Generating URL of wikimedia checksums file for a given language"
  (let ((message               (_ "generating URL for checksums file"))
	(replace-this-language "xx")
	(with-this-language    (language-code-normalized language-code 
							 :put put))
	(url                   nil))
    (put-message-start message :put put)
    (put-message-value (_ "*wikimedia-site*")      *wikimedia-site* :put put)
    (put-message-value (_ "*wikimedia-checksums-template*")
		       *wikimedia-checksums-template* :put put)
    (when (regexp:match replace-this-language *wikimedia-checksums-template*)
      (setq url (format nil "~a~a" *wikimedia-site*
			(join
			 (regexp:regexp-split replace-this-language
					      *wikimedia-checksums-template*)
			 with-this-language))))
    (put-message url :put put)
    (put-message-value-done message language-code :put put)
    url))

(defun language-code-to-dump-url (language-code &key (put nil))
  "Generating URL of wikimedia dump file for a given language"
  (let* ((message               (_ "generating URL for dump file"))
	 (dummy-date            "yyyymmdd")
	 (replace-this-language "xx")
	 (with-this-language    (language-code-normalized        language-code))
	 (file-name             (language-code-to-dump-file-name language-code))
	 (file-date             (sql-select-file-date file-name))
	 (url                   nil))
    (put-message-start message :put put)
    (put-message-value (_ "*wikimedia-site*")      *wikimedia-site* :put put)
    (put-message-value (_ "*wikimedia-dump-template*")
		       *wikimedia-dump-template* :put put)
    ;; replace language
    (when (regexp:match replace-this-language *wikimedia-dump-template*)
      (setq url (format nil "~a~a" *wikimedia-site*
			(join
			 (regexp:regexp-split replace-this-language
					      *wikimedia-dump-template*)
			 with-this-language))))
    ;; replace date
    (when (regexp:match dummy-date url)
      (setq url    	(join
			 (regexp:regexp-split dummy-date url)
			 file-date)))
    (put-message url :put put)
    (put-message-value-done message language-code :put put)
    url))

(defun language-code-to-checksums-file-name (language-code &key (put nil))
  "Generating file-name of checksums file for a given language"
  (put-message-start (_ "generating checksums file-name") :put put)
  (let* ((replace-this-language  "xx")
	 (with-this-language     (language-code-normalized language-code
							   :put put))
	 (file-name (join (regexp:regexp-split 
			   replace-this-language
			   *whereis-file-checksums-template*)
			  with-this-language)))
    (put-message-value language-code file-name :put put)
    (put-message-done (_ "generating checksums file-name") :put put)
    file-name))

(defun language-code-to-dump-file-name (language-code &key (put nil))
  "Generating file-name of dump file for a given language-code"
  (sql-select-dump-name language-code :put put))

(defun language-code-to-checksums-path (language-code &key (put nil))
  "Generating pathname of checksums file for a given language-code"
  (put-message-start (_ "generating checksums pathname"))
  (let* ((file-name  (language-code-to-checksums-file-name language-code))
	 (path-name  (merge-pathnames
		      (parse-namestring *whereis-directory-working*)
		      (parse-namestring file-name))))
    (put-message path-name :put put)
    (put-message-done (_ "generating checksums pathname") :put put)
    path-name))

(defun language-code-to-dump-path (language-code &key (put nil))
  "Generating pathname of dump file for a given language-code"
  (put-message-start (_ "generating dump pathname") :put put)
  (let* ((file-name  (language-code-to-dump-file-name language-code :put put))
	 (path-name  (merge-pathnames
		      (parse-namestring *whereis-directory-working*)
		      (parse-namestring file-name))))
    (put-message path-name :put put)
    (put-message-done (_ "generating dump pathname") :put put)
    path-name))

(defun delete-files-and-state-language-and-die (language-code &key (put nil))
  "Deleting files and state info, for a given language-code"
  ;; 2) Delete mode
  (put-message (_ "-----delete-mode-begin------------------------") :put t)
  (put-message-start (_ "deleting language") :put put)
  (log-test-set '(delete-working-files-for-language-code
		  delete-state-for-language-code)
		language-code :put put)
  (put-message-value-done (_ "deleting language") language-code :put put)
  (put-message (_ "-----delete-mode-done-------------------------") :put nil))

(defun delete-directory-wp (path-name &key (put nil))
  "Delete directory"
  (when (ext:probe-pathname path-name)
    (shell-remove-directory path-name :put put)
    ;; may be terabytes, millions of files
    (sleep-while 5 #'ext:probe-pathname path-name))
  ;; confirm the directory is gone
  (not (ext:probe-pathname path-name)))

(defun delete-directory-images (&key (put nil))
  "Delete all images, math, thumbs, and working files"
  (delete-directory-wp *whereis-directory-images*       :put put))

(defun delete-directory-images-bad (&key (put nil))
  "Delete bad images directory"
  (delete-directory-wp *whereis-directory-images-bad*   :put put))

(defun delete-directory-images-math (&key (put nil))
  "Delete math directory"
  (delete-directory-wp *whereis-directory-images-math*  :put put))

(defun delete-directory-images-thumb (&key (put nil))
  "Delete thumb directory"
  (delete-directory-wp *whereis-directory-images-thumb* :put put))

(defun delete-directory-working (&key (put nil))
  "Delete working directory"
  (delete-directory-wp *whereis-directory-working*      :put put))

(defun delete-state-for-language-code (language-code &key (put nil))
  "Deleting state information for given language-code"
  (null (sql-delete-language language-code :put put)))

(defun delete-working-files-for-language-code (language-code &key (put nil))
  "Removing working files for given language-code"
  (shell-remove-all-working-files language-code :put put))

(defun drop-database-p (database-name &key (put nil))
  "Dropping database and confirming"
  (let ((message-drop-database (_ "dropping database"))
	(message-no-database   (_ "database not found"))
	(result                nil)) ; t - database dropped
    (put-message-start message-drop-database :put put)
    (if (assert-database-p database-name :put put)
	(progn
	  (sql-drop-database database-name :put put)
	  ;; dropping can take time
	  (sleep-while 5 #'assert-database-p database-name :put put))
      (put-message-value message-no-database database-name :put put))
    (setq result (null (assert-database-p database-name :put put)))
    ;; confirm that the database is gone
    (if result
	(put-message-value-done message-drop-database database-name :put put)
      (put-message-value-fail message-drop-database database-name))
    result))

(defun drop-database-for-language-code-p (language-code &key (put nil))
  "Dropping database and confirming"
  (let ((database-name (language-code-to-database-name language-code :put put)))
    (drop-database-p database-name :put put)))

(defun delete-interwiki-links (language-code &key (put nil))
  "Deleting interwiki links (interlanguage links actually)"
  (loop
   for language-code-2 in (mirror-language-list :put put)
   as  database-name = (language-code-to-database-name language-code-2 :put put)
   when (assert-database-p database-name :put put)
   do (sql-delete-interwiki-link database-name language-code :put put))
  t)

(defun drop-language-and-die (language-code &key (put nil))
  "Dropping database, files, and state info, for a given language"
  ;; 2) Drop mode
  (put-message (_ "-----drop-mode-begin--------------------------") :put t)
  (put-message-start (_ "dropping language") :put put)
  (log-test-set '(delete-interwiki-links
		  delete-state-for-language-code
		  drop-database-for-language-code-p
		  delete-working-files-for-language-code)
		language-code :put put)
  (put-message-value-done (_ "dropping language") language-code :put put)
  (put-message (_ "-----drop-mode-done---------------------------") :put nil))

(defun dump-database-p (database-name &key (put nil))
  "Dumping database and confirming"
  (let* ((message-dump-database (_ "dumping database"))
	 (message-no-database   (_ "database not found"))
	 (file-name             (database-name-to-database-dump-file-name
				 database-name :put put))
	 (path-name             (merge-pathnames 
				 (parse-namestring *whereis-directory-working*)
				 (parse-namestring file-name)))
	 (result                nil)) ; t - database dumped
    (put-message-start message-dump-database :put put)
    (if (assert-database-p database-name :put put)
	(progn
	  (shell-mysqldump-to-file database-name path-name :put put)
	  ;; dumping can take time
	  (sleep-until 5 #'assert-database-dump-p database-name :put put))
      (put-message-value message-no-database database-name :put put))
    (setq result (assert-database-dump-p database-name :put put))
    ;; confirm that the database is dumped
    (if result
	(put-message-value-done message-dump-database database-name :put put)
      (put-message-value-fail message-dump-database database-name))
    result))

(defun dump-database-for-language-code-p (language-code &key (put nil))
  "Dumping database and confirming"
  (let ((database-name (language-code-to-database-name language-code :put put)))
    (dump-database-p database-name :put put)))

(defun dump-database-language-and-die (language-code &key (put nil))
  "Dump database to file `xxwiki.sql' in working directory for a given language"
  ;; 2) Dump mode
  (put-message (_ "-----dump-mode-begin--------------------------") :put t)
  (put-message-start (_ "dropping language") :put put)
  (log-test-set '(dump-database-for-language-code-p)
		language-code :put put)
  (put-message-value-done (_ "dumping language") language-code :put put)
  (put-message (_ "-----dump-mode-done---------------------------") :put nil))

(defun update-database-for-language-code-p (language-code &key (put nil))
  "Updating database for given language code, and confirming"
  (let* ((database-name           (language-code-to-database-name language-code
								  :put put))
	 (message-update-database (_ "updating database"))
	 (message-no-database     (_ "database not found"))
	 (result                  nil)) ; t - database updated
    (put-message-start message-update-database :put put)
    (if (assert-database-p database-name :put put)
	(progn
	  (shell-mediawiki-farm-update language-code :put put)
	  ;; updating can take time
	  (sleep 2))
      (put-message-value message-no-database database-name :put put))
    (setq result (assert-database-p database-name :put put))
    ;; confirm that the database is updated
    (if result
	(put-message-value-done message-update-database database-name :put put)
      (put-message-value-fail message-update-database database-name))
    result))

(defun update-database-language-and-die (language-code &key (put nil))
  "Update database, to latest MediaWiki schema, for a given language"
  ;; 2) Update mode
  (put-message (_ "-----update-mode-begin------------------------") :put t)
  (put-message-start (_ "updating language") :put put)
  (log-test-set '(update-database-for-language-code-p)
		language-code :put put)
  (put-message-value-done (_ "updating language") language-code :put put)
  (put-message (_ "-----update-mode-done-------------------------") :put nil))

(defun restore-default-and-die (&key (put nil))
  "Dropping databases and files.  Starting over with default configuration"
  (put-message
      (_ "
+-----------------------------------------------------------------------------+
| WARNING  WARNING  WARNING   WARNING   WARNING   WARNING   WARNING   WARNING |
|                                                                             |
| This option may DELETE more than you expect or want:                        |
| 1)  Delete config files      : /etc/mediawiki/LocalSettings.php             |
|                              : /etc/mediawiki/LocalSettings_wpmirror.php    |
|                              : /etc/mysql/conf.d/wp-mirror.cnf              |
|                              : /etc/wpmirror/local.conf                     |
|                              : /etc/wpmirror/default.conf                   |
|     Delete database template : /usr/share/mediawiki/mai../database_farm.sql |
| 2)  Delete image files       : /var/lib/mediawiki/images/bad-images/        |
|                              : /var/lib/mediawiki/images/math/              |
|                              : /var/lib/mediawiki/images/thumb/             |
| 3)  Delete working files     : /var/lib/mediawiki/images/wp-mirror/         |
| 4)  Drop databases           : wikidb, *wiki, wpmirror                      |
|     Drop database users      : wikiadmin, wikiuser                          |
| The original default configuration is restored (mirror of `simple' wiki).   |
|                                                                             |
| WARNING  WARNING  WARNING   WARNING   WARNING   WARNING   WARNING   WARNING |
+-----------------------------------------------------------------------------+
") :put t)
  (when (yes-or-no-p "Do you wish to continue")
    (put-message (_ "-----restore-default-mode-begin---------------") :put t)
    ;; 1) drop databases: `wikidb', `xxwiki', and `wpmirror'
    (loop
     for database-name in (sql-show-databases-wpmirror :user *db-debian-user*
						       :put put)
     do (log-test-set '(drop-database-p) database-name :put put))
    ;; 2) drop database users: `wikiuser' and `wikiadmin'
    (log-test-set '(assert-dbms-drop-accounts-p)       :put put)
    ;; 3) delete image directory: `/var/lib/mediawiki/images/'
    (log-test-set '(delete-directory-images-bad
		    delete-directory-images-math
		    delete-directory-images-thumb
		    delete-directory-working) :put put)
    ;;    delete a few stray files
    (when (ext:probe-directory *whereis-directory-images*)
      (ext:cd *whereis-directory-images*)
      (shell-remove-file "*log"                                       :put put)
      (shell-remove-file "*sh"                                        :put put))
    ;; 4) delete all config files
    (loop
     for (dir-symbol file-symbol) in *wpmirror-config-delete-list* 
     as dir-path     = (symbol-value dir-symbol)
     as file-name    = (symbol-value file-symbol)
     as dummy        = (put-message-value dir-path file-name :put put)
     as path-name    = (merge-pathnames (parse-namestring dir-path)
					(parse-namestring file-name))
     when (file-exists-p path-name :put put)
     do (put-flag-message-value :info (_ "deleting config file") path-name)
     and do (shell-remove-file path-name :put put)
     finally
       (shell-mysql-restart :put put)
       (put-flag-message :info (_ "done")))
    (put-message (_ "-----restore-default-mode-done----------------") :put nil)))

(defun dump-file-remove-p (language-code &key (put nil))
  "Removing any old dump files and state info"
  (put-message-start (_ "deleting old files and state info") :put put)
  (shell-remove-old-working-files language-code :put put)
  (put-message-start (_ "deleting state info") :put put)
  (sql-delete-language-old language-code :put put)
  (put-message-value-done (_ "deleting state info") language-code :put put)
  (put-message-value-done (_ "deleting old files and state info") language-code
			  :put put))

(defun ichunks-complete-p (file-name &key (put nil))
  "Determining if all `ichunk's are all processed"
  ;; if no `images' records exist, and
  ;; if no `ichunk's remain to be processed,
  ;; then insert `xxwiki-yyyymmdd-images-x-xx' records into the
  ;; `wpmirror.file' table
  (put-message-start (_ "ichunks-complete-p") :put put)
  (let* ((language  (sql-select-file-language file-name :put put))
	 (date      (sql-select-file-date     file-name :put put))
	 (result (and (zerop (sql-select-count-images  language :put put))
		      (zerop (sql-select-count-ichunk-remaining
			      language :put put))
		      (> (sql-select-count-ichunk-done language :put put)
			 0))))
    (when result
      (put-message-value-done (_ "ichunks complete") result :put put)
      (put-message-start (_ "inserting `images' records"))
      (loop named images-validate
	with hex = (the simple-string "0123456789abcdef")
	for c1 of-type standard-char across hex
	do
	(loop
	  for c2 of-type standard-char across hex
	  do
	  (sql-insert-file language date
			   (format nil "~a-~a-images-~a-~a~a"
				   (language-code-to-database-name language)
				   date c1 c1 c2)
			   "images" :put nil))
	(sleep 1))
      (put-message-done (_ "inserting `images' records")))
    result))

(defun ichunk-edit-p (ichunk-name)
  "Editing `ichunk' shell scripts"
  ;;
  ;; Design note:
  ;;
  ;; `Wikix' has an option `-p' (meaning parallel) that lets one
  ;; generate 16 shell scripts which may then be run concurrently.
  ;; In these scripts, `curl' is always invoked with a `--retry 7'
  ;; option.  This makes sense when trying to run so many scripts
  ;; concurrently.  However, `wp-mirror' runs its scripts
  ;; sequentially, so `--retry 0' is better.
  ;;
  (let ((temp-file (concatenate 'string ichunk-name ".temp"))
	(replace-this-string "--retry 7")     ; gain speed with fewer retries
	(with-this-string                     ; 0 or 1 retries are good
	 (format nil "--retry ~d" *ichunk-curl-retry*)))
    (with-open-file (f ichunk-name
		       :direction :input
		       :if-does-not-exist nil)
      (with-open-file (g temp-file
			 :direction :output
			 :if-exists :supersede)
	(do ((l (read-line f nil 'eof) (read-line f nil 'eof)))
	    ((eq l 'eof) nil)
	  (cond ((regexp:match replace-this-string l)
		 (write-line (join (regexp:regexp-split replace-this-string l)
				   with-this-string) g))
		(t (write-line l g))))))
    (null (shell-rename-file temp-file (file-namestring ichunk-name)))))

(defun join (list-of-strings join-string)
  "Joining strings"
  (let ((n (length list-of-strings)))
    (cond ((eql n 0) '())
	  ((eql n 1) (first list-of-strings))
	  ((eql n 2) (concatenate 'string
				  (first list-of-strings)
				  join-string
				  (second list-of-strings)))
	  (t (concatenate 'string
			  (first list-of-strings)
			  join-string
			  (second list-of-strings)
			  join-string
			  (join (cddr list-of-strings) join-string))))))

(defun wait-for-internet-access (file-name &key (put nil))
  "Waiting for internet access, if processing the file involves a download"
  ;; We assume lock has been grabbed
  (let* ((message (_ "wait for internet access, if download required"))
	 (file-type (sql-select-file-type  file-name :put put))
	 (old-state (sql-select-file-state file-name :put put))
	 (input     "start")
	 (new-state (fsm-transition old-state input     :put put))
	 (fsm-func  (fsm-function   file-type new-state :put put)))
    ;; 0) when a download is required, wait for internet access.
    ;;    sometimes your Internet Service Provider (ISP) will go down
    ;;    for a few hours during the night
    (when (member fsm-func *fsm-function-needs-internet-access*)
      (put-message message)
      (loop named wait-access
	;; actually this will not detect a failure, if your traffic is
	;; going through a caching web proxy, and the failure lies
	;; between the proxy and wikimedia
	until (assert-internet-access-to-wikimedia-site-p :put put)
	do
	(put-message-value (_ "no access to wikipedia - sleep")
			   *monitor-poll-sec*)
	(sleep *monitor-poll-sec*)))))

(defun main-mirror (&key (put t))
  "Purpose: Main function when wp-mirror is in Mirror Mode."
  ;; 2) Mirror mode
  (put-message (_ "-----mirror-mode-begin------------------------") :put t)
  ;; load initial record into finite-state-machine
  (when (eql *main-mode* :first-mirror)
    (log-test-set '(fsm-boot
		    release-all-file-semaphores) :put put))
  (loop named mirror-pass-multiple
    for i from 0 to 1
    ;; first pass does all the work
    ;; second pass retries any failed `xchunks' and `ichunks'
    while (first-mirror-exists-p :put put) ; all mirrors exit if the first exits
    do
    ;;(when (and (zerop i) (i-am-the-first-mirror-p :put put))
    (when (and (zerop i) (eql *main-mode* :first-mirror))
      (sql-update-fail-to-start :put put))   ; give failed dump a second chance
    (loop named mirror-pass-single
      ;;   2a) decide order in which files should be processed,
      ;;       select one, and grab its lock (set its semaphore to 0)
      for file-name of-type string = (sql-select-next-file :put put)
      while (and (first-mirror-exists-p :put put)
					 ; all mirrors exit if the first does
		 file-name)              ; if no file found, exit loop
        do
          ;; 2b) wait if necessary
          (wait-for-internet-access file-name :put put)
	  ;; 2c) process the file
	  ;;(put-message-value (_ "file before") 
	  ;;		     (sql-select-file file-name :put put) :put put)
	  (fsm-process-file file-name :put put) ; <--- heavy lifting
	  ;; 2d) release its lock (set semaphore to 1)
	  (sql-update-file-semaphore file-name 1 :put put)
	  ;;(put-message-value (_ "file after") 
	  ;;		       (sql-select-file file-name :put put) :put put)
	  ;; 2e) see if `ichunk's all done, if so generate `images'
	  (ichunks-complete-p file-name :put put)
	  (put-message "----------" :put put)
	  ;;(die) ; uncomment to `single-step' for debug purposes
	  )
    ;;(when (and (zerop i) (i-am-the-first-mirror-p :put put))
    (when (and (zerop i) (eql *main-mode* :first-mirror))
      (sql-update-fail-to-pending :put put) ; give failed chunks a second chance
      ;;(sql-update-images-to-pending) ; run `chown' again
      ))
  (put-message (_ "-----mirror-mode-done-------------------------") :put put))


;;;;--------------------------------------------------------------------------+
;;;; Monitor mode:                                                            |
;;;;   0) Initialize parameters (command line, configuration file)            |
;;;;   1) Assert prerequisites                                                |
;;;;   2) Determine monitor mode (:auto, :gui, :screen, :text)                |
;;;;--------------------------------------------------------------------------+


(defvar       *report*                  nil)

(defstruct report
  ;; timestamp
  timestamp
  ;; wiki
  language
  database-name
  ;; activities (boolean)
  dump-file-null
  dump-file-downloading
  dump-file-validating
  dump-file-decompressing
  dump-file-page-counting
  dump-file-image-counting
  xchunk-generating
  ichunk-generating
  images-counting
  images-validating
  ;; downloading
  file-dump-name
  file-dump-size
  file-dump-xml-name
  file-dump-xml-size
  ;; estimating
  page-count
  image-count
  chunk-count
  ;; chunk generating
  xchunks-generated
  xchunks-generated-progress
  ichunks-generated
  ichunks-generated-progress
  ;; ichunks
  ichunks-running
  ichunks-done
  ichunks-failed
  ichunks-remaining
  ichunks-progress
  images-count
  images-start
  images-created
  images-valid
  images-imported
  images-sum
  images-count-progress
  images-valid-progress
  images-imported-progress
  ;; xchunks
  xchunks-running
  xchunks-done
  xchunks-failed
  xchunks-remaining
  xchunks-progress
  pages-imported
  ;; InnoDB
  innodb-buffer-pool-pages-total
  innodb-buffer-pool-pages-dirty
  innodb-buffer-pool-pages-dirty-progress
  innodb-buffer-pool-pages-insert
  innodb-buffer-pool-pages-insert-progress
  innodb-buffer-pool-pages-misc
  innodb-buffer-pool-pages-misc-progress
  ;; System
  partition-images-free
  partition-images-size
  partition-images-free-to-size-ratio
  partition-innodb-free
  partition-innodb-size
  partition-innodb-free-to-size-ratio
)

(defun monitor-data-collect (lang &key (put nil))
  "Collecting status data suitable for monitors"
  (put-message-start (_ "collecting status data for monitor") :put put)
  (let* ((database-name    (language-code-to-database-name    lang :put put))
	 (file-dump-name   (sql-select-dump-name              lang :put put))
	 (file-dump-xml-name (sql-select-dump-xml-name        lang :put put))
	 (chunk-count      (ceiling (/ (sql-select-file-pages
					file-dump-xml-name :put put)
				       *xchunk-page-count*)))
	 (ichunk-count     (sql-select-count-ichunk           lang :put put))
	 (ichunk-done      (sql-select-count-ichunk-done      lang :put put))
	 (ichunk-fail      (sql-select-count-ichunk-fail      lang :put put))
	 (ichunk-remaining (sql-select-count-ichunk-remaining lang :put put))
	 (ichunk-running   (sql-select-count-ichunk-running   lang :put put))
	 (xchunk-count     (sql-select-count-xchunk           lang :put put))
	 (xchunk-done      (sql-select-count-xchunk-done      lang :put put))
	 (xchunk-fail      (sql-select-count-xchunk-fail      lang :put put))
	 (xchunk-remaining (sql-select-count-xchunk-remaining lang :put put))
	 (xchunk-running   (sql-select-count-xchunk-running   lang :put put))
	 (pages-imported   (if (assert-database-p    database-name :put put)
			       (sql-select-count-xxwiki-page  lang :put put)
			     0))
	 (images-count     (sql-select-count-images           lang :put put))
	 (images-start     (sql-select-count-images-start     lang :put put))
	 (images-created   (sql-select-count-images-created   lang :put put))
         (images-valid     (sql-select-count-image-valid           :put put))
	 ;; hack - storing number of validated images in the `pages' field
	 ;(images-valid     (sql-select-sum-images-pages       lang :put put))
	 ;(images-valid     (sql-select-count-images-valid     lang :put put))
 	 (images-imported  (if (assert-database-p    database-name :put put)
			       (sql-select-count-xxwiki-image lang :put put)
			     0))
	 (images-sum       (sql-select-sum-images-images      lang :put put))
	 (innodb-buffer-pool-pages-dirty 
	  (sql-show-innodb-buffer-pool-pages-dirty    :put put))
	 (innodb-buffer-pool-pages-insert
	  (sql-show-innodb-buffer-pool-pages-insert   :put put))
	 (innodb-buffer-pool-pages-misc
	  (sql-show-innodb-buffer-pool-pages-misc     :put put))
	 (innodb-buffer-pool-pages-total
	  (sql-show-innodb-buffer-pool-pages-total    :put put))
	 ;; System
	 (partition-images-free
	  (system-partition-free-images               :put put))
	 (partition-images-size
	  (system-partition-size-images               :put put))
	 (partition-innodb-free
	  (system-partition-free-innodb               :put put))
	 (partition-innodb-size
	  (system-partition-size-innodb               :put put)))
    (put-message-value (_ "chunk-count")     chunk-count         :put put)
    (put-message-value (_ "ichunk-count")    ichunk-count        :put put)
    (put-message-value (_ "ichunk-done")     ichunk-done         :put put)
    (put-message-value (_ "ichunk-fail")     ichunk-fail         :put put)
    (put-message-value (_ "xchunk-count")    xchunk-count        :put put)
    (put-message-value (_ "xchunk-done")     xchunk-done         :put put)
    (put-message-value (_ "xchunk-fail")     xchunk-fail         :put put)
    (put-message-value (_ "images-count")    images-count        :put put)
    (put-message-value (_ "images-start")    images-start        :put put)
    (put-message-value (_ "images-created")  images-created      :put put)
    (put-message-value (_ "images-valid")    images-valid        :put put)
    (put-message-value (_ "images-imported") images-imported     :put put)
    (put-message-value (_ "images-sum")      images-sum          :put put)
    (put-message-value (_ "innodb-buffer-pool-pages-dirty") 
		       innodb-buffer-pool-pages-dirty            :put put)
    (put-message-value (_ "innodb-buffer-pool-pages-insert") 
		       innodb-buffer-pool-pages-insert           :put put)
    (put-message-value (_ "innodb-buffer-pool-pages-misc") 
		       innodb-buffer-pool-pages-misc             :put put)
    (put-message-value (_ "innodb-buffer-pool-pages-total") 
		       innodb-buffer-pool-pages-total            :put put)
    (setq *report*
	(make-report
	 ;; timestamp
	 :timestamp	          (formatted-timestamp)
	 ;; wiki
	 :language                lang
	 :database-name           database-name
	 ;; activities (boolean)
	 :dump-file-null          (zerop (sql-select-count-file lang :put put))
	 :dump-file-downloading   (string= (sql-select-file-state
					    file-dump-name :put put)
					   "start")
	 :dump-file-validating    (string= (sql-select-file-state
					    file-dump-name :put put)
					   "created")
	 :dump-file-decompressing (string= (sql-select-file-state
					    file-dump-name :put put)
					   "valid")
	 :dump-file-page-counting (and
				   (string= (sql-select-file-state
					     file-dump-xml-name :put put)
					    "start")
				   (zerop (sql-select-file-pages
					   file-dump-xml-name :put put)))
	 :dump-file-image-counting (and
				    (string= (sql-select-file-state
					      file-dump-xml-name :put put)
					     "start")
				    (zerop (sql-select-file-images
					    file-dump-xml-name :put put)))
	 :xchunk-generating      (sql-select-xchunk-wip lang :put put)
	 :ichunk-generating      (sql-select-ichunk-wip lang :put put)
	 :images-counting        (not
				  (zerop
				   (sql-select-count-images-start lang
								  :put put)))
	 :images-validating      (not
				  (zerop
				   (sql-select-count-images-created lang
								    :put put)))
	 ;; downloading
	 :file-dump-name         file-dump-name
	 :file-dump-size         (sql-select-file-size file-dump-name :put put)
	 :file-dump-xml-name     file-dump-xml-name
	 :file-dump-xml-size     (sql-select-file-size file-dump-xml-name
						       :put put)
	 ;; estimating
	 :page-count             (sql-select-file-pages file-dump-xml-name
							:put put)
	 :image-count            (sql-select-file-images file-dump-xml-name
							 :put put)
	 :chunk-count            chunk-count
	 ;; chunk generating
	 :xchunks-generated      xchunk-count
	 :xchunks-generated-progress (cond ((zerop chunk-count) 0)
					   ((> xchunk-count chunk-count) 1.0)
					   (t (float
					       (/ xchunk-count chunk-count))))
	 :ichunks-generated      ichunk-count
	 :ichunks-generated-progress (cond ((zerop chunk-count) 0)
					   ((> ichunk-count chunk-count) 1.0)
					   (t (float
					       (/ ichunk-count chunk-count))))
	 ;; ichunks
	 :ichunks-running        ichunk-running
	 :ichunks-done           ichunk-done
	 :ichunks-failed         ichunk-fail
	 :ichunks-remaining      (- ichunk-count ichunk-done ichunk-fail)
	 :ichunks-progress       (cond ((zerop chunk-count) 0)
				       ((> (+ ichunk-done ichunk-fail)
					   chunk-count) 1.0)
				       (t (float 
					   (/ (+ ichunk-done ichunk-fail)
					      chunk-count))))
	 ;; xchunks
	 :xchunks-running        xchunk-running
	 :xchunks-done           xchunk-done
	 :xchunks-failed         xchunk-fail
	 :xchunks-remaining      (- xchunk-count xchunk-done xchunk-fail)
	 :xchunks-progress       (cond ((zerop chunk-count) 0)
				       ((> (+ xchunk-done xchunk-fail)
					   chunk-count) 1.0)
				       (t (float
					   (/ (+ xchunk-done xchunk-fail)
					      chunk-count))))
	 :pages-imported         pages-imported
	 ;; images
	 :images-sum                 images-sum
	 :images-count               images-count
	 :images-count-progress      (cond ((zerop images-count) 0)
					   ((> (- images-count images-start)
					       images-count) 1.0)
					   (t (float
					       (/ (- images-count images-start)
						  images-count))))
	 :images-valid               images-valid
	 :images-valid-progress      (cond ((zerop images-sum) 0)
					   ((> images-valid images-sum) 1.0)
					   (t (float (/ images-valid
							images-sum))))
	 :images-imported            images-imported
	 :images-imported-progress   (cond ((zerop images-sum) 0)
					   ((> images-imported images-sum) 1.0)
					   (t (float (/ images-imported
							images-sum))))
	 ;; InnoDB
	 :innodb-buffer-pool-pages-total innodb-buffer-pool-pages-total
	 :innodb-buffer-pool-pages-dirty innodb-buffer-pool-pages-dirty
	 :innodb-buffer-pool-pages-dirty-progress
	 (if (zerop innodb-buffer-pool-pages-total)
	     0
	   (float (/ innodb-buffer-pool-pages-dirty
		     innodb-buffer-pool-pages-total)))
	 :innodb-buffer-pool-pages-insert innodb-buffer-pool-pages-insert
	 :innodb-buffer-pool-pages-insert-progress 
	 (if (zerop innodb-buffer-pool-pages-total)
	     0
	   (float (/ innodb-buffer-pool-pages-insert 
		     innodb-buffer-pool-pages-total)))
	 :innodb-buffer-pool-pages-misc innodb-buffer-pool-pages-misc
	 :innodb-buffer-pool-pages-misc-progress 
	 (if (zerop innodb-buffer-pool-pages-total)
	     0
	   (float (/ innodb-buffer-pool-pages-misc
		     innodb-buffer-pool-pages-total)))
	 ;; System
	 :partition-images-free    partition-images-free
	 :partition-images-size    partition-images-size
	 :partition-images-free-to-size-ratio
	 (if (zerop partition-images-size)
	     0
	   (float (/ partition-images-free
		     partition-images-size)))
	 :partition-innodb-free    partition-innodb-free
	 :partition-innodb-size    partition-innodb-size
	 :partition-innodb-free-to-size-ratio
	 (if (zerop partition-innodb-size)
	     0
	   (float (/ partition-innodb-free
		     partition-innodb-size)))
       ))))

(defun windowing-system-running-p (&key (put nil))
  "Determining if a windowing system is running"
  (put-message-start (_ "probing running windowing system") :put put)
  (let ((windowing-system nil))
    (with-open-stream (s (ext:run-shell-command
			  (format nil "ps -wef | ~a ~a"
				  *whereis-grep* *whereis-x*)
			  :output :stream))
      (setq windowing-system (read-line s nil nil)))
    (if (null windowing-system)
	(put-message-fail (_ "probing running windowing system"))
      (put-message-value-done (_ "probing running windowing system")
			      windowing-system :put put))
    windowing-system))

(defun monitor-mode-gui-screen-or-text (&key (put nil))
  "Determining best monitor mode (gui, screen, or text)"
  (put-message-start (_ "choosing monitor mode") :put put)
  (put-message-value (_ "monitor mode") *monitor-mode* :put put)
  (when (eql *monitor-mode* :auto)
    (put-message-value (_ "probing for :gui") *monitor-mode* :put put)
    (if (and (member :clx *features*)
	     (windowing-system-running-p))
	(setq *monitor-mode* :gui)
      (progn
	(put-message-value (_ "probing for :screen") *monitor-mode* :put put)
	(if (member :screen *features*)
	    (setq *monitor-mode* :screen)
	  (progn
	    (put-message-value (_ "falling back on :text") *monitor-mode*
			       :put put)
	    (setq *monitor-mode* :text))))))
  (put-message-value-done (_ "choosing monitor mode") *monitor-mode* :put put)
  *monitor-mode*)

(defparameter *net-wm-icons* 
  (make-array (+ 2 (* 16 16) 2 (* 32 32)) 
	      :element-type '(unsigned-byte 32)
	      :initial-contents
	      '(16 16
4294901760 4294901760 4294901760 4294901760
4294901760 4294901760 4294901760 4294901760 338034905 3657433343 0 184483840
234881279 3053453567 3221225727 1879048447 0 0 0 0 0 0 0 1224737023
3305111807 3875537151 0 0 2063597823 1291845887 0 67109119 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
50266112 3422552319 0 0 3070230783 2063597823 2986344703 771752191 0 0 0
0 0 0 0 0 0 3422552319 0 0 3372220671 1509949695 704643327 3355443455
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4294901760 0 3422552319 0 134152192 3187671295 251658495 0 3439329535 0 0
0 0 0 0 0 0 0 3422552319 0 0 2332033279 1342177535 167772415 3338666239
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4294901760 0 3422552319 0 0 436207871 3322085628 3456106751 1375731967
4278255360 4026597120 3758161664 3489726208 3204513536 2952855296 2684419840
2399207168 2130771712 1845559040 1593900800 1308688128 1040252672 755040000
486604544 234946304 4278255360 4043374336 3774938880 3506503424 3221290752
2952855296 2667642624 2399207168 2130771712 1862336256 1627453957 1359017481
1073805064 788591627 503379721 218169088 4278255360 4043374336 3758161664
3506503424 3221290752 2952855296 2684419840 2415984384 2130771712 1862336256
1577123584 1308688128 1040252672 755040000 486604544 218169088 4278190335
4026532095 3758096639 3489661183 3221225727 2952790271 2667577599 2415919359
2130706687 1862271231 1593835775 1325400319 1056964863 771752191 520093951
234881279 4278190335 4026532095 3758096639 3489661183 3221225727 2952790271
2667577599 2415919359 2130706687 1862271231 1593835775 1325400319 1056964863
771752191 503316735 234881279 4278190335 4026532095 3758096639 3489661183
3221225727 2952790271 2684354815 2399142143 2130706687 1862271231 1593835775
1325400319 1040187647 771752191 520093951 234881279 4294901760 4043243520
3774808064 3506372608 3221159936 2952724480 2684289024 2399076352 2147418112
1862205440 1593769984 1308557312 1040121856 771686400 503250944 234815488
4294901760 4060020736 3758030848 3506372608 3221159936 2952724480 2684289024
2415853568 2130640896 1862205440 1593769984 1308557312 1040121856 771686400
503250944 234815488 4294901760 4043243520 3774808064 3489595392 3237937152
2952724480 2684289024 2415853568 2147418112 1862205440 1593769984 1325334528
1056899072 788463616 503250944 234815488
32 32
4294901760 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
0 0 0 0 0 0 0 0 0 268369920 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4294901760 1509949695 3120562431 4009754879 4194304255 3690987775 2130706687
83886335 0 50331903 1694499071 3170894079 3992977663 4211081471 3657433343
1879048447 16777471 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 3087007999
2281701631 1191182591 1040187647 2030043391 4127195391 2566914303 0 16777471
3254780159 2181038335 1191182591 973078783 2030043391 4177527039 2130706687
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4294901760 4294901760 0 0 0 0 0 2214592767 4093640959 0 0 0 0 0 0 0
2298478847 3909091583 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
0 2214592767 3607101695 0 0 0 0 0 0 0 1946157311 4093640959 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4294901760 0 0 536871167 1191182591 2281701631 3019899135 637534463 0 0 0
100597760 251592704 33488896 0 3321889023 2919235839 0 0 0 0 0 0 0 0 0
0 0 0 0 0 0 0 0 0 2550137087 4278190335 4278190335 3405775103 570425599
0 0 0 0 0 0 2046820607 4043309311 620757247 4294901760 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
33488896 0 0 218104063 1291845887 3841982719 3388997887 0 0 0 0 0
1996488959 4093640959 1073742079 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
0 0 0 0 0 0 1761607935 4278190335 150995199 0 0 67109119 2550137087
3909091583 889192703 0 0 4294901760 4294901760 4294901760 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 0 0 0 0 0
2181038335 3925868799 0 0 218104063 3070230783 3623878911 570425599 0 0 0
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 805306623 3288334591 1795162367
1040187647 1023410431 2231369983 4211081471 1694499071 0 369099007 3456106751
3825205503 1174405375 872415487 872415487 872415487 872415487 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4293984270 2046951677 3422552319 4110418175 4177527039 3405775103 1409286399
0 0 1409286399 4278190335 4278190335 4278190335 4278190335 4278190335
4278190335 4278190335 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
0 0 0 0 0 0 0 0 0 0 0 0 4294901760 4294901760 4294901760 4294901760
4294901760 4294901760 4294901760 4294901760 4294901760 4294901760 4294901760
4294901760 42 94901760 4294901760 4294901760 4294901760 0 0 0 0 0 0 0 0
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 ;0
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 4278255360 4144037632 4009819904
3875602176 3741384448 3607166720 3472948992 3338731264 3204513536 3053518592
2936078080 2801860352 2650865408 2516647680 2382429952 2264989440 2113994496
1996553984 1862336256 1728118528 1577123584 1459683072 1325465344 1191247616
1040252672 922812160 771817216 637599488 503381760 385941248 234946304
100728576 4278255360 4144037632 4009819904 3875602176 3724607232 3607166720
3472948992 3338731264 3204513536 3070295808 2936078080 2801860352 2667642624
2516647680 2399207168 2264989440 2130771712 1996553984 1845559040 1728118528
1593900800 1459683072 1308688128 1191247616 1057029888 922812160 788594432
637599488 503381760 369164032 234946304 117505792 4278255360 4144037632
4009819904 3875602176 3741384448 3607166720 3472948992 3338731264 3204513536
3053518592 2919300864 2801860352 2650865408 2533424896 2399207168 2264989440
2113994496 1996553984 1862336256 1728118528 1593900800 1459683072 1325465344
1191247616 1040252672 906034944 771817216 654376704 503381760 369164032
234946304 117505792 4278255360 4144037632 4009819904 3858824960 3741384448
3607166720 3472948992 3338731264 3204513536 3070295808 2936078080 2801860352
2667642624 2533424896 2382429952 2264989440 2130771712 1979776768 1862336256
1728118528 1577123584 1442905856 1325465344 1191247616 1040252672 922812160
771817216 637599488 503381760 369164032 234946304 100728576 4278255360
4144037632 4009819904 3875602176 3741384448 3607166720 3472948992 3338731264
3204513536 3070295808 2919300864 2801860352 2667642624 2533424896 2399207168
2264989440 2113994496 1996553984 1862336256 1728118528 1593900800 1442905856
1342241795 1174470400 1057029888 906034944 788594432 654376704 503381760
385941248 251723520 100728576 4278190335 4160749823 4026532095 3892314367
3741319423 3623878911 3472883967 3338666239 3221225727 3070230783 2952790271
2818572543 2667577599 2533359871 2399142143 2264924415 2147483903 1996488959
1862271231 1728053503 1593835775 1459618047 1325400319 1191182591 1056964863
922747135 788529407 654311679 520093951 385876223 251658495 117440767
4278190335 4160749823 4026532095 3892314367 3741319423 3623878911 3489661183
3355443455 3221225727 3087007999 2936013055 2801795327 2667577599 2533359871
2399142143 2281701631 2130706687 1996488959 1862271231 1728053503
1593835775 1459618047 1325400319 1191182591 1056964863 922747135 788529407
654311679 520093951 385876223 234881279 100663551 4278190335 4160749823
4026532095 3892314367 3758096639 3623878911 3489661183 3355443455 3221225727
3087007999 2936013055 2801795327 2667577599 2550137087 2415919359 2264924415
2130706687 1996488959 1862271231 1728053503 1593835775 1459618047 1325400319
1191182591 1056964863 922747135 788529407 654311679 503316735 369099007
251658495 100663551 4278190335 4160749823 4026532095 3892314367 3758096639
3623878911 3489661183 3355443455 3204448511 3087007999 2936013055 2818572543
2667577599 2533359871 2399142143 2264924415 2130706687 1996488959 1879048447
1728053503 1593835775 1459618047 1325400319 1191182591 1056964863 922747135
788529407 654311679 520093951 385876223 251658495 117440767 4278190335
4160749823 4026532095 3892314367 3758096639 3623878911 3489661183 3355443455
3221225727 3087007999 2952790271 2818572543 2667577599 2533359871 2399142143
2264924415 2147483903 2013266175 1862271231 1744830719 1610612991 1476395263
1342177535 1191182591 1056964863 922747135 788529407 654311679 520093951
385876223 251658495 100663551 4294901760 4160684032 4026466304 3909025792
3774808064 3623813120 3489595392 3355377664 3237937152 3103719424 2952724480
2818506752 2684289024 2550071296 2415853568 2281635840 2147418112 2013200384
1878982656 1744764928 1593769984 1476329472 1325334528 1207894016 1056899072
939458560 788463616 654245888 520028160 385810432 251592704 117374976
4294901760 4177461248 4043243520 3909025792 3774808064 3640590336 3506372608
3355377664 3221159936 3086942208 2952724480 2818506752 2701066240 2550071296
2415853568 2281635840 2147418112 2013200384 1878982656 1727987712 1610547200
1476329472 1325334528 1191116800 1073676288 922681344 788463616 654245888
520028160 385810432 251592704 100597760 4294901760 4177461248 4043243520
3909025792 3774808064 3640590336 3489595392 3372154880 3237937152 3103719424
2952724480 2818506752 2700935170 2550071296 2415853568 2281635840 2147418112
2013200384 1878982656 1744764928 1610547200 1459552256 1342111744 1191116800
1056899072 922681344 788463616 671023104 520028160 385810432 251592704
100597760 4294901760 4177461248 4043243520 3909025792 3774808064 3640590336
3489595392 3372154880 3237937152 3086942208 2969501696 2818506752 2684289024
2550071296 2432630784 2281635840 2147418112 2013200384 1862205440 1744764928
1610547200 1476329472 1342111744 1191116800 1056899072 922681344 788463616
654245888 520028160 385810432 251592704 117374976 4294901760 4177461248
4043243520 3909025792 3774808064 3623813120 3506372608 3372154880 3237937152
3103719424 2952724480 2835283968 2684289024 2550071296 2432630784 2281635840
2147418112 2046492676 1862205440 1744764928 1610547200 1476329472
1342111744 1207894016 1056899072 939458560 788463616 654245888 536281096
385810432 251592704 134152192)))

(defun monitor-mode-gui (&key (put nil))
  "Running the monitor in gui mode"
  (put-message-start (_ "attempting gui mode") :put put)
  (let* ((host               '())
	 (display            (xlib:open-display       host))
	 (screen             (first (xlib:display-roots display)))
	 (black              (xlib:screen-black-pixel screen))
	 (white              (xlib:screen-white-pixel screen))
	 (root-window        (xlib:screen-root        screen))
	 ;; take the most preferred font available
	 (favorite-font-available
	              (first
		       (reduce #'append
			       (mapcar #'(lambda (x) (xlib:list-font-names
						      display x))
				       *monitor-gui-font-preference-list*))))
	 (font        (xlib:open-font display  favorite-font-available))
	 (width       (* (+ 1 10 7 26 4 2)                        ; a b c d e
			 (xlib:max-char-width font)))
	 (height      (* (+ 25 1)                                 ; lines 
			 (floor (* 1.4 (xlib:font-ascent font))))); line spacing
	 ;; X11 colors (see `/etc/X11/rgb.txt')
	 (gcontexts   (make-hash-table :test 'equal))
	 (gtext       (xlib:alloc-color
		       (xlib:window-colormap root-window) "black"))
	 (gback       (xlib:alloc-color
		       (xlib:window-colormap root-window) "LightYellow"))
	 (gfill       (xlib:alloc-color
		       (xlib:window-colormap root-window) "RoyalBlue"))
	 (gedge       (xlib:alloc-color
		       (xlib:window-colormap root-window) "YellowGreen"))
	 (net_wm_name      (xlib:intern-atom display "_NET_WM_NAME"))
	 (net_wm_icon_name (xlib:intern-atom display "_NET_WM_ICON_NAME"))
	 (window      (create-window-for-gui root-window width height gback))
	 (window-list (list window))
	 (time-out    *monitor-poll-sec*))
    (setf (gethash 'text  gcontexts)
	  (xlib:create-gcontext
	   :drawable      root-window
	   :font          font
	   :background    gback
	   :foreground    gtext))
    (setf (gethash 'fill gcontexts)
	  (xlib:create-gcontext
	   :drawable      root-window
	   :background    gback
	   :foreground    gfill))
    (setf (gethash 'edge  gcontexts)
	  (xlib:create-gcontext
	   :drawable      root-window
	   :background    gback
	   :foreground    gedge))
    (unwind-protect
	(loop named gui       ; user terminates with button-press (see below)
	  for language-code-list = (mirror-language-list :put put)
	  as  time-out =  (if *monitor-gui-one-window-per-language*
			      (max (ceiling *monitor-poll-sec*
					    (length language-code-list))
				   *monitor-poll-sec-min*)
			    *monitor-poll-sec*)
	  ;; create the correct number of windows
	  when *monitor-gui-one-window-per-language*
	  do 
	    (loop named window-create-destroy
	      with len-lang = (length language-code-list)
	      with len-wind = (length window-list)
	      for i from 0 to (1- (max len-lang len-wind))
	        ;; if user dropped languages, destroy excess windows
	        when (> i (1- len-lang))
	          do (xlib:destroy-window (first (last window-list)))
		  and do (setf window-list (butlast window-list))
		  end
	        ;; if user added languages, append more windows
	        when (> i (1- len-wind))
	          do (setf window-list
			   (append window-list 
				   (list
				    (create-window-for-gui
				     root-window width height gback)))))
	  ;; map (display) windows
	  do
	  (loop named window-map
	    for language-code in language-code-list
	    as  wind in window-list by (if *monitor-gui-one-window-per-language*
					   #'cdr
					 #'identity)
	    do
	      (monitor-data-collect language-code :put put)
	      (debug-message-start "event-case")
	      (xlib:map-window wind)
	      (xlib:event-case (display :discard-p      t
					:force-output-p t
			                :timeout        time-out)
			       (:button-press () nil)
			       (:destroy-window ()              ; <termination>
						(return-from gui t) t)
			       (:exposure (count)
					  (when (zerop count)
					    (monitor-mode-gui-show wind
								   gcontexts))
					  nil)
			       (:key-press ()                   ; <termination>
					   (return-from gui t) t)
			       )
	      (xlib:send-event wind :exposure nil)
	      (debug-message-value "event-case" (formatted-timestamp)))))
    (put-message-start (_ "cleanup"))
    (xlib:display-finish-output display)
    (loop
       for wind in window-list
       do (xlib:destroy-window wind))
    (loop
       for g being each hash-value of gcontexts
       do (xlib:free-gcontext g))
    (loop
       while (xlib:discard-current-event display))
    (put-message-done (_ "cleanup"))
    (xlib:close-display display)
    (put-message-done  (_ "attempting gui mode"))))

(defun create-window-for-gui (root-window width height gback)
  "Create window"
  (let ((wind  (xlib:create-window
		:parent        root-window
		:x             0
		:y             0
		:width         width
		:height        height
		:background    gback
		:event-mask    (xlib:make-event-mask 
				:button-press
				:exposure
				:key-press
				:structure-notify))))
    (xlib:change-property wind
			  :wm_name   "WP-MIRROR - Monitor Mode"
			  :string    8
			  :transform #'char-code)
    (xlib:change-property wind
			  :_net_wm_name   "WP-MIRROR - Monitor Mode"
			  :string    8
			  :transform #'char-code)
    (xlib:change-property wind
			  :wm_icon_name   "wp-mirror"
			  :string    8
			  :transform #'char-code)
    (xlib:change-property wind
			  :_net_wm_icon_name   "wp-mirror"
			  :string    8
			  :transform #'char-code)
    wind))

(defun monitor-mode-gui-show (my-window gcontexts &key (put nil))
  "Showing status in gui format"
  (let* ((w         my-window)
	 (ww        (xlib:drawable-width   w))
	 (wh        (xlib:drawable-height  w))
	 (wr        (xlib:drawable-root    w))
	 (g         gcontexts)
	 (gt        (gethash 'text  g))
	 (f         (xlib:gcontext-font   gt))
	 (fa        (xlib:font-ascent      f))
	 (line      0))
    ;; clear window first (lest new stuff get written on top of old stuff)
    (xlib:clear-area w)
    ;; timestamp
    (let* ((str (report-timestamp *report*))
	   (x   (- ww (xlib:text-extents f str)))
	   (y   (* 1 fa)))
      (xlib:draw-glyphs w gt x y str))

    ;; language code
    (monitor-mode-gui-show-line w g (incf line)
      :d (string-upcase (report-language *report*)))

    ;; dumpfile
    (monitor-mode-gui-show-line w g (incf line)
      :a "Dumpfile:"
      :d (when (null (report-dump-file-null     *report*))
	   (report-file-dump-name *report*)))

    (monitor-mode-gui-show-line w g (incf line)
      :b "downloading"
      :c (format-integer-for-human (report-file-dump-size *report*))
      :d (cond ((report-dump-file-null        *report*) 0)
	       ((report-dump-file-downloading *report*) :pulse)
	       (t 1)))

    (monitor-mode-gui-show-line w g (incf line)
      :b "validating"
      :d (if (report-dump-file-validating *report*) :pulse
	   (if (> (report-file-dump-size *report*) 0) 1 0)))
    
    (monitor-mode-gui-show-line w g (incf line)
      :b "decompressing"
      :c (format-integer-for-human (report-file-dump-xml-size *report*))
      :d (if (report-dump-file-decompressing *report*) :pulse 
	   (if (> (report-file-dump-xml-size *report*) 0) 1 0)))

    ;; estimates
    (monitor-mode-gui-show-line w g (incf line)
      :a "Estimates:")

    (monitor-mode-gui-show-line w g (incf line)
      :b "counting pages"
      :c (format nil "~10:d" (report-page-count *report*))
      :d (if (report-dump-file-page-counting *report*) :pulse
	   (if (> (report-page-count *report*) 0) 1 0)))

    (monitor-mode-gui-show-line w g (incf line)
      :b "counting image files"
      :c (format nil "~10:d" (report-image-count *report*))
      :d (if (report-dump-file-image-counting *report*) :pulse 
	   (if (> (report-image-count *report*) 0) 1 0)))

    ;; chunks
    (monitor-mode-gui-show-line w g (incf line)
      :a "Chunks:")

    (monitor-mode-gui-show-line w g (incf line)
      :b "generating xchunks"
      :c (format nil "~10:d" (report-xchunks-generated *report*))
      :d (report-xchunks-generated-progress *report*))

    (monitor-mode-gui-show-line w g (incf line)
      :b "generating ichunks"
      :c (format nil "~10:d" (report-ichunks-generated *report*))
      :d (report-ichunks-generated-progress *report*))

    ;; images
    (monitor-mode-gui-show-line w g (incf line)
      :a "Images:")

    (monitor-mode-gui-show-line w g (incf line)
      :b "downloading"
      :c (format nil "~10:d" (+ (report-ichunks-done      *report*)
				(report-ichunks-failed    *report*)))
      :d (report-ichunks-progress *report*))

    (monitor-mode-gui-show-line w g (incf line)
      :b "counting"
      :c (format nil "~10:d" (report-images-sum  *report*))
      :d (report-images-count-progress *report*))

    (monitor-mode-gui-show-line w g (incf line)
      :b "validating"
      :c (format nil "~10:d" (report-images-valid  *report*))
      :d (report-images-valid-progress *report*))

    (monitor-mode-gui-show-line w g (incf line)
      :b "importing"
      :c (format nil "~10:d" (report-images-imported  *report*))
      :d (report-images-imported-progress *report*))

    ;; articles
    (monitor-mode-gui-show-line w g (incf line)
      :a "Articles:")

    (monitor-mode-gui-show-line w g (incf line)
      :b "importing          "
      :c (format nil "~10:d" (report-pages-imported    *report*))
      :d (report-xchunks-progress *report*))

    ;; rule
    ;(monitor-mode-gui-show-rule w g (incf line))
    ;(incf line)

    ;; InnoDB
    (monitor-mode-gui-show-line w g (incf line)
      :a "InnoDB:")

    (monitor-mode-gui-show-line w g (incf line)
      :b "buffer pool pages dirty" 
      :c (format nil "~10:d" (report-innodb-buffer-pool-pages-dirty
				      *report*))
      :d (report-innodb-buffer-pool-pages-dirty-progress *report*))

    (monitor-mode-gui-show-line w g (incf line)
      :b "buffer pool pages insert"
      :c (format nil "~10:d" (report-innodb-buffer-pool-pages-insert *report*))
      :d (report-innodb-buffer-pool-pages-insert-progress *report*))

    (monitor-mode-gui-show-line w g (incf line)
      :b "buffer pool pages misc"
      :c (format nil "~10:d" (report-innodb-buffer-pool-pages-misc *report*))
      :d (report-innodb-buffer-pool-pages-misc-progress *report*))
    ;; System
    (monitor-mode-gui-show-line w g (incf line)
      :a "System:")

    (monitor-mode-gui-show-line w g (incf line)
      :b "images partition free"
      :c (format-integer-for-human (report-partition-images-free *report*))
      :d (report-partition-images-free-to-size-ratio *report*))
    (monitor-mode-gui-show-line w g (incf line)
      :b "InnoDB partition free"
      :c (format-integer-for-human (report-partition-innodb-free *report*))
      :d (report-partition-innodb-free-to-size-ratio *report*))
    ))

(defun monitor-mode-gui-show-rule (w gcontexts l)
  "Showing a rule under line l"
  (let* ((ww       (xlib:drawable-width   w))
	 (g        gcontexts)
	 (ge       (gethash 'edge  g))
	 (gf       (gethash 'fill  g))
	 (gt       (gethash 'text  g))
	 (f        (xlib:gcontext-font  gt))
	 (fa       (xlib:font-ascent     f))
	 (fiws     (xlib:text-extents f " ")); inter-word space
	 (lh       (floor (* 1.4 fa)))       ; line spacing
	 (ax       (+       fiws))           ; col a x
	 (ay       (* l lh))
	 (y        (- ay (floor (* lh 0.3)))))
    (xlib:draw-line w gt ax y (- ww ax) y)))

(defun monitor-mode-gui-show-line (w gcontexts l &key 
				     (a nil) (b nil) (c nil) (d nil) (put nil))
  "Showing one line for monitor screen mode"
  ;; we very carefully define (x,y,w,h) for each cell (a,b,c,d,e)
  (let* ((d        (if (numberp d)
		       (max 0 (min d 1.0))   ; [0,1]
		     d))                     ; string
	 (e        (when (numberp d)
		     (format nil "~5,2f%" (* d 100)))) ; percent progress
	 (g        gcontexts)
	 (ge       (gethash 'edge  g))
	 (gf       (gethash 'fill  g))
	 (gt       (gethash 'text  g))
	 (f        (xlib:gcontext-font  gt))
	 (fa       (xlib:font-ascent     f))
	 (fd       (xlib:font-direction  f)) ; used for I18N
	 (fmcw     (xlib:max-char-width  f)) 
	 (fiws     (xlib:text-extents f " ")); inter-word space
	 (lh       (floor (* 1.4 fa)))       ; line spacing
	 (aw       (*  1 fmcw))              ; col a width
	 (bw       (* 10 fmcw))              ; col b width
	 (cw       (*  7 fmcw))
	 (dw       (* 26 fmcw))
	 (dw-bar   (cond ((numberp d) 
			  (floor (* dw d))) ; progress bar - filled
			 ((eql d :pulse)
			  (floor (* dw (/ (get-decoded-time) 60)))))) ; pulsed
	 (ew       (*  4 fmcw))
	 (ax       (+       fiws))           ; col a x
	 (bx       (+ ax aw fiws))           ; col b x
	 (cx       (+ bx bw fiws))
	 (cx-just  (when (stringp c)
		     (- (+ cx cw) (xlib:text-extents f c)))) ; right justify
	 (dx       (+ cx cw fiws))
	 (ex       (+ dx dw fiws))
	 (ex-just  (when (stringp e)
		     (- (+ ex ew) (xlib:text-extents f e)))) ; right justify
	 (ah       lh)                       ; col a height
	 (bh       lh)                       ; col b height
	 (ch       lh)
	 (dh       lh)
	 (dh-bar   fa)        ; progress bar thinner for readability
	 (eh       lh)
	 (ay       (* l lh))                 ; col a y
	 (by       ay)                       ; col b y
	 (cy       ay)
	 (dy       ay)
	 (dy-bar   (- ay fa)) ; progress bar thinner for readablity
	 (ey       ay))
    (unless (null a) (xlib:draw-glyphs    w gt ax      ay a))
    (unless (null b) (xlib:draw-glyphs    w gt bx      by b))
    (unless (null c) (xlib:draw-glyphs    w gt cx-just cy c))       ; justify
    (cond ((stringp d) (xlib:draw-glyphs  w gt dx      dy d))
	  ((and (numberp d) (eql d 1))
	   (xlib:draw-rectangle w gf dx dy-bar dw-bar dh-bar   t))  ; fill
	  ((or (numberp d) (eql d :pulse))
	   (xlib:draw-rectangle w ge dx dy-bar dw     dh-bar nil)   ; edge
	   (xlib:draw-rectangle w gf dx dy-bar dw-bar dh-bar   t))  ; fill
	  (t nil))
    (cond ((eql d :pulse) (xlib:draw-glyphs    w gt ex      ey "pulse"))
          ((and (stringp e) (string= e " 0.00%")) nil)
          ((and (stringp e) (string= e "100.00%"))
	                  (xlib:draw-glyphs    w gt ex      ey "done"))
          ((stringp e)    (xlib:draw-glyphs    w gt ex-just ey e))) ; justify
    ))

(defun monitor-mode-screen (&key (put nil))
  "Running the monitor in screen mode"
  (put-message-start (_ "attempting screen mode"))
  (unwind-protect
      (screen:with-window
       (defparameter *monitor-screen-stream* screen:*window*)
       (multiple-value-bind (height width)
	   (screen:window-size *monitor-screen-stream*)
	 (defparameter *monitor-screen-height* height)
	 (defparameter *monitor-screen-width*  width)
	 (when (and (> *monitor-screen-height* 0)
		    (> *monitor-screen-width*  0))
	   (loop named screen    ; user terminates with Control-C 
	     do
	     (loop
	       for language-code in (mirror-language-list :put put)
	       do
	         (monitor-data-collect language-code :put put)
		 (monitor-mode-screen-show)
		 (sleep *monitor-poll-sec*))))))
    (put-message-done (_ "attempting screen mode"))))

(defun monitor-mode-screen-show (&key (put nil))
  "Showing status in screen format"
  (let ((w    *monitor-screen-stream*)
	(wh   (1- *monitor-screen-height*))
	(ww   (1- *monitor-screen-width* ))
	(line 0))

    ;; clear screen first
    (screen:window-cursor-off  w)
    (screen:clear-window       w)

    ;; borders
    (monitor-mode-screen-show-box)

    ;; caption
    (monitor-mode-screen-show-line line
      :a "WP-MIRROR"
      :d "MONITOR MODE")
    (screen:set-window-cursor-position      w  0 (- ww 1
			(length (report-timestamp *report*))))
    (princ (report-timestamp *report*) w)

    ;; language code
    (monitor-mode-screen-show-line (incf line)
      :d (string-upcase (report-language *report*)))

    ;; dump file
    (monitor-mode-screen-show-line (incf line)
      :a "Dumpfile:            "
      :d (if (null (report-dump-file-null *report*))
	     (report-file-dump-name *report*)
	   (format nil "~a" (language-code-to-database-name
			     (report-language *report*)))))
	   
    (monitor-mode-screen-show-line (incf line)
      :b "downloading        "
      :c (format nil "~10@a" (format-integer-for-human
			      (report-file-dump-size *report*)))
      :d (cond ((report-dump-file-null        *report*) 
		(format-progress-bar-string 0))
	       ((report-dump-file-downloading *report*) :pulse)
	       (t (format-progress-bar-string 1))))

    (monitor-mode-screen-show-line (incf line)
      :b "validating         " 
      :d (if (report-dump-file-validating *report*)
	     (format-progress-bar-pulse)
	   (if (> (report-file-dump-size *report*) 0)
	       (format-progress-bar-string 1)
	     (format-progress-bar-string 0))))

    (monitor-mode-screen-show-line (incf line)
      :b "decompressing      " 
      :c (format nil "~10@a" (format-integer-for-human 
			      (report-file-dump-xml-size *report*)))
      :d (if (report-dump-file-decompressing *report*)
	     (format-progress-bar-pulse)
	   (if (> (report-file-dump-xml-size *report*) 0)
	       (format-progress-bar-string 1)
	     (format-progress-bar-string 0))))

    ;; estimates
    (monitor-mode-screen-show-line (incf line)
      :a "Estimates:           ")

    (monitor-mode-screen-show-line (incf line)
      :b "counting pages     " 
      :c (format nil "~10:d" (report-page-count *report*))
      :d (if (report-dump-file-page-counting *report*)
	     (format-progress-bar-pulse)
	   (if (> (report-page-count *report*) 0)
	       (format-progress-bar-string 1)
	     (format-progress-bar-string 0))))

    (monitor-mode-screen-show-line (incf line)
      :b "counting image files"
      :c (format nil "~10:d" (report-image-count *report*))
      :d (if (report-dump-file-image-counting *report*)
	     (format-progress-bar-pulse)
	   (if (> (report-image-count *report*) 0)
	       (format-progress-bar-string 1)
	     (format-progress-bar-string 0))))

    ;; chunks
    (monitor-mode-screen-show-line (incf line)
      :a "Chunks:              ")

    (monitor-mode-screen-show-line (incf line)
      :b "generating xchunks "
      :c (format nil "~10:d" (report-xchunks-generated *report*))
      :d (format-progress-bar-string
	  (report-xchunks-generated-progress *report*)))

    (monitor-mode-screen-show-line (incf line)
      :b "generating ichunks "
      :c (format nil "~10:d" (report-ichunks-generated *report*))
      :d (format-progress-bar-string 
	  (report-ichunks-generated-progress *report*)))

    ;; images
    (monitor-mode-screen-show-line (incf line)
      :a "Images:              ")

    (monitor-mode-screen-show-line (incf line)
      :b "downloading        "
      :c (format nil "~10:d" (+ (report-ichunks-done      *report*)
				(report-ichunks-failed    *report*)))
      :d (format-progress-bar-string (report-ichunks-progress *report*)))

    (monitor-mode-screen-show-line (incf line)
      :b "counting           "
      :c (format nil "~10:d" (report-images-sum *report*))
      :d (format-progress-bar-string
	  (report-images-count-progress *report*)))

    (monitor-mode-screen-show-line (incf line)
      :b "validating         "
      :c (format nil "~10:d" (report-images-valid *report*))
      :d (format-progress-bar-string
	  (report-images-valid-progress *report*)))

    (monitor-mode-screen-show-line (incf line)
      :b "importing           "
      :c (format nil "~10:d" (report-images-imported *report*))
      :d (format-progress-bar-string
	  (report-images-imported-progress *report*)))

    ;; articles
    (monitor-mode-screen-show-line (incf line)
      :a "Articles:          ")

    (monitor-mode-screen-show-line (incf line)
      :b "importing          "
;;      :c (format nil "~10:d" (report-xchunks-done      *report*))
      :c (format nil "~10:d" (report-pages-imported      *report*))
      :d (format-progress-bar-string (report-xchunks-progress *report*)))

    ;; InnoDB
    (monitor-mode-screen-show-line (incf line)
      :a "InnoDB:              ")

    (monitor-mode-screen-show-line (incf line)
      :b "buffer pool dirty  " 
      :c (format nil "~10:d" (report-innodb-buffer-pool-pages-dirty *report*))
      :d (format-progress-bar-string
	  (report-innodb-buffer-pool-pages-dirty-progress *report*)))

    (monitor-mode-screen-show-line (incf line)
      :b "buffer pool insert "
      :c (format nil "~10:d" (report-innodb-buffer-pool-pages-insert *report*))
      :d (format-progress-bar-string
	  (report-innodb-buffer-pool-pages-insert-progress *report*)))

    (monitor-mode-screen-show-line (incf line)
      :b "buffer pool misc   "
      :c (format nil "~10:d" (report-innodb-buffer-pool-pages-misc *report*))
      :d (format-progress-bar-string
	  (report-innodb-buffer-pool-pages-misc-progress *report*)))
    ;; System
    (monitor-mode-screen-show-line (incf line)
      :a "System:")

    (monitor-mode-screen-show-line (incf line)
      :b "images partition free"
      :c (format nil "~10@a" (format-integer-for-human 
			      (report-partition-images-free *report*)))
      :d (format-progress-bar-string
	  (report-partition-images-free-to-size-ratio *report*)))
    (monitor-mode-screen-show-line (incf line)
      :b "InnoDB partition free"
      :c (format nil "~10@a" (format-integer-for-human 
			      (report-partition-innodb-free *report*)))
      :d (format-progress-bar-string
	  (report-partition-innodb-free-to-size-ratio *report*)))
    
    ;; park cursor in corner
    (screen:set-window-cursor-position      w wh ww)
))

(defun monitor-mode-screen-show-box (&key (put nil))
  "Showing box bordering the screen"
  (let ((w  *monitor-screen-stream*)
	(wh (1- *monitor-screen-height*))
	(ww (1- *monitor-screen-width* )))
    ;; edges
    (loop 
      for i of-type integer from 0 to ww do 
      (screen:set-window-cursor-position    w  0  i) (princ "-" w)
      (screen:set-window-cursor-position    w wh  i) (princ "-" w))
    (loop 
      for i of-type integer from 1 to wh do
      (screen:set-window-cursor-position    w  i  0) (princ "|" w)
      (screen:set-window-cursor-position    w  i ww) (princ "|" w))
    ;; corners
    (screen:set-window-cursor-position      w  0  0) (princ "+" w)
    (screen:set-window-cursor-position      w wh  0) (princ "+" w)
    (screen:set-window-cursor-position      w  0 ww) (princ "+" w)
    (screen:set-window-cursor-position      w wh ww) (princ "+" w)))

(defun monitor-mode-screen-show-line (y &key (a nil) (b nil) (c nil) (d nil))
  "Showing one line for monitor screen mode"
  ;; left-align four strings (a,b,c,d) on columns (a-col,b-col,c-col,d-col)
  (let ((w  *monitor-screen-stream*)
	(wh (1- *monitor-screen-height*))
	(a-col   2)
	(b-col   4)
	(c-col  23)
	(d-col  34))
    (when (< y wh)
      (unless (null a) (screen:set-window-cursor-position w y a-col)
	      (princ a w))
      (unless (null b) (screen:set-window-cursor-position w y b-col)
	      (princ b w))
      (unless (null c) (screen:set-window-cursor-position w y c-col) 
	      (princ c w))
      (unless (null d) (screen:set-window-cursor-position w y d-col) 
	      (princ d w)))
    ))

(defun monitor-mode-text (&key (put nil))
  "Running the monitor in text mode"
  (loop                  ; user terminates with Control-C 
    do
    (loop 
      for language-code in (mirror-language-list :put put)
      do
        (put-message-start (_"monitor mode text loop") :put t)
	(monitor-data-collect language-code :put put)
	(monitor-mode-text-show :put t)
	(sleep *monitor-poll-sec*))))

(defun monitor-mode-text-show (&key (put t))
  "Showing status in text format"
  (put-message-value (report-timestamp                      *report*)
		     (format nil "~a"
			     (language-code-to-database-name
			      (report-language              *report*))) :put t)
  (put-message-value (_ "dump file name")
		     (report-file-dump-name                 *report*)   :put t)
  (put-message-value (_ "downloading dump file")
		     (report-dump-file-downloading          *report*)   :put t)
  (put-message-value (_ "validating dump file")
		     (report-dump-file-validating           *report*)   :put t)
  (put-message-value (_ "decompressing dump file")
		     (report-dump-file-decompressing        *report*)   :put t)
  (put-message-value (_ "counting pages in dump file")
		     (report-dump-file-page-counting        *report*)   :put t)
  (put-message-value (_ "counting images in dump file")
		     (report-dump-file-image-counting       *report*)   :put t)
  (when (> (report-page-count *report*) 0)
    (put-message-value (_ "pages in dump file")
		       (report-page-count                   *report*)   :put t)
    (put-message-value (_ "xchunks expected")  
		       (report-chunk-count                  *report*)   :put t)
    (put-message-value (_ "xchunks generating/generated")
		       (format nil "~d/~d"
			       (report-xchunk-generating    *report*)
			       (report-xchunks-generated    *report*))  :put t))
  (when (> (report-ichunks-generated *report*) 0)
    (put-message-value (_ "ichunks generating/generated")
		       (format nil "~d/~d"
			       (report-ichunk-generating    *report*)
			       (report-ichunks-generated    *report*))  :put t)
    (put-message-value (_ "ichunks running")   
		       (report-ichunks-running              *report*)   :put t)
    (put-message-value (_ "ichunks done/failed/remaining")
		       (format nil "~d/~d/~d" 
			       (report-ichunks-done         *report*)
			       (report-ichunks-failed       *report*)
			       (report-ichunks-remaining    *report*))  :put t)
    (put-message-value (_ "ichunks progress") 
		       (format-progress-bar-string
			(report-ichunks-progress            *report*))  :put t))
  (put-message-value (_ "xchunks running")
		     (report-xchunks-running                *report*)   :put t)
  (put-message-value (_ "xchunks done/failed/remaining")
		     (format nil "~d/~d/~d" 
			     (report-xchunks-done           *report*)
			     (report-xchunks-failed         *report*)
			     (report-xchunks-remaining      *report*))  :put t)
  (put-message-value (_ "articles imported") 
		     (report-pages-imported                 *report*)   :put t)
  (put-message-value (_ "xchunks progress") 
		     (format-progress-bar-string
		      (report-xchunks-progress              *report*))  :put t)
  (when (> (report-ichunks-generated *report*) 0)
    (put-message-value (_ "counting download images")
		       (report-images-counting              *report*)   :put t)
    (put-message-value (_  "images counting")
		       (format nil "~d"
			       (report-images-sum           *report*))  :put t)
    (put-message-value (_  "images counting")
		       (format-progress-bar-string
			(report-images-count-progress       *report*))  :put t)
    (put-message-value (_  "images validating")
		       (format nil "~d"
			       (report-images-valid         *report*))  :put t)
    (put-message-value (_  "images validating")
		       (format-progress-bar-string
			(report-images-valid-progress       *report*))  :put t)
    (put-message-value (_  "images imported")
			(report-images-imported             *report*)   :put t)
    (put-message-value (_  "images imported")
		       (format-progress-bar-string
			(report-images-imported-progress    *report*))  :put t))
  ;; InnoDB
  (put-message-value (_ "InnoDB buffer pages dirty")
		     (format-progress-bar-string
		      (report-innodb-buffer-pool-pages-dirty-progress
		                                            *report*))  :put t)
  (put-message-value (_ "InnoDB buffer pages insert")
		     (format-progress-bar-string
		      (report-innodb-buffer-pool-pages-insert-progress
		                                            *report*))  :put t)
  (put-message-value (_ "InnoDB buffer pages misc")
		     (format-progress-bar-string
		      (report-innodb-buffer-pool-pages-misc-progress
		                                            *report*))  :put t)
  ;; System
  (put-message-value (_ "images partition free/size")
		     (format nil "~a/~a"
			     (format-integer-for-human
			      (report-partition-images-free *report*))
			     (format-integer-for-human
			      (report-partition-images-size *report*))) :put t)
  (put-message-value (_ "InnoDB partition free/size")
		     (format nil "~a/~a"
			     (format-integer-for-human
			      (report-partition-innodb-free *report*))
			     (format-integer-for-human
			      (report-partition-innodb-size *report*))) :put t)
  )

(defun main-monitor (&key (put nil))
  "Purpose: Main function when wp-mirror is in Monitor Mode."
  ;; 2) Determine monitor mode (:auto, :gui, :screen, :text)
  (put-message (_ "-----monitor-mode-begin-----------------------") :put t)
  (monitor-mode-gui-screen-or-text :put put)
  (case *monitor-mode*
    (:gui    (monitor-mode-gui    :put put))
    (:screen (monitor-mode-screen :put put))
    (:text   (monitor-mode-text   :put put)))
  (put-message (_ "-----monitor-mode-done------------------------") :put put))


;;;;--------------------------------------------------------------------------+
;;;; Main:                                                                    |
;;;;   Determine main mode of operation :first-mirror, :next-mirror, :monitor |
;;;;--------------------------------------------------------------------------+


(defun main (&key (put nil))
  (declare (optimize (safety 3)))
  (debug-message (_ "Hello, Dr. Miller."))
  ;; 0) Initialize
  (initialize-system              :put put)
  ;; 1) Assert prerequisites
  (assert-prerequisite-software   :put put)
  (assert-prerequisite-hardware   :put put)
  ;; 2) Main modes
  (case *main-mode*
	(:add
	 (add-language-and-die
          *main-language*         :put put))
	(:delete          
	 (delete-files-and-state-language-and-die
	  *main-language*         :put put))
	(:drop
	 (drop-language-and-die 
	  *main-language*         :put put))
	(:dump
	 (dump-database-language-and-die
	  *main-language*         :put put))
	((:first-mirror :next-mirror)
	 (main-mirror             :put put))
	(:monitor                     
	 (main-monitor            :put put))
	(:restore-default
	 (restore-default-and-die :put put))
	(:update
	 (update-database-language-and-die
	  *main-language*         :put put))
	(otherwise nil))
  ;; 3) Finalize
  (finalize-system                :put put)
  (debug-message (_ "Goodbye, Dr. Miller."))
  (ext:exit))

(main :put nil)
